#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "3d.h"
#include "brush.h"
#include "camera.h"
#include "clip.h"
#include "display.h"
#include "edbrush.h"
#include "error.h"
#include "keyhit.h"
#include "map.h"
#include "memory.h"
#include "message.h"
#include "mouse.h"
#include "newgroup.h"
#include "quest.h"
#include "undo.h"


static int ClipBrush(brush_t *b,plane_t *p,int flip)
{
   int j;
   brush_t *temp;
   brush_my_t t;

   t.numfaces=b->num_planes+1;
   t.faces=Q_malloc(sizeof(face_my_t)*t.numfaces);
   if (!t.faces)
   {
      HandleError("ClipBrush","Out of memory!");
      return 0;
   }
   for (j=0;j<b->num_planes;j++)
   {
      t.faces[j].misc=0;
      t.faces[j].numpts=0;
      t.faces[j].normal=b->plane[j].normal;
      t.faces[j].dist  =b->plane[j].dist;
      t.faces[j].tex   =b->plane[j].tex;
   }
// add new plane
   t.faces[j].misc=0;
   t.faces[j].numpts=0;

   if (flip)
   {
      t.faces[j].normal.x=-p->normal.x;
      t.faces[j].normal.y=-p->normal.y;
      t.faces[j].normal.z=-p->normal.z;
      t.faces[j].dist    =-p->dist;
   }
   else
   {
      t.faces[j].normal.x= p->normal.x;
      t.faces[j].normal.y= p->normal.y;
      t.faces[j].normal.z= p->normal.z;
      t.faces[j].dist    = p->dist;
   }
   t.faces[j].tex=b->plane[0].tex;

// create final brush
   temp=B_New(BR_NORMAL);
   if (!temp)
   {
      Q_free(t.faces);
      HandleError("ClipBrush","Out of memory!");
      return 0;
   }

   if (!BuildBrush(&t,temp))
   {
//      HandleError("ClipBrush","Error building brush!");
      return 0;
   }

   temp->EntityRef=M.WorldSpawn;
   temp->Group=FindVisGroup(M.WorldGroup);
   B_Link(temp);

   AddDBrush(temp);

   return 1;
}

static void C_clipplane_split(void)
{
   plane_t *p;
   brushref_t *br1;
   brushref_t *br;
   int split,error;

   br1=M.display.bsel;

	M.display.bsel=NULL;
	M.display.num_bselected=0;
	UpdateAllViewports();
	NewMessage("Select the clipping plane. Right click when finished.");
   do
   {
      CheckMoveKeys();
      UpdateMap();
		UpdateMouse();
//		ClearKeys();
	} while (mouse.button!=2);

   while (mouse.button==2) UpdateMouse();
	
	if (M.display.num_bselected!=1)
	{
		NewMessage("Exactly one clipping plane must be selected!");
		M.display.bsel = br1;
		ClearSelBrushes();
		return;
	}
   if (M.display.bsel->Brush->bt->type!=BR_CLIP)
      NewMessage("Warning! Brush isn't a clipping plane!");

   p=&M.display.bsel->Brush->plane[0];
   ClearSelBrushes();

   if ((fabs(p->normal.x)<0.01) &&
       (fabs(p->normal.y)<0.01) &&
       (fabs(p->normal.z)<0.01))
   {
      HandleError("clipplane_split","Invalid clipping plane!");
      M.display.bsel=br1;
      ClearSelBrushes();
      return;
   }

   M.display.bsel=br1;
   SUndo(UNDO_NONE,UNDO_DELETE);

   split=error=0;
   for (br=br1;br;br=br->Next)
   {
      if (!br->Brush->bt->flags&BR_F_CSG)
      {
         HandleError("clipplane_split","Can only split normal brushes");
         continue;
      }
   
      if (!ClipBrush(br->Brush,p,0)) error++;
      if (!ClipBrush(br->Brush,p,1)) error++;
      DeleteABrush(br->Brush);
      split+=2;
   }

   NewMessage("%i brushes split, %i not split.",split-error,error);

   M.display.bsel=br1;
   ClearSelBrushes();

   Cmd_SetRedraw(2);
}


static void C_clipplane_create(void)
{
   brush_t *b;
   plane_t *p;
   int i;
   int dx,dy,dz;

   b=B_New(BR_CLIP);
   if (!b)
   {
      HandleError("clipplane_create","Out of memory!");
      return;
   }

   b->num_planes=1;
   b->num_verts=3;
   b->num_edges=3;

   b->plane=p=Q_malloc(sizeof(plane_t));
   memset(p,0,sizeof(plane_t));

   b->verts=Q_malloc(sizeof(vec3_t)*3);
   b->tverts=Q_malloc(sizeof(vec3_t)*3);
   b->sverts=Q_malloc(sizeof(vec3_t)*3);

   b->edges=Q_malloc(sizeof(edge_t)*3);

   if (!p || !b->verts || !b->tverts || !b->sverts || !b->edges)
   {
      Q_free(p);
      Q_free(b->verts);
      Q_free(b->sverts);
      Q_free(b->tverts);
      Q_free(b->edges);
      Q_free(b);
      HandleError("clipplane_create","Out of memory!");
      return;
   }

   p->num_verts=3;
   p->verts=Q_malloc(sizeof(int)*3);
   if (!p->verts)
   {
      Q_free(p);
      Q_free(b->verts);
      Q_free(b->sverts);
      Q_free(b->tverts);
      Q_free(b->edges);
      Q_free(b);
      HandleError("clipplane_create","Out of memory!");
      return;
   }

   for (i=0;i<3;i++)
   {
      p->verts[i]=i;
      b->edges[i].startvertex=i;
      b->edges[i].endvertex  =(i+1)%3;

      b->verts[i].x=SnapPointToGrid(M.display.vport[M.display.active_vport].camera_pos.x);
      b->verts[i].y=SnapPointToGrid(M.display.vport[M.display.active_vport].camera_pos.y);
      b->verts[i].z=SnapPointToGrid(M.display.vport[M.display.active_vport].camera_pos.z);

      switch (i)
      {
      case 0:
         Move90(M.display.active_vport,MOVE_UP,&dx,&dy,&dz,64);
         break;
      case 1:
         Move90(M.display.active_vport,MOVE_LEFT,&dx,&dy,&dz,64);
         break;
      case 2:
         Move90(M.display.active_vport,MOVE_RIGHT,&dx,&dy,&dz,64);
         break;
      }

      b->verts[i].x+=dx;
      b->verts[i].y+=dy;
      b->verts[i].z+=dz;

      Move90(M.display.active_vport,MOVE_FORWARD,&dx,&dy,&dz,128);
      b->verts[i].x+=dx;
      b->verts[i].y+=dy;
      b->verts[i].z+=dz;
   }
   
	b->EntityRef = M.WorldSpawn;
	b->Group = FindVisGroup(M.WorldGroup);

   B_Link(b);

   Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"clipplane_create",C_clipplane_create,"Create a new clipping plane."},
{"clipplane_split",C_clipplane_split,"Split selected brushes with a clipping plane."},
};

static void RegCmds(void) __attribute__ ((constructor));
static void RegCmds(void)
{
   Cmd_Register(cmds,sizeof(cmds));
}

