#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "3d.h"
#include "entity.h"
#include "error.h"
#include "memory.h"
#include "quest.h"
#include "status.h"


/*
Figure out the path of a jump from a certain Quake 3-style jumppad. The
math is sort of like this:

(a,b,c) -> (d,e,f)

d=a+vx*t
e=b+vy*t
f=c+vz*t-400*t^2

vz-800*t=0
vz=800*t

f=c+800*t*t-400*t*t=c+400*t^2
t^2=(f-c)/400
t=(f-c)^0.5/20

The constants below control it. TODO: Ideally, they should be arguments to
this function so one may adjust them for different maps, but that'll
have to wait for now.

TOTO: You'd probably get better accuracy if you calculated destination
coordinates directly instead of stepping along the path.

*/

#define NUMPTS   32
#define STEP     0.1
#define GRAVITY  800.0
static void C_calc_quake3_jumppad(void)
{
   vec3_t to;
   const char *c1,*c2;
   entity_t *e1,*e2;

   vec3_t pos,vel;
   float t;

   int i;


   if (M.display.num_eselected!=2)
   {
      HandleError("calc_quake3_jumppad","Must have two entities selected!");
      return;
   }

   FreePts();
   M.pts=Q_malloc(sizeof(vec3_t)*NUMPTS);
   M.tpts=Q_malloc(sizeof(vec3_t)*NUMPTS);
   if (!M.pts || !M.tpts)
   {
      Q_free(M.pts);
      Q_free(M.tpts);
      M.pts=M.tpts=NULL;
      HandleError("calc_quake3_jumppad","Out of memory!");
      return;
   }
   status.draw_pts=1;
   M.npts=NUMPTS;

   c1=GetKeyValue(M.display.esel->Entity,"classname");
   c2=GetKeyValue(M.display.esel->Next->Entity,"classname");

   if (!strcmp(c1,"target_position"))
      e1=M.display.esel->Entity,e2=M.display.esel->Next->Entity;
   else
      e2=M.display.esel->Entity,e1=M.display.esel->Next->Entity;

   c1=GetKeyValue(e1,"origin");
   sscanf(c1,"%f %f %f",&to.x,&to.y,&to.z);
   pos=e2->center;

   t=to.z-pos.z;
   if (t<0)
      t=0;

   t=sqrt(t)/sqrt(GRAVITY/2.0);

   vel.z=t*GRAVITY;
   vel.x=(to.x-pos.x)/t;
   vel.y=(to.y-pos.y)/t;

   for (i=0;i<NUMPTS;i++)
   {
      M.pts[i]=pos;
      pos.x+=vel.x*STEP;
      pos.y+=vel.y*STEP;
      pos.z+=vel.z*STEP;

      vel.z-=GRAVITY*STEP;
   }

   Cmd_SetRedraw(2);
}

static cmd_reg_t cmds[]={
{"calc_quake3_jumppad",C_calc_quake3_jumppad,"Calculate the path of flight when one hits a Quake 3-style jumppad."},
};

static void RegCmds(void) __attribute__ ((constructor));
static void RegCmds(void)
{
   Cmd_Register(cmds,sizeof(cmds));
}

