#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "error.h"
#include "message.h"
#include "video.h"

#ifdef _UNIX
#include <sys/stat.h>
#endif


typedef struct
{
   char  id;
   char  version;
   char  encoding;
   char  bpp;
   short xmin;
   short ymin;
   short xmax;
   short ymax;
   short hres;
   short vres;
   char  pal[48];
   char  reserved;
   char  nplanes;
   short bpl;
   short palinfo;
   char  filler[58];
} PCXHeader /*__attribute__ ((packed))*/;


static int LastFileName=0;

static void C_screenshot(void)
{
   FILE *fp;
   char filename[80];
   char pal[768];
   int i, j;
   PCXHeader PCXHead;

#ifdef _UNIX
   struct stat buf;
#endif

   for (i=LastFileName;i<1000;i++)
   {
      sprintf(filename,"quest%03i.pcx",i);
#ifdef DJGPP
      if (!__file_exists(filename))
#endif
#ifdef _UNIX
      if (stat(filename,&buf))
#endif
         break;
   }
   LastFileName=i;

   fp = fopen(filename, "wb");

   if (!fp)
   {
      HandleError("screenshot","Unable to create %s",filename);
      return;
   }

   memset(&PCXHead,0,sizeof(PCXHead));

   PCXHead.id=10;
   PCXHead.version=5;
   PCXHead.encoding=1;
   PCXHead.bpp=8;

   PCXHead.xmin=PCXHead.ymin=0;
   PCXHead.xmax=video.ScreenWidth-1;
   PCXHead.ymax=video.ScreenHeight-1;

   PCXHead.hres=0x9600;
   PCXHead.vres=0x9600;

   PCXHead.nplanes=1;
   PCXHead.bpl=video.ScreenWidth;
   PCXHead.palinfo=2;

   if (fwrite(&PCXHead,1,sizeof(PCXHead),fp)!=sizeof(PCXHead))
   {
      fclose(fp);
      HandleError("screenshot","Unable to write to %s",filename);
      return;
   }

   for (i=0; i<video.ScreenHeight; i++)
   {
      unsigned char *buf;
      int last,num;

      buf=&video.ScreenBuffer[i*video.ScreenWidth];
      last=*buf++;
      num=1;
      for (j=1;j<video.ScreenWidth;j++,buf++)
      {
         if (last==*buf)
         {
            num++;
            if (num==63)
            {
               fputc(0xC0|num,fp);
               fputc(last,fp);
               num=0;
            }
         }
         else
         {
            if (num)
            {
               if ((num==1) && ((last & 0xC0)!=0xC0))
               {
                  fputc(last,fp);
               }
               else
               {
                  fputc(0xC0|num,fp);
                  fputc(last,fp);
               }
               num=0;
            }
            last=*buf;
            num=1;
         }
      }
      if (num)
      {
         if ((num==1) && ((last & 0xc0)!=0xC0))
         {
            fputc(last,fp);
         }
         else
         {
            fputc(0xC0|num,fp);
            fputc(last,fp);
         }
         num=0;
      }
   }

   fputc(12, fp);
   
/* GETS THE CURRENT PAL FROM VIDEO CARD
   (using quest.pal wouldn't hold correct gamma corrected values) */
   GetPal(pal);

   fwrite(pal,1,768,fp);

   fclose(fp);

	NewMessage("Screenshot taken.");
}

static cmd_reg_t scr_shot_cmds[]={
{"screenshot",C_screenshot,"Take a screenshot."},
};

static void RegCmds(void) __attribute__ ((constructor));
static void RegCmds(void)
{
   Cmd_Register(scr_shot_cmds,sizeof(scr_shot_cmds));
}

