/*
edent.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "edent.h"

#include "3d.h"
#include "brush.h"
#include "button.h"
#include "camera.h"
#include "display.h"
#include "edbrush.h"
#include "edentity.h"
#include "edit.h"
#include "edvert.h"
#include "entclass.h"
#include "entity.h"
#include "error.h"
#include "keyboard.h"
#include "map.h"
#include "memory.h"
#include "message.h"
#include "mouse.h"
#include "newgroup.h"
#include "popup.h"
#include "quest.h"
#include "qui.h"
#include "status.h"
#include "texlock.h"
#include "times.h"
#include "undo.h"
#include "video.h"


static entity_t *PasteEntity_Askws(entity_t *eo)
{
   QUI_window_t *w;
   unsigned char *tempbuf;
   int bp,b_keep,b_use1,b_use2,b_merge;
   int i,y;

	w=&Q.window[POP_WINDOW_1+Q.num_popups];
	w->size.x=450;
	w->size.y=140;
	w->pos.x=(video.ScreenWidth-w->size.x)/2;
	w->pos.y=(video.ScreenHeight-w->size.y)/2;

   PushButtons();
	b_merge=AddButtonText(0,0,0,"Merge");
   b_use1=AddButtonText(0,0,0,"Use existing");
   b_use2=AddButtonText(0,0,0,"Use new");
   b_keep=AddButtonText(0,0,0,"Keep both");

   button[b_merge].y=button[b_use1].y=button[b_use2].y=button[b_keep].y=
      w->pos.y+w->size.y-button[b_merge].sy-4;
   button[b_merge].x=w->pos.x+4;
   button[b_use1].x=button[b_merge].x+button[b_merge].sx+4;
   button[b_use2].x=button[b_use1 ].x+button[b_use1 ].sx+4;
   button[b_keep].x=button[b_use2 ].x+button[b_use2 ].sx+4;

	QUI_PopUpWindow(POP_WINDOW_1+Q.num_popups,"Pasting 'worldspawn'",&tempbuf);
	Q.num_popups++;

   y=w->pos.y+30;

   QUI_DrawStr(w->pos.x+5,y,6,15,0,FALSE,
      "You're attempting to paste a 'worldspawn' entity!");
   y+=20;

   QUI_DrawStr(w->pos.x+5,y,6,15,0,FALSE,
      "Existing:");
   QUI_DrawStr(w->pos.x+90,y,6,15,0,FALSE,
      "%i extra keys",M.WorldSpawn->numkeys-1);
   y+=20;

   QUI_DrawStr(w->pos.x+5,y,6,15,0,FALSE,
      "New:");
   QUI_DrawStr(w->pos.x+90,y,6,15,0,FALSE,
      "%i extra keys",eo->numkeys-1);
   y+=20;

   QUI_DrawStr(w->pos.x+5,y,6,15,0,FALSE,
      "If you're unsure about what to do, select merge.");
   y+=20;

	DrawButtons();

	RefreshPart(w->pos.x,w->pos.y,w->pos.x+w->size.x,w->pos.y+w->size.y);

	while (1)
	{
      UpdateMouse();
      bp=UpdateButtons();
      if (bp!=-1)
         break;
	}

	/* Pop down the window */
	Q.num_popups--;
	QUI_PopDownWindow(POP_WINDOW_1+Q.num_popups,&tempbuf);

	RemoveButton(b_merge);
	RemoveButton(b_use1);
	RemoveButton(b_use2);
	RemoveButton(b_keep);
   PopButtons();

	RefreshPart(w->pos.x,w->pos.y,w->pos.x+w->size.x,w->pos.y+w->size.y);

   if (bp==b_keep)
      return NULL;

   if (bp==b_use1)
      return M.WorldSpawn;

   if (bp==b_use2)
   {
      for (i=0;i<M.WorldSpawn->numkeys;i++)
      {
         Q_free(M.WorldSpawn->key[i]);
         Q_free(M.WorldSpawn->value[i]);
      }
      Q_free(M.WorldSpawn->key);
      Q_free(M.WorldSpawn->value);
      M.WorldSpawn->numkeys=0;
      M.WorldSpawn->key=M.WorldSpawn->value=NULL;
   }

   for (i=0;i<eo->numkeys;i++)
   {
      if (!GetKeyValue(M.WorldSpawn,eo->key[i]))
         SetKeyValue(M.WorldSpawn,eo->key[i],eo->value[i]);
   }

   return M.WorldSpawn;
}


static void RemoveT(entity_t *ign,const char *n1,const char *n2)
{
   entity_t *e;
   char      t[256];
   int       found;

   if (GetKeyValue(ign,n1))
   {
      strcpy(t,GetKeyValue(ign,n1));

      found=0;
      for (e=M.EntityHead;e;e=e->Next)
      {
         if (e!=ign)
         {
            if (GetKeyValue(e,n1))
            {
               found=1;
               break;
            }
         }
      }
      if (!found)
      {
         for (e=M.EntityHead;e;e=e->Next)
         {
            if (GetKeyValue(e,n2))
            {
               if (!strcmp(t,GetKeyValue(e,n2)))
               {
                  RemoveKeyValue(e,n2);
               }
            }
         }
      }
   }
}

void DeleteEntity(entity_t *e)
{
	brush_t *b;

   if (e==M.WorldSpawn)
   {
      HandleError("DeleteEntity","Tried to delete worldspawn.");
      return;
   }
   
   RemoveT(e,"target","targetname");
   RemoveT(e,"targetname","target");
   
   /* Update brush-entity ref */
   for (b=M.BrushHead; b; b=b->Next)
   {
      if (b->EntityRef == e)
         b->EntityRef=M.WorldSpawn;
   }
   
   if (e->Next != NULL) e->Next->Last = e->Last;
   if (e->Last != NULL) e->Last->Next = e->Next;
   if (e==M.EntityHead) M.EntityHead=e->Next;
   while (e->numkeys != 0)
      RemoveKeyValue(e, e->key[0]);
   Q_free(e);
   M.num_entities--;
}

