/*
camera.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "camera.h"

#include "config.h"
#include "entity.h"
#include "matrix.h"
#include "quest.h"
#include "s_3d.h"
#include "status.h"


/* Rotate a viewport from one direction to another slowly */
static void RotateDisplay(vec3_t from,vec3_t delta)
{
   int   i,j;
   float sr[3];
   float er[3];
   float dr[3];
   float cr[3];
   viewport_t *v;

   v=S_cur_vport;

   sr[0]=from.x;        sr[1]=from.y;        sr[2]=from.z;
   er[0]=sr[0]+delta.x; er[1]=sr[1]+delta.y; er[2]=sr[2]+delta.z;

   if (cv_cam_turn_frames->i<=0)
   {
      v->rot.x=er[0];
      v->rot.y=er[1];
      v->rot.z=er[2];
      return;
   }
   
   for (i=0;i<3;i++)
   {
      dr[i]=er[i]-sr[i];
      if (dr[i]>180)
         dr[i]-=360;
      if (dr[i]<-180)
         dr[i]+=360;

      dr[i]/=cv_cam_turn_frames->i;

      cr[i]=sr[i];
   }


   v->axis_aligned=0;
	for (i=0;i<cv_cam_turn_frames->i;i++)
	{
      v->rot.x=cr[0];
      v->rot.y=cr[1];
      v->rot.z=cr[2];
      S_UpdateCurViewport();
      for (j=0;j<3;j++)
         cr[j]+=dr[j];
   }
   v->axis_aligned=1;

   v->rot.x=er[0];
   v->rot.y=er[1];
   v->rot.z=er[2];
}


void GetRotValues(viewport_t *v,float *rx,float *ry,float *rz)
{
   if (v->axis_aligned)
   {
#define SNAP(x) \
   if (x>0) \
      x=floor((x+45)/90)*90; \
   else \
   if (x<0) \
      x=ceil((x-45)/90)*90;

      SNAP(v->rot.x)
      SNAP(v->rot.y)
      SNAP(v->rot.z)

#undef Snap
   }

   v->rot.x-=360*(int)(v->rot.x/360); if (v->rot.x<0) v->rot.x+=360;
   v->rot.y-=360*(int)(v->rot.y/360); if (v->rot.y<0) v->rot.y+=360;
   v->rot.z-=360*(int)(v->rot.z/360); if (v->rot.z<0) v->rot.z+=360;
   *rx=v->rot.x;
   *ry=v->rot.y;
   *rz=v->rot.z;
}


#define MOVEGEN() \
{ \
   GenerateIRotMatrix(m,-rot.x,-rot.y,-rot.z); \
\
   d[0]=d[1]=d[2]=0; \
   switch (dir) \
   { \
      case MOVE_UP: \
         d[1]=1; \
         break; \
      case MOVE_DOWN: \
         d[1]=-1; \
         break; \
\
      case MOVE_LEFT: \
         d[0]=-1; \
         break; \
      case MOVE_RIGHT: \
         d[0]=1; \
         break; \
\
      case MOVE_FORWARD: \
         d[2]=1; \
         break; \
      case MOVE_BACKWARD: \
         d[2]=-1; \
         break; \
   } \
\
   for (i=0;i<3;i++) \
   { \
      e[i]=d[0]*m[i][0]+ \
           d[1]*m[i][1]+ \
           d[2]*m[i][2]; \
   } \
\
   delta->x=e[0]*amt; if (fabs(delta->x-(int)delta->x)<0.001) delta->x=(int)delta->x; \
   delta->y=e[1]*amt; if (fabs(delta->y-(int)delta->y)<0.001) delta->y=(int)delta->y; \
   delta->z=e[2]*amt; if (fabs(delta->z-(int)delta->z)<0.001) delta->z=(int)delta->z; \
}

void Move(viewport_t *v,int dir, vec3_t *delta, int amt)
{
   matrix_t m;
   scalar_t d[3],e[3];
   int i;
   vec3_t rot;

   GetRotValues(v,&rot.x,&rot.y,&rot.z);

   MOVEGEN()
}

void Move90(viewport_t *v,int dir, vec3_t *delta, int amt)
{
   matrix_t m;
   scalar_t d[3],e[3];
   int i;
   vec3_t rot;

   GetRotValues(v,&rot.x,&rot.y,&rot.z);

#define Snap(x) \
   if (x>0) \
      x=floor((x+45)/90)*90; \
   else \
   if (x<0) \
      x=floor((x-45)/90)*90;

   Snap(rot.x);
   Snap(rot.y);
   Snap(rot.z);

#undef Snap

   MOVEGEN()
}

void MoveCamera(viewport_t *v,int dir)
{
	vec3_t d;

	Move(v,dir, &d, cv_cam_pan_speed->i);
	v->camera_pos.x += d.x;
	v->camera_pos.y += d.y;
	v->camera_pos.z += d.z;
}

void Cam_LookDown(void)
{
   if (S_cur_vport->axis_aligned)
   	RotateDisplay(S_cur_vport->rot,(vec3_t){+90,0,0});
   else
      S_cur_vport->rot.x+=5;
}

void Cam_LookUp(void)
{
   if (S_cur_vport->axis_aligned)
   	RotateDisplay(S_cur_vport->rot,(vec3_t){-90,0,0});
   else
      S_cur_vport->rot.x-=5;
}

void Cam_TurnLeft(void)
{
   if (S_cur_vport->axis_aligned)
   	RotateDisplay(S_cur_vport->rot,(vec3_t){0,0,-90});
   else
      S_cur_vport->rot.z-=5;
}

void Cam_TurnRight(void)
{
   if (S_cur_vport->axis_aligned)
   	RotateDisplay(S_cur_vport->rot,(vec3_t){0,0,+90});
   else
      S_cur_vport->rot.z+=5;
}

void Cam_RollLeft(void)
{
   if (!S_cur_vport->axis_aligned)
      S_cur_vport->rot.y-=5;
}

void Cam_RollRight(void)
{
   if (!S_cur_vport->axis_aligned)
      S_cur_vport->rot.y+=5;
}


static vec3_t az_vec;
static void AutoZoom_VP(viewport_t *v)
{
   int j;
   
   v->camera_pos.x = az_vec.x;
   v->camera_pos.y = az_vec.y;
   v->camera_pos.z = az_vec.z;
   
   for (j=0;j<100;j+=cv_cam_pan_speed->i)
      MoveCamera(v,MOVE_BACKWARD);
}

void AutoZoom(vec3_t w)
{
   az_vec=w;
   S_R_ForEachViewport(AutoZoom_VP);
}


static vec3_t s_pos,s_ang;

static void FindStartPos(void)
{
   entity_t *e;
   const char *c;
   float ang;

   s_ang.x=90;
   s_ang.y=0;
   s_ang.z=90;
   s_pos.x=s_pos.y=s_pos.z=0;
   for (e=M.EntityHead;e;e=e->Next)
   {
      if (!(c=GetKeyValue(e,"classname"))) continue;
      if (strcmp(c,"info_player_start")) continue;
      if (!(c=GetKeyValue(e,"origin"))) continue;
      sscanf(c,"%f %f %f",&s_pos.x,&s_pos.y,&s_pos.z);
      if ((c=GetKeyValue(e,"angle")))
      {
         sscanf(c,"%f",&ang);
         s_ang.x=90;
         s_ang.y=0;
         s_ang.z=90-ang;
      }
      return;
   }
}

static void SetStartPos(viewport_t *v)
{
   v->camera_pos=s_pos;
   v->rot=s_ang;
   
   v->zoom_amt=1;
   v->active=1;
}

void InitViewport(viewport_t *v)
{
   FindStartPos();
   SetStartPos(v);
}

void InitViewports(void)
{
   FindStartPos();
   S_R_ForEachViewport(SetStartPos);
}

