#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "camera.h"
#include "geom.h"
#include "group.h"
#include "map.h"
#include "memory.h"
#include "quest.h"
#include "s_3d.h"
#include "s_error.h"
#include "s_gui.h"
#include "texdef.h"


static void C_curve_create(void)
{
   brush_t *b;
   face_t *f;
   int i;
   vec3_t d;
   int x,y;

   int sizex,sizey,nx,ny;

   {
      int    i,j;
      char **keys;
      char **values;

      keys = (char **)Q_malloc(sizeof(char *) * 4);
      values = (char **)Q_malloc(sizeof(char *) * 4);
   
      if (!keys || !values)
      {
         S_Error("q3curve.create","Out of memory!");
         return;
      }
   
      for (i=0; i<4; i++)
      {
         keys[i] = (char *)Q_malloc(sizeof(char) * 30);
         values[i] = (char *)Q_malloc(sizeof(char) * 10);
         if (!keys[i] || !values[i])
         {
            S_Error("q3curve.create","Out of memory!");
            return;
         }
      }
   
      strcpy(keys[0], "X size per patch");
      strcpy(keys[1], "Y size per patch");
      strcpy(keys[2], "Patches in X axis");
      strcpy(keys[3], "Patches in Y axis");
      sprintf(values[0], "64");
      sprintf(values[1], "64");
      sprintf(values[2], "1");
      sprintf(values[3], "1");

      j=S_GUI_PopEntity("Create curve", keys, values, 4);
      
      sizex = atoi(values[0]);
      sizey = atoi(values[1]);
      nx = atoi(values[2]);
      ny = atoi(values[3]);

      for (i=0; i<4; i++)
      {
         Q_free(keys[i]);
         Q_free(values[i]);
      }
      Q_free(keys);
      Q_free(values);

      if (!j) return;
   }

   b=B_New(BRUSH_Q3_CURVE);
   if (!b)
   {
      S_Error("q3curve.create","Out of memory!");
      return;
   }

   b->x.q3c.sizex=nx*2+1;
   b->x.q3c.sizey=ny*2+1;

   b->num_faces=nx*ny;
   b->num_verts=(nx*2+1)*(ny*2+1);
   b->num_edges=0;

   b->faces=f=Q_malloc(sizeof(face_t)*b->num_faces);
   if (!b->faces)
   {
      S_Error("q3curve.create","Out of memory!");
      return;
   }
   memset(f,0,sizeof(face_t)*b->num_faces);

   b->verts=Q_malloc(sizeof(vec3_t)*b->num_verts);

   b->edges=NULL;

   b->x.q3c.a=Q_malloc(sizeof(vec2_t)*b->num_verts);

   if (!f || !b->verts || !b->x.q3c.a)
   {
      Q_free(f);
      Q_free(b->verts);
      Q_free(b->x.q3c.a);
      Q_free(b);
      S_Error("q3curve.create","Out of memory!");
      return;
   }

   InitTexdef(&b->tex);

   for (i=0,y=0;y<ny*2+1;y++)
   {
      for (x=0;x<nx*2+1;x++,i++)
      {
         b->verts[i]=S_cur_vport->camera_pos;
         SnapVecToGrid(&b->verts[i]);
   
         Move90(S_cur_vport,MOVE_RIGHT,&d,sizex/2*(x-nx));
         VEC_Add(b->verts[i],b->verts[i],d);
   
         Move90(S_cur_vport,MOVE_DOWN,&d,sizey/2*(y-ny));
         VEC_Add(b->verts[i],b->verts[i],d);
   
         Move90(S_cur_vport,MOVE_FORWARD,&d,128);
         VEC_Add(b->verts[i],b->verts[i],d);

         b->x.q3c.a[i].x=x*0.5;
         b->x.q3c.a[i].y=y*0.5;
      }
   }
   
   for (y=0,f=b->faces;y<ny;y++)
   {
      for (x=0;x<nx;x++,f++)
      {
         f->num_verts=9;
         f->verts=Q_malloc(sizeof(int)*f->num_verts);
         if (!f->verts)
         {
            Q_free(b->faces);
            Q_free(b->verts);
            Q_free(b);
            S_Error("q3curve.create","Out of memory!");
            return;
         }

         f->verts[0]= y*2   *(nx*2+1)+x*2+0;
         f->verts[1]= y*2   *(nx*2+1)+x*2+1;
         f->verts[2]= y*2   *(nx*2+1)+x*2+2;
         f->verts[3]=(y*2+1)*(nx*2+1)+x*2+0;
         f->verts[4]=(y*2+1)*(nx*2+1)+x*2+1;
         f->verts[5]=(y*2+1)*(nx*2+1)+x*2+2;
         f->verts[6]=(y*2+2)*(nx*2+1)+x*2+0;
         f->verts[7]=(y*2+2)*(nx*2+1)+x*2+1;
         f->verts[8]=(y*2+2)*(nx*2+1)+x*2+2;
      }
   }

   b->Entity = M.WorldSpawn;
   b->Group = GetDefaultGroup();

   B_Link(b);
   
   Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"q3curve.create",C_curve_create,{"Curve/Create..."},{},"Create a Quake 3-style curve."},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/crcurve","Create Quake 3-style curves."}};

PLUGIN_FOOT_CMDS

