#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "config.h"
#include "csg.h"
#include "edsel.h"
#include "group.h"
#include "memory.h"
#include "quest.h"
#include "s_error.h"
#include "s_message.h"
#include "status.h"


static void CSG_MakeHollow(brush_t *b)
{
   int i,j;
   brush_t *temp;
   cbrush_t t;

   for (i=0;i<b->num_faces;i++)
   {
      t.num_faces=b->num_faces+1;
      t.faces=Q_malloc(sizeof(cface_t)*t.num_faces);
      if (!t.faces)
      {
         S_Error("CSG Hollow","Out of memory!");
         return;
      }
      for (j=0;j<b->num_faces;j++)
      {
         if (!CSG_FaceToCFace(&t.faces[j],b,j))
         {
            S_Error("CSG Hollow","Out of memory!");
            return;
         }
      }

/* new plane */
      if (!CSG_FaceToCFace(&t.faces[j],b,i))
      {
         S_Error("CSG Hollow","Out of memory!");
         return;
      }
      t.faces[j].p.n.x =-t.faces[j].p.n.x;
      t.faces[j].p.n.y =-t.faces[j].p.n.y;
      t.faces[j].p.n.z =-t.faces[j].p.n.z;
      t.faces[j].p.dist=-t.faces[j].p.dist+cv_edit_snap_size->i;

      temp=CSG_BuildBrush(NULL,&t);
      if (!temp)
      {
         S_Error("CSG Hollow","Error building brush!");
         return;
      }
      
      temp->Entity=M.WorldSpawn;
      temp->Group=FindVisGroup(b->Group);
      B_Link(temp);
   }
}

static void C_hollow(void)
{
   bsel_t *b;
   int warn;

   if (!M.sel.num_bselected)
   {
      S_Error("CSG Hollow","No brushes selected!");
      return;
   }
   ClearSelVerts();

   warn=0;
	for (b=M.sel.bsel;b;b=b->Next)
	{
      if (b->Brush->bt->type!=BRUSH_NORMAL)
      {
         if (!warn)
         {
            S_NewMessage("Warning: Can only use normal brushes for CSG operations.");
            warn=1;
         }
      }
      else
      {
         CSG_MakeHollow(b->Brush);
         B_Unlink(b->Brush);
         B_Free(b->Brush);
      }
   }

   ClearSelBrushes();

   Cmd_SetRedraw(2);
}


static void CSG_MakeRoom(brush_t *b)
{
   int i,j;
   brush_t *temp;
   cbrush_t t;

   for (i=0;i<b->num_faces;i++)
   {
      t.num_faces=b->num_faces*2;
      t.faces=Q_malloc(sizeof(cface_t)*t.num_faces);
      if (!t.faces)
      {
         S_Error("CSG Room","Out of memory!");
         return;
      }
      for (j=0;j<b->num_faces;j++)
      {
         if (!CSG_FaceToCFace(&t.faces[j],b,j))
         {
            S_Error("CSG Room","Out of memory!");
            return;
         }
      }
/* add new planes */
      for (;j<t.num_faces;j++)
      {
         if (!CSG_FaceToCFace(&t.faces[j],b,j-b->num_faces))
         {
            S_Error("CSG Room","Out of memory!");
            return;
         }
   
         t.faces[j].p.dist-=cv_edit_snap_size->i;
         if (j-b->num_faces==i)
         {
            t.faces[j].p.n.x=-t.faces[j].p.n.x;
            t.faces[j].p.n.y=-t.faces[j].p.n.y;
            t.faces[j].p.n.z=-t.faces[j].p.n.z;
            t.faces[j].p.dist=-t.faces[j].p.dist;
         }
      }

/* Create final brush */
      temp=CSG_BuildBrush(NULL,&t);
      if (!temp)
      {
         S_Error("CSG_Room","Error building brush!");
         return;
      }

      temp->Entity=M.WorldSpawn;
      temp->Group=FindVisGroup(b->Group);
      B_Link(temp);
   }
}

static void C_room(void)
{
   bsel_t *b;
   int warn;
   
   if (!M.sel.num_bselected)
   {
      S_Error("CSG Room","No brushes selected!");
      return;
   }
   ClearSelVerts();

   warn=0;
	for (b=M.sel.bsel;b;b=b->Next)
	{
      if (b->Brush->bt->type!=BRUSH_NORMAL)
      {
         if (!warn)
         {
            S_NewMessage("Warning: Can only use normal brushes for CSG operations.");
            warn=1;
         }
      }
      else
      {
         CSG_MakeRoom(b->Brush);
         B_Unlink(b->Brush);
         B_Free(b->Brush);
      }
   }

   ClearSelBrushes();

   Cmd_SetRedraw(2);
}



static cmd_reg_t cmds[]={
{"csg.hollow",C_hollow,{"Brush/CSG/Hollow","h"},{},"CSG hollow",NULL},
{"csg.room",C_room,{"Brush/CSG/Room","r"},{},"CSG room",NULL},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/csg/rmhl","CSG make hollow/room command."}};

PLUGIN_FOOT_CMDS

