#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"


#include "qmap.h"

#include "s_texture.h"
#include "q3_jpg.h"
#include "game_reg.h"

#include "config.h"
#include "map.h"
#include "memory.h"
#include "s_dir.h"
#include "s_error.h"
#include "s_message.h"
#include "s_types.h"
#include "status.h"
#include "texcat.h"
#include "token.h"


/*
Total texels:
54330809 : 192
  207mb  276mb
  155mb  207mb
  103mb  138mb

10559784 : 160
   40mb   53mb
   30mb   40mb
   20mb   26mb

 9905081 : 128
          50mb

 3455180 :  96
   13mb   17mb
    9mb   13mb
    6mb    8mb
*/


static config_var_t
   *cv_texture_path,
   *cv_shader_path,
   *cv_entity_path;

config_var_t
   *cv_quake3_scale_above,
   *cv_quake3_scale_below;
   

typedef struct G_texdef_s
{
   unsigned int contents,flags,value;
} G_q3_texdef_t;


static int QT_Init(texdef_t *t)
{
   if (!t->g)
   {
      t->g=Q_malloc(sizeof(G_q3_texdef_t));
      if (!t->g) return 0;
   }
   memset(t->g,0,sizeof(G_q3_texdef_t));
   return 1;
}

static void QT_Free(texdef_t *t)
{
   if (t->g)
   {
      Q_free(t->g);
      t->g=NULL;
   }
}

static int QT_Copy(texdef_t *dst,const texdef_t *src)
{
   if (!src->g) return 1;
   if (!dst->g)
   {
      dst->g=Q_malloc(sizeof(G_q3_texdef_t));
      if (!dst->g) return 0;
   }
   *(dst->g)=*(src->g);
   return 1;
}


static int Q3_LoadTexInfo(texdef_t *t)
{
   if (!t->g)
   {
      t->g=Q_malloc(sizeof(G_q3_texdef_t));
      if (!t->g) return ERROR_NOMEM;
   }
   memset(t->g,0,sizeof(G_q3_texdef_t));

   if (TokenAvailable(0))
   {
      GETTOKEN(0,TOKEN_NUMBER);
      t->g->contents=atoi(token);

      GETTOKEN(0,TOKEN_NUMBER);
      t->g->flags=atoi(token);

      GETTOKEN(0,TOKEN_NUMBER);
      t->g->value=atoi(token);
   }

   return ERROR_NO;
}

static void Q3_SaveTexInfo(texdef_t *tex,FILE *fp)
{
   if (tex->g->contents ||
       tex->g->flags ||
       tex->g->value)
   {
      fprintf(fp," %i %i %i",
         tex->g->contents,
         tex->g->flags,
         tex->g->value);
   }
}

static game_reg_t G_Quake3=
{
   "Quake 3",
   {
      QMap_Load,
      QMap_Save
   },
   {
      QT_Init,
      QT_Free,
      QT_Copy
   },
};


typedef struct
{
   int real; /* 0 if not a shader */
   int disp_file;
   int light; /* 0 if no light */
   int flags;
#define SHADER_SKY         1
#define SHADER_LAVA        2
#define SHADER_WATER       4
#define SHADER_SLIME       8
#define SHADER_TRANS      16
#define SHADER_NODRAW     32
#define SHADER_NOTSOLID   64
#define SHADER_FOG       128
} shader_t;


#if TODO
static void Q3_ModifyFlags(void)
{
static const char *q3_contents[32]=
{
NULL           ,NULL           ,NULL           ,NULL           ,// 00000001
NULL           ,NULL           ,NULL           ,NULL           ,// 00000010
NULL           ,NULL           ,NULL           ,NULL           ,// 00000100
NULL           ,NULL           ,NULL           ,NULL           ,// 00001000
NULL           ,NULL           ,NULL           ,NULL           ,// 00010000
NULL           ,NULL           ,NULL           ,NULL           ,// 00100000
NULL           ,NULL           ,NULL           ,"Detail"       ,// 01000000
NULL           ,NULL           ,NULL           ,NULL           ,// 10000000
};
static const char *q3_flags[32]=
{
NULL           ,NULL           ,NULL           ,NULL           ,// 00000001
NULL           ,NULL           ,NULL           ,NULL           ,// 00000010
NULL           ,NULL           ,NULL           ,NULL           ,// 00000100
NULL           ,NULL           ,NULL           ,NULL           ,// 00001000
NULL           ,NULL           ,NULL           ,NULL           ,// 00010000
NULL           ,NULL           ,NULL           ,NULL           ,// 00100000
NULL           ,NULL           ,NULL           ,NULL           ,// 01000000
NULL           ,NULL           ,NULL           ,NULL           ,// 10000000
};

   Quake2Flags_ModifyFlags(q3_contents,q3_flags);
}
#endif


#if 0
static int Q3_TexBSPFlags(texdef_t *t)
{
   int i;
   tex_name_t *tn;
   q3_tn_int_t *sh;
   
   for (i=n_tnames,tn=tnames;i;i--,tn++)
      if (!stricmp(t->name,tn->name))
         break;

   if (!i) return 0;
   if (!tn->x) return 0;
   sh=(q3_tn_int_t *)tn->x;

   if (sh->flags&SHADER_SKY)       return TEX_FULLBRIGHT;

   i=0;
   if (sh->flags&SHADER_NODRAW)    i|=TEX_NODRAW;
   if (sh->flags&
       (SHADER_WATER|SHADER_SLIME|SHADER_LAVA|SHADER_FOG
       |SHADER_NOTSOLID|SHADER_TRANS)
      )
      i|=TEX_FULLBRIGHT|TEX_NONSOLID;

   return i;
}
#endif

static void tex_description(char *dest,const texture_t *t)
{
   int i;
   shader_t *sh;

   sh=t->g;

   if (!sh->real)
   {
      strcpy(dest,"raw image");
      return;
   }

   i=sprintf(dest,"shader ");
   
   if (sh->light)
      i+=sprintf(&dest[i],"light=%i ",sh->light);
   if (sh->flags&SHADER_SKY)
      i+=sprintf(&dest[i],"sky ");
   if (sh->flags&SHADER_WATER)
      i+=sprintf(&dest[i],"water ");
   if (sh->flags&SHADER_LAVA)
      i+=sprintf(&dest[i],"lava ");
   if (sh->flags&SHADER_SLIME)
      i+=sprintf(&dest[i],"slime ");
   if (sh->flags&SHADER_FOG)
      i+=sprintf(&dest[i],"fog ");
   if (sh->flags&SHADER_TRANS)
      i+=sprintf(&dest[i],"trans ");
   if (sh->flags&SHADER_NODRAW)
      i+=sprintf(&dest[i],"nodraw ");
   if (sh->flags&SHADER_NOTSOLID)
      i+=sprintf(&dest[i],"notsolid ");

   if (dest[0])
      dest[strlen(dest)-1]=0;
}


typedef struct
{
   s_u8 id_len  __attribute__ ((packed));
   s_u8 cm_type  __attribute__ ((packed));
   s_u8 image_type  __attribute__ ((packed));

   s_u16 cm_start  __attribute__ ((packed));
   s_u16 cm_len  __attribute__ ((packed));
   s_u8 cm_size  __attribute__ ((packed));

   s_u16 org_x  __attribute__ ((packed));
   s_u16 org_y  __attribute__ ((packed));
   s_u16 width  __attribute__ ((packed));
   s_u16 height  __attribute__ ((packed));
   s_u8 bpp  __attribute__ ((packed));

   s_u8 flags  __attribute__ ((packed));
} tga_header_t __attribute__ ((packed));

static void load_tga(const char *name,void *tex,const texture_t *t)
{ /* TODO: scaling could be optimized */
   FILE *f;
   int baseofs;
   unsigned char *data,*d;

   int sx,sy;
   int dsx,dsy;
   int scale;

   int i,j,k;
   int x,y;
   int c[3];
   
   tga_header_t h;


   f=fopen(name,"rb");
   if (!f)
   {
      S_NewMessage("quake3.load_tga: Can't open '%s'!",name);
      return;
   }

   baseofs=0;

   fseek(f,baseofs,SEEK_SET);
   fread(&h,1,sizeof(tga_header_t),f);

   sx=h.width;
   sy=h.height;

   dsx=sx;
   dsy=sy;

   scale=0;
   if (dsx>cv_quake3_scale_above->i && dsy>cv_quake3_scale_above->i) dsx/=2,dsy/=2,scale++;
   while (scale<3 && (dsx>cv_quake3_scale_below->i || dsy>cv_quake3_scale_below->i))
      dsx/=2,dsy/=2,scale++;

#ifdef COUNT_TEXEL
   count_texel(dsx,dsy);
   fclose(f);
   return;
#endif

   fseek(f,baseofs+sizeof(tga_header_t)+h.id_len,SEEK_SET);

   data=Q_malloc(dsx*dsy*3);
   if (!data)
   {
      fclose(f);
      S_NewMessage("quake3.load_tga: Out of memory, can't load '%s'!",name);
      return;
   }
   memset(data,0,dsx*dsy*3);

   switch (h.image_type)
   {
   case 2:
      if (h.bpp!=32 && h.bpp!=24)
      {
         Q_free(data);
         fclose(f);
         S_NewMessage("quake3.load_tga: Unsupported .tga: type=%i and bpp=%i",h.image_type,h.bpp);
         return;
      }
   
      for (i=0;i<h.width*h.height;i++)
      {
         y=i/h.width;
         x=i-y*h.width;
         y=h.height-y-1;

         x>>=scale;
         y>>=scale;

         c[2]=fgetc(f);
         c[1]=fgetc(f);
         c[0]=fgetc(f);
         
         if (h.bpp==32)
            fgetc(f);

         if (x<0 || x>=dsx || y<0 || y>=dsy)
            continue;
         
         d=&data[3* (x+ dsx*y ) ];

         *d++ +=c[0]>>(2*scale);
         *d++ +=c[1]>>(2*scale);
         *d++ +=c[2]>>(2*scale);
      }
      break;

   case 10:
      if (h.bpp!=32 && h.bpp!=24)
      {
         Q_free(data);
         fclose(f);
         S_NewMessage("quake3.load_tga: Unsupported .tga: type=%i and bpp=%i",h.image_type,h.bpp);
         return;
      }
      
      for (i=0;i<h.width*h.height;)
      {
         j=((k=fgetc(f))&127)+1;
         if (k&128)
         {
            c[2]=fgetc(f)>>(2*scale);
            c[1]=fgetc(f)>>(2*scale);
            c[0]=fgetc(f)>>(2*scale);
            if (h.bpp==32)
               fgetc(f);
               
            while (j--)
            {
               y=i/h.width;
               x=i-y*h.width;
               y=h.height-y-1;
      
               x>>=scale;
               y>>=scale;

               i++;
               
               if (x<0 || x>=dsx || y<0 || y>=dsy)
                  continue;
               
               d=&data[3* (x+ dsx*y ) ];
               
               *d++ +=c[0];
               *d++ +=c[1];
               *d++ +=c[2];
            }
         }
         else
         {
            while (j--)
            {
               c[2]=fgetc(f)>>(2*scale);
               c[1]=fgetc(f)>>(2*scale);
               c[0]=fgetc(f)>>(2*scale);
               if (h.bpp==32)
                  fgetc(f);

               y=i/h.width;
               x=i-y*h.width;
               y=h.height-y-1;
      
               x>>=scale;
               y>>=scale;

               i++;
               if (x<0 || x>=dsx || y<0 || y>=dsy)
                  continue;
               d=&data[3* (x+ dsx*y ) ];
               
               *d++ +=c[0];
               *d++ +=c[1];
               *d++ +=c[2];
            }
         }
      }
      break;

   default:
      Q_free(data);
      fclose(f);
      S_NewMessage("quake3.load_tga: Unsupported .tga: type=%i and bpp=%i",h.image_type,h.bpp);
      return;
   }

   S_T_AddData24(tex,sx,sy,dsx,dsy,data);

   Q_free(data);
   fclose(f);
}


static void load_tga_size(const char *name,void *tex,const texture_t *t)
{
   FILE *f;
   tga_header_t h;

   f=fopen(name,"rb");
   if (!f)
      return;
   fread(&h,1,sizeof(tga_header_t),f);

   S_T_AddSize(tex,h.width,h.height);

   fclose(f);
}


static char base_tex_dir[MAX_PATH];

typedef struct
{
   char name[MAX_PATH];
   int type; /* 0 if .tga, 1 if .jpg */
   int used;
} tex_file_t;

static tex_file_t *tex_files;
static int num_tex_files;


static void load_tex(void *tex,const texture_t *t)
{
   shader_t *s;
   tex_file_t *tf;
   char name[MAX_PATH];
   
   s=t->g;
   tf=&tex_files[s->disp_file];
   strcpy(name,base_tex_dir);
   strcat(name,tf->name);
   switch (tf->type)
   {
   case 0:
      strcat(name,".tga");
      load_tga(name,tex,t);
      break;
   case 1:
      strcat(name,".jpg");
      quake3_load_jpg(name,tex,t);
      break;
   default:
      S_NewMessage("Can't load '%s'!",t->name);
      break;
   }
}


static void load_tex_size(void *tex,const texture_t *t)
{
   shader_t *s;
   tex_file_t *tf;
   char name[MAX_PATH];
   
   s=t->g;
   tf=&tex_files[s->disp_file];
   strcpy(name,base_tex_dir);
   strcat(name,tf->name);
   switch (tf->type)
   {
   case 0:
      strcat(name,".tga");
      load_tga_size(name,tex,t);
      break;
   case 1:
      strcat(name,".jpg");
      quake3_load_jpg_size(name,tex,t);
      break;
   }
}


static int tex_file_cmp(const void *v1,const void *v2)
{
   const tex_file_t *t1=v1;
   const tex_file_t *t2=v2;
   return stricmp(t1->name,t2->name);
}

static void AddTexFile(const char *name,int type)
{
   tex_file_t *t;

   t=Q_realloc(tex_files,sizeof(tex_file_t)*(num_tex_files+1));
   if (!t)
   {
      S_Error("games.quake3","Out of memory!");
      return;
   }
   tex_files=t;
   t=&tex_files[num_tex_files++];
   strcpy(t->name,&name[strlen(base_tex_dir)]);
   t->name[strlen(t->name)-4]=0;
   t->type=type;
   t->used=0;
}

static void AddDir_r(const char *dirname)
{
   struct S_directory_s *dir;
   S_filedir_t f;
   char buf[MAX_PATH];

   strcpy(buf,dirname);
   strcat(buf,"*");
   dir=S_Dir_Open(buf,FILE_DIREC);
   if (!dir)
   {
      S_NewMessage("games.quake3: Can't open directory '%s'!",dirname);
      return;
   }
   while (S_Dir_Read(dir,&f))
   {
      if (f.name[0]=='.') continue;
      strcpy(buf,dirname);
      strcat(buf,f.name);
      strcat(buf,"/");
      AddDir_r(buf);
   }
   S_Dir_Close(dir);
   
   strcpy(buf,dirname);
   strcat(buf,"*.jpg");
   dir=S_Dir_Open(buf,FILE_NORMAL);
   if (!dir)
   {
      S_NewMessage("games.quake3: Can't open directory '%s'!",dirname);
      return;
   }
   while (S_Dir_Read(dir,&f))
   {
      strcpy(buf,dirname);
      strcat(buf,f.name);
      AddTexFile(buf,1);
   }
   S_Dir_Close(dir);

   strcpy(buf,dirname);
   strcat(buf,"*.tga");
   dir=S_Dir_Open(buf,FILE_NORMAL);
   if (!dir)
   {
      S_NewMessage("games.quake3: Can't open directory '%s'!",dirname);
      return;
   }
   while (S_Dir_Read(dir,&f))
   {
      strcpy(buf,dirname);
      strcat(buf,f.name);
      AddTexFile(buf,0);
   }
   S_Dir_Close(dir);
}


static int LookupFile(char *name)
{
   char *n;
   int i;

   if (strncmp(name,"textures/",9)) return -1;
   n=&name[9];
   if (!strncmp(&n[strlen(n)-4],".jpg",4) || !strncmp(&n[strlen(n)-4],".tga",4))
      n[strlen(n)-4]=0;
   for (i=0;i<num_tex_files;i++)
   {
      if (!stricmp(tex_files[i].name,n))
      {
         tex_files[i].used++;
         return i;
      }
   }
   return -1;
}


static tex_category_t *tc_shaders,*tc_nshaders;


static void LoadShaderFile(const char *fname)
{
   char name[256];
   int brace_level;
   int i;

   int disp_file;
   int disp_i;

   shader_t sh;
   shader_t *shader;

   texture_t *tex;
   

   if (!TokenFile(fname,TOKEN_C|TOKEN_ALLNAME,NULL))
   {
      S_Error("games.quake3.load_shader","Can't open '%s'!",fname);
      return;
   }

   while (TokenGet(1,-1))
   {
      disp_file=disp_i=-1;
      
      strcpy(name,token);
      i=LookupFile(token);
      if (i!=-1)
      {
         disp_i=1;
         disp_file=i;
      }

      TokenGet(1,-1);
      if (strcmp(token,"{"))
      {
         S_Error("games.quake3.load_shader","In '%s' line %i: Expected '{', got '%s'!",
            fname,token_linenum,
            token);
         TokenDone();
         return;
      }

      brace_level=1;
      memset(&sh,0,sizeof(sh));
      do
      {
         TokenGet(1,-1);
         if (!strcmp(token,"{"))
            brace_level++;
         else
         if (!strcmp(token,"}"))
            brace_level--;

/* general shader information */
         else
         if (!strcmp(token,"surfaceparm"))
         {
            TokenGet(1,-1);
            if (!strcmp(token,"fog"))
               sh.flags|=SHADER_FOG;
            else
            if (!strcmp(token,"nodraw"))
               sh.flags|=SHADER_NODRAW;
            else
            if (!strcmp(token,"nonsolid"))
               sh.flags|=SHADER_NOTSOLID;
            else
            if (!strcmp(token,"trans"))
               sh.flags|=SHADER_TRANS;
               
            else
            if (!strcmp(token,"lava"))
               sh.flags|=SHADER_LAVA;
            else
            if (!strcmp(token,"slime"))
               sh.flags|=SHADER_SLIME;
            else
            if (!strcmp(token,"water"))
               sh.flags|=SHADER_WATER;
            else
            if (!strcmp(token,"sky"))
               sh.flags|=SHADER_SKY;
         }
         else
         if (!strcmp(token,"q3map_surfacelight"))
         {
            TokenGet(1,-1);
            sh.light=atof(token);
         }

         else
         if (!strcmp(token,"qer_editorimage"))
         {
            TokenGet(1,-1);
            i=LookupFile(token);
            if (i!=-1 && disp_i<2)
            {
               disp_i=2;
               disp_file=i;
            }
         }

         else
         if (!strcmp(token,"map"))
         {
            TokenGet(1,-1);
            if (token[0]!='$')
            { /* TODO: discard shaders with missing textures? */
               i=LookupFile(token);
               if (i!=-1 && disp_i<=0)
               {
                  disp_i=0;
                  disp_file=i;
               }
            }
         }

         else
         {
            while (TokenAvailable(0))
               TokenGet(0,TOKEN_ALLNAME);
         }
            
      } while (brace_level);

      if (disp_file==-1)
         continue;

      sh.disp_file=disp_file;
      tex_files[disp_file].used+=100;
      
      sh.real=1;

      shader=Q_malloc(sizeof(shader_t));
      if (!shader)
      {
         S_Error("games.quake3.load_shader","Out of memory!");
         TokenDone();
         return;
      }
      *shader=sh;

      if (!strnicmp(name,"textures/",9))
         strcpy(name,&name[9]);
      tex=S_T_GetTexture(S_T_AddTexture(name,-1,-1,shader));
      if (tex)
         TC_AddName(tc_shaders,tex->name);
   }

   TokenDone();
}

static void LoadShaders(void)
{
   struct S_directory_s *d;
   S_filedir_t f;
   char name[256];
   int len;

   len=strrchr(cv_shader_path->s,'/')-cv_shader_path->s+1;
   strcpy(name,cv_shader_path->s);
   
   d=S_Dir_Open(cv_shader_path->s,FILE_NORMAL);
   if (!d)
   {
      S_NewMessage("games.quake3: Can't find .shader files!");
      return;
   }

   while (S_Dir_Read(d,&f))
   {
      strcpy(&name[len],f.name);
      LoadShaderFile(name);
   }

   S_Dir_Close(d);
}


#ifdef COUNT_TEXEL
static int total_texel;
void count_texel(int x,int y)
{
   total_texel+=x*y;
}
#endif


static tex_funcs_t quake3_tex={load_tex,load_tex_size,tex_description};


#include "plugin.h"

PLUGIN_HEAD

static int Done(void);

static plugin_t plugin=
{{Init,Done},{"games/quake3","Game plugin for Quake 3 and games/mods based on it."}};

static int Init(void)
{
   cv_texture_path=Cfg_DefVar("quake3.texture_path","/mnt/msdos_c/quake3/quest/textures/","Texture path.");
   cv_shader_path=Cfg_DefVar("quake3.shader_path","/mnt/msdos_c/quake3/quest/scripts/*.shader","Texture path.");
   cv_entity_path=Cfg_DefVar("quake3.entity_path","games/quake3/es/*","Path to the entity set.");

   cv_quake3_scale_above=Cfg_DefVar("quake3.scale_above","64",
      "All textures larger than this will always be scaled down to half their size.");
   cv_quake3_scale_below=Cfg_DefVar("quake3.scale_below","128",
      "All textures will be scaled down until they are smaller than this.");

   Cfg_SetInt(cv_texture_align_rotate,0);
   Cfg_SetInt(cv_texture_align_shiftx,0);
   Cfg_SetInt(cv_texture_align_shifty,0);
   Cfg_SetFloat(cv_texture_align_scalex,0.5);
   Cfg_SetFloat(cv_texture_align_scaley,0.5);
   
   if (!Map_Close("Change game"))
      return 0;
   Map_Unload();

   if (!G_Load(&G_Quake3))
   {
      S_Error("games/quake3","Can't load, another game is already loaded!");
      return 0;
   }

   QMap_Init();

   qmap_loadtexinfo=Q3_LoadTexInfo;
   qmap_savetexinfo=Q3_SaveTexInfo;

   S_NewMessage("quake3: Finding textures.");
   strcpy(base_tex_dir,cv_texture_path->s);
   AddDir_r(base_tex_dir);
   qsort(tex_files,num_tex_files,sizeof(tex_file_t),tex_file_cmp);

   S_T_Init(1,&quake3_tex);

   TC_Init();
   TC_Load("games/quake3/texcat.txt");
   tc_shaders=TC_New("| Shaders",TC_DONT_FREE_NAMES);
   tc_nshaders=TC_New("| Raw images",TC_DONT_FREE_NAMES);
   /* TODO: create categories based on directories */
   
   S_NewMessage("quake3: Loading shaders");
   LoadShaders();

   S_NewMessage("quake3: Creating shaders");
   {
      int i;
      shader_t *sh;
      texture_t *tex;

      for (i=0;i<num_tex_files;i++)
      {
      /* TODO: filter away uninteresting raw images here someway */
         if (S_T_LoadCheck(tex_files[i].name))
            continue;
         if (strlen(tex_files[i].name)>5 &&
             !strncmp(&tex_files[i].name[strlen(tex_files[i].name)-5],"blend",5))
            continue;
         if (strlen(tex_files[i].name)>4 &&
             !strncmp(&tex_files[i].name[strlen(tex_files[i].name)-4],"glow",4))
            continue;

         sh=Q_malloc(sizeof(shader_t));
         if (!sh) S_Abort("games.quake3","Out of memory!");
         memset(sh,0,sizeof(shader_t));
         sh->real=0;
         sh->disp_file=i;
         tex=S_T_GetTexture(S_T_AddTexture(tex_files[i].name,-1,-1,sh));

         if (tex)
         {
            if (tex_files[i].used)
               TC_AddName(tc_nshaders,tex->name);
            else
               TC_AddName(tc_shaders,tex->name);
         }
      }
   }

   return 1;
}

static int Done(void)
{
   if (!Map_Close("Unload Quake 3"))
      return 0;
   Map_Unload();
   G_Unload();
   TC_Done();
   S_T_Clear();
   Map_New();
   return 1;
}

PLUGIN_FOOT

