#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "entity.h"
#include "geom.h"
#include "map.h"
#include "quest.h"
#include "s_message.h"
#include "undo.h"


static void C_snap(void)
{
   if (!M.sel.bsel && !M.sel.esel)
   {
      S_NewMessage("edit.snap: Nothing selected!");
      return;
   }

   if (M.sel.vsel)
   {
      vsel_t *vs;
      
      for (vs=M.sel.vsel;vs;vs=vs->Next)
      {
         U_BrushModify(vs->Brush);
         SnapVecToGrid(&vs->Brush->verts[vs->num]);
      }
   }
   else
   {
      vec3_t old,delta;
      bsel_t *bs;
      brush_t *b;
      int i;

      for (bs=M.sel.bsel;bs;bs=bs->Next)
      {
         b=bs->Brush;
         if (!b->num_verts) continue;
         U_BrushModify(b);

         old=b->verts[0];
         SnapVecToGrid(&b->verts[0]);
         VEC_Sub(delta,b->verts[0],old);
         for (i=1;i<b->num_verts;i++)
         {
            VEC_Add(b->verts[i],b->verts[i],delta);
            SnapVecToGrid(&b->verts[i]);
         }
      }
   }

   {
      esel_t *es;
      vec3_t origin;
      const char *c;
      char buf[128];

      for (es=M.sel.esel;es;es=es->Next)
      {
         if (!(c=GetKeyValue(es->Entity,"origin"))) continue;
         origin.x=origin.y=origin.z=0;
         sscanf(c,"%f %f %f",&origin.x,&origin.y,&origin.z);
         SnapVecToGrid(&origin);
         sprintf(buf,"%i %i %i",(int)origin.x,(int)origin.y,(int)origin.z);
         U_EntityModify(es->Entity);
         SetKeyValue(es->Entity,"origin",buf);
      }
   }

   Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"edit.snap",C_snap,{"Edit/Snap to grid","control+n"},{},"Snap the selected brushes and entities to the grid."}
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/edit/snap","Commands to snap objects to the grid."}};

PLUGIN_FOOT_CMDS

