#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "edprim.h"

#include "brush.h"
#include "camera.h"
#include "config.h"
#include "map.h"
#include "memory.h"
#include "group.h"
#include "quest.h"
#include "s_error.h"
#include "s_texture.h"
#include "status.h"
#include "texdef.h"


#define GR ((sqrt(5) + 1) / 2)
#define GA (3 * PI / 10)
#define GR_2 (GR / 2)


/*
Used below to add edges using tables.
*/
#define ADDEDGES(num) \
   edge_t *e; \
   int *j; \
   for (i=num,e=b->edges,j=edges;i;i--,e++) \
   { \
      e->startvertex=*j++; \
      e->endvertex=*j++; \
   }


static brush_t *AllocBrush(int nverts,int nedges,int nfaces)
{
   brush_t *b;

	b = B_New(BRUSH_NORMAL);
	if (!b)
	{
		S_Error("Add Brush","Out of memory!");
		return NULL;
	}

	b->num_verts = nverts;
	b->num_edges = nedges;
	b->num_faces = nfaces;

	b->verts  = (vec3_t *)Q_malloc(sizeof(vec3_t) * b->num_verts);
	if (!b->verts)
	{
		S_Error("Add Brush","Out of memory!");
		return NULL;
	}

	b->edges = (edge_t *)Q_malloc(sizeof(edge_t) * b->num_edges);
	if (!b->edges)
	{
		S_Error("Add Brush","Out of memory!");
		return NULL;
	}

	b->faces = (face_t *)Q_malloc(sizeof(face_t) * b->num_faces);
	if (!b->faces)
	{
		S_Error("Add Brush","Out of memory!");
		return NULL;
	}
   memset(b->faces,0,sizeof(face_t)*b->num_faces);

   return b;
}


static brush_t *AddBrush_Cube(int x,int y,int z,int xsize,int ysize,int zsize)
{
	brush_t *b;
   face_t *f;
	int nsides, i;

	nsides = 4;

   b=AllocBrush(8,12,6);
   if (!b) return NULL;

   x=SnapPointToGrid(x);
   y=SnapPointToGrid(y);
   z=SnapPointToGrid(z);

	for (i=0; i<b->num_faces; i++)
	{
		b->faces[i].verts = (int *)Q_malloc(sizeof(int) * nsides);
		b->faces[i].num_verts = nsides;
		if (!b->faces[i].verts)
		{
			S_Error("Add Brush", "Out of memory!");
			return NULL;
		}
	}

   xsize/=2;
   ysize/=2;
   zsize/=2;

	b->verts[0].x = x - xsize;
	b->verts[0].y = y - ysize;
	b->verts[0].z = z + zsize;
	b->verts[1].x = x - xsize;
	b->verts[1].y = y + ysize;
	b->verts[1].z = z + zsize;
	b->verts[2].x = x + xsize;
	b->verts[2].y = y + ysize;
	b->verts[2].z = z + zsize;
	b->verts[3].x = x + xsize;
	b->verts[3].y = y - ysize;
	b->verts[3].z = z + zsize;
	b->verts[4].x = x - xsize;
	b->verts[4].y = y - ysize;
	b->verts[4].z = z - zsize;
	b->verts[5].x = x - xsize;
	b->verts[5].y = y + ysize;
	b->verts[5].z = z - zsize;
	b->verts[6].x = x + xsize;
	b->verts[6].y = y + ysize;
	b->verts[6].z = z - zsize;
	b->verts[7].x = x + xsize;
	b->verts[7].y = y - ysize;
	b->verts[7].z = z - zsize;


   {
static int edges[12*2]=
{0,1, 1,2, 2,3, 3,0, 4,5, 5,6, 6,7, 7,4, 0,4, 1,5, 2,6, 3,7};
      ADDEDGES(12)
   }

#define Cfaces(a, i, j, k, l) \
	{ \
		f->verts[0] = i; \
		f->verts[1] = j; \
      f->verts[2] = k; \
      f->verts[3] = l; \
      f++; \
	}
   f=b->faces;
   Cfaces(0, 0,1,2,3);
   Cfaces(1, 4,0,3,7);
   Cfaces(2, 7,3,2,6);
   Cfaces(3, 6,2,1,5);
   Cfaces(4, 1,0,4,5);
   Cfaces(5, 4,7,6,5);


   for (i=0; i<b->num_faces; i++)
   {
		if (!InitFace(&b->faces[i]))
         return NULL;
   }

	b->Entity = M.WorldSpawn;
	b->Group = GetDefaultGroup();

   B_Link(b);

	return b;
}


static brush_t *AddBrush_Prism(int x,int y,int z,int nsides,int radius,int height,float eccent)
{
	brush_t *b;
	int i,j;
	float tempd;

   b=AllocBrush(nsides*2,nsides*3,2+nsides);
   if (!b) return NULL;

	for (i=0; i<2; i++)
	{
		b->faces[i].verts = (int *)Q_malloc(sizeof(int) * nsides);
		b->faces[i].num_verts = nsides;
		if (!b->faces[i].verts)
		{
			S_Error("Add Brush", "Out of memory!");
			return NULL;
		}
	}

	for (i=0; i<nsides; i++)
	{
		b->faces[i+2].verts = (int *)Q_malloc(sizeof(int) * 4);
		b->faces[i+2].num_verts = 4;
		if (!b->faces[i+2].verts)
		{
			S_Error("Add Brush", "Out of memory!");
			return NULL;
		}
	}

	/* Calculate top vertices */
	for (i=0;i<nsides;i++)
	{
		tempd = -((i*2*PI)/nsides);
		tempd = cos(tempd);
		tempd *=  ((tempd<0.01)&&(tempd>-0.01)) ? 0 : radius;
		b->verts[i].x = SnapPointToGrid(tempd);

		tempd = -((i*2*PI)/nsides);
		tempd = sin(tempd);
		tempd *=  ((tempd<0.01)&&(tempd>-0.01)) ? 0 : radius*eccent;
		b->verts[i].y = SnapPointToGrid(tempd);

		b->verts[i].z = (height / 2);
	}

	/* Calculate bottom vertices */
	for (i=0;i<nsides;i++)
	{
		tempd = ((i*2*PI)/nsides);
		tempd = cos(tempd);
		tempd *=  ((tempd<0.01)&&(tempd>-0.01)) ? 0 : radius;
		b->verts[i+nsides].x = SnapPointToGrid(tempd);

		tempd = ((i*2*PI)/nsides);
		tempd = sin(tempd);
		tempd *=  ((tempd<0.01)&&(tempd>-0.01)) ? 0 : radius*eccent;
		b->verts[i+nsides].y = SnapPointToGrid(tempd);

		b->verts[i+nsides].z = -(height / 2);
	}

	/*
		 SHOULD PASS THE VERTICES THROUGH A ROTATION MATRIX
		 (orient the prism so that its top and bottom faces
		 are parallel to the view plane)
	 */

	for (i=0;i<b->num_verts;i++)
	{
		b->verts[i].x += x;
		b->verts[i].y += y;
		b->verts[i].z += z;
	}

	/* Create top edges */
	for (i=0;i<nsides-1;i++)
	{
		b->edges[i].startvertex = i;
		b->edges[i].endvertex   = i + 1;
	}
	b->edges[nsides-1].startvertex = nsides - 1;
	b->edges[nsides-1].endvertex   = 0;

	/* Create bottom edges */
	for (i=0;i<nsides-1;i++)
	{
		b->edges[nsides+i].startvertex = nsides + i;
		b->edges[nsides+i].endvertex   = nsides + i + 1;
	}
	b->edges[nsides*2-1].startvertex = nsides*2 - 1;
	b->edges[nsides*2-1].endvertex   = nsides;

	/* Create side edges */
	b->edges[nsides*2].startvertex = 0;
	b->edges[nsides*2].endvertex   = nsides;
	for (i=1;i<nsides;i++)
	{
		b->edges[nsides*2+i].startvertex = i;
		b->edges[nsides*2+i].endvertex   = nsides*2 - i;
	}

	/* Do the top face's vertices */
	for (j=0;j<nsides;j++)
	{
		b->faces[0].verts[j] =  j;
	}

	/* Do the bottom face's vertices */
	for (j=0;j<nsides;j++)
	{
		b->faces[1].verts[j] = j + nsides;
	}

	/* Do the side faces' vertices */
	b->faces[2].verts[0] =  1;
	b->faces[2].verts[1] =  0;
	b->faces[2].verts[2] =  nsides;
	b->faces[2].verts[3] =  nsides*2 - 1;
	for (i=1;i<nsides-1;i++)
	{
		b->faces[i+2].verts[0] = i + 1;
		b->faces[i+2].verts[1] = i;
		b->faces[i+2].verts[2] = nsides*2 - i;
		b->faces[i+2].verts[3] = nsides*2 - i - 1;
	}
	b->faces[nsides+1].verts[0] = 0;
	b->faces[nsides+1].verts[1] = nsides - 1;
	b->faces[nsides+1].verts[2] = nsides + 1;
	b->faces[nsides+1].verts[3] = nsides;

	/* update misc. face info */
	for (i=0; i<b->num_faces; i++)
	{
		if (!InitFace(&b->faces[i]))
         return NULL;
	}

	b->Entity = M.WorldSpawn;
	b->Group = GetDefaultGroup();

   B_Link(b);

	return b;
}


/* Pyramid primitive by Gyro Gearloose. */
static brush_t *AddBrush_Pyramid(int x,int y,int z,int nsides,int radius,int height,float eccent)
{
	brush_t *b;
	int i,j;
	float tempd;

   b=AllocBrush(nsides+1,nsides*2,nsides+1);
   if (!b) return NULL;

	/* Allocate bottom face */
	b->faces[0].verts = (int *)Q_malloc(sizeof(int) * nsides);
	b->faces[0].num_verts = nsides;
	if (!b->faces[0].verts)
	{
		S_Error("Add Brush", "Out of memory!");
		return NULL;
	}

	/* Allocate side faces */
	for (i=0; i<nsides; i++)
	{
		b->faces[i+1].verts = (int *)Q_malloc(sizeof(int) * 3);
		b->faces[i+1].num_verts = 3;
		if (!b->faces[i+1].verts)
		{
			S_Error("Add Brush", "Out of memory!");
			return NULL;
		}
	}

	/* Create top vertex */
	b->verts[0].x = 0.0;
	b->verts[0].y = 0.0;
	b->verts[0].z = (height / 2);

	/* Create bottom vertices */
	for (i=0;i<nsides;i++)
	{
		tempd = ((i*2*PI)/nsides);
		tempd = cos(tempd);
		tempd *=  ((tempd<0.01)&&(tempd>-0.01)) ? 0 : radius;
		b->verts[i+1].x = SnapPointToGrid(tempd);

		tempd = ((i*2*PI)/nsides);
		tempd = sin(tempd);
		tempd *=  ((tempd<0.01)&&(tempd>-0.01)) ? 0 : radius*eccent;
		b->verts[i+1].y = SnapPointToGrid(tempd);

		b->verts[i+1].z = -(height / 2);
	}

	// Translate vertices of the object over to where the camera is
	for (i=0;i<nsides+1;i++)
	{
		b->verts[i].x += x;
		b->verts[i].y += y;
		b->verts[i].z += z;
	}

	/* Create bottom edges (edges 0 to nsides-1) */
	for (i=0;i<nsides-1;i++)
	{
		b->edges[i].startvertex = i + 1;
		b->edges[i].endvertex   = i + 2;
	}
	b->edges[nsides-1].startvertex = nsides;
	b->edges[nsides-1].endvertex   = 1;

	/* Create side edges (edges nsides to nsides*2-1) */
	b->edges[nsides].startvertex = 0;
	b->edges[nsides].endvertex   = nsides;
	for (i=1;i<nsides;i++)
	{
		b->edges[nsides+i].startvertex = 0;
		b->edges[nsides+i].endvertex   = nsides - i;
	}

	/* Do the bottom face's vertices */
	for (j=0;j<nsides;j++)
	{
		b->faces[0].verts[j] = j + 1;
	}

	/* Do the side faces' vertices */
	for (i=1;i<nsides;i++)
	{
		b->faces[i].verts[0] = i;
		b->faces[i].verts[1] = 0;
		b->faces[i].verts[2] = i+1;
	}
	b->faces[nsides].verts[0] = nsides;
	b->faces[nsides].verts[1] = 0;
	b->faces[nsides].verts[2] = 1;

	/* update misc. face info */
	for (i=0; i<b->num_faces; i++)
		if (!InitFace(&b->faces[i]))
         return NULL;

	b->Entity = M.WorldSpawn;
	b->Group = GetDefaultGroup();

   B_Link(b);

	return b;
}


static brush_t *AddBrush_Dodec(int x,int y,int z,int radius,float eccent)
{
	brush_t *b;
	int i;
	float tempd;

	// 20 vertices, 30 edges, 12 (!) faces
   b=AllocBrush(20,30,12);
   if (!b) return NULL;

	for (i=0; i<b->num_faces; i++)
	{
		b->faces[i].verts = (int *)Q_malloc(sizeof(int) * 5);
		b->faces[i].num_verts = 5;
		if (!b->faces[i].verts)
		{
			S_Error("Add Brush", "Unable to allocate brush face vertex list");
			return NULL;
		}
	}

	for (i=0; i<5; i++)
	{
		tempd = i * 2 * PI / 5;
		if ((tempd<0.01)&&(tempd>-0.01)) tempd=0;

		/* Calculate top vertices */
		b->verts[i].x = radius * cos(tempd);
		b->verts[i].y = eccent * radius * sin(tempd);
		b->verts[i].z = 0;

		/* Calculate second level vertices */
		b->verts[i+5].x = GR * radius * cos(tempd);
		b->verts[i+5].y = eccent * GR * radius * sin(tempd);
		b->verts[i+5].z = radius;

		/* Rotate by ... radians */
		tempd += PI/5;
		if ((tempd<0.01)&&(tempd>-0.01)) tempd=0;

		/* Calculate third level vertices */
		b->verts[i+10].x = 2 * sin(GA) * radius * cos(tempd);
		b->verts[i+10].y = 2 * sin(GA) * eccent * radius * sin(tempd);
		b->verts[i+10].z = radius * GR;

		/* Calculate bottom level vertices */
		b->verts[i+15].x = radius * cos(tempd);
		b->verts[i+15].y = eccent * radius * sin(tempd);
		b->verts[i+15].z = radius * (1 + GR);
	}

	for (i=0;i<b->num_verts;i++)
	{
		b->verts[i].x += x;
		b->verts[i].y += y;
		b->verts[i].z += z;
		b->verts[i].x = rint(b->verts[i].x * 10000) / 10000;
		b->verts[i].y = rint(b->verts[i].y * 10000) / 10000;
		b->verts[i].z = rint(b->verts[i].z * 10000) / 10000;
	}

   {
static int edges[30*2]=
{
 0, 1,  1, 2,  2, 3,  3, 4,  4, 0,
 0, 5,  1, 6,  2, 7,  3, 8,  4, 9,
 5,10, 10, 6,  6,11, 11, 7,  7,12, 12, 8,  8,13, 13, 9,  9,14, 14, 5,
10,15, 11,16, 12,17, 13,18, 14,19,
15,16, 16,17, 17,18, 18,19, 19,15,
};
      ADDEDGES(30)
   }

	/*
		 The vertices _must_ form each face in the order given,
		 otherwise the normal will have the wrong sign, and the
		 object will fail the convex test.
	*/

#define Dface(a, x, y) { b->faces[a].verts[x] = y; }

	Dface( 0, 0,  0);
	Dface( 0, 1,  1);
	Dface( 0, 2,  2);
	Dface( 0, 3,  3);
	Dface( 0, 4,  4);

	Dface( 1, 0,  0);
	Dface( 1, 1,  5);
	Dface( 1, 2, 10);
	Dface( 1, 3,  6);
	Dface( 1, 4,  1);

	Dface( 2, 0,  1);
	Dface( 2, 1,  6);
	Dface( 2, 2, 11);
	Dface( 2, 3,  7);
	Dface( 2, 4,  2);

	Dface( 3, 0,  2);
	Dface( 3, 1,  7);
	Dface( 3, 2, 12);
	Dface( 3, 3,  8);
	Dface( 3, 4,  3);

	Dface( 4, 0,  3);
	Dface( 4, 1,  8);
	Dface( 4, 2, 13);
	Dface( 4, 3,  9);
	Dface( 4, 4,  4);

	Dface( 5, 0,  4);
	Dface( 5, 1,  9);
	Dface( 5, 2, 14);
	Dface( 5, 3,  5);
	Dface( 5, 4,  0);

	Dface( 6, 0, 16);
	Dface( 6, 1, 11);
	Dface( 6, 2,  6);
	Dface( 6, 3, 10);
	Dface( 6, 4, 15);

	Dface( 7, 0, 17);
	Dface( 7, 1, 12);
	Dface( 7, 2,  7);
	Dface( 7, 3, 11);
	Dface( 7, 4, 16);

	Dface( 8, 0, 18);
	Dface( 8, 1, 13);
	Dface( 8, 2,  8);
	Dface( 8, 3, 12);
	Dface( 8, 4, 17);

	Dface( 9, 0, 19);
	Dface( 9, 1, 14);
	Dface( 9, 2,  9);
	Dface( 9, 3, 13);
	Dface( 9, 4, 18);

	Dface(10, 0, 15);
	Dface(10, 1, 10);
	Dface(10, 2,  5);
	Dface(10, 3, 14);
	Dface(10, 4, 19);

	Dface(11, 0, 19);
	Dface(11, 1, 18);
	Dface(11, 2, 17);
	Dface(11, 3, 16);
	Dface(11, 4, 15);

	/* update misc. face info */
	for (i=0; i<b->num_faces; i++)
		if (!InitFace(&b->faces[i]))
         return NULL;

	b->Entity = M.WorldSpawn;
	b->Group = GetDefaultGroup();

   B_Link(b);

	return b;
}


static brush_t *AddBrush_Icos(int x,int y,int z,int radius,float eccent)
{
	brush_t *b;
	int i;
   face_t *f;

	// 12 vertices, 30 edges, 20 faces
   b=AllocBrush(12,30,20);
   if (!b) return NULL;

	for (i=0; i<b->num_faces; i++)
	{
		b->faces[i].verts = (int *)Q_malloc(sizeof(int) * 3);
		b->faces[i].num_verts = 3;
		if (!b->faces[i].verts)
		{
			S_Error("Add Brush", "Unable to allocate brush face vertex list");
			return NULL;
		}
	}

#define Ivert(a, j, k, l) \
	{ \
		b->verts[a].x = j; b->verts[a].y = k; b->verts[a].z = l; \
	}	
	Ivert(0, 0, 0.5, GR_2);
	Ivert(1, 0,-0.5, GR_2);
	Ivert(2, 0,-0.5,-GR_2);
	Ivert(3, 0, 0.5,-GR_2);

	Ivert(4, GR_2, 0, 0.5);
	Ivert(5,-GR_2, 0, 0.5);
	Ivert(6,-GR_2, 0,-0.5);
	Ivert(7, GR_2, 0,-0.5);

	Ivert(8,  0.5, GR_2, 0);
	Ivert(9,  0.5,-GR_2, 0);
	Ivert(10,-0.5,-GR_2, 0);
	Ivert(11,-0.5, GR_2, 0);

	for (i=0;i<b->num_verts;i++)
	{
		b->verts[i].x *= radius;
		b->verts[i].y *= eccent * radius;
		b->verts[i].z *= radius;
		b->verts[i].x += x;
		b->verts[i].y += y;
		b->verts[i].z += z;
		b->verts[i].x = SnapPointToGrid(b->verts[i].x);
		b->verts[i].y = SnapPointToGrid(b->verts[i].y);
		b->verts[i].z = SnapPointToGrid(b->verts[i].z);
	}

   {
static int edges[30*2]=
{
 0, 1,  0,11,  0, 4,  0, 5,  0, 8,  1,10,  1, 4,  1, 5,  1, 9,  2,10,
 2, 3,  2, 6,  2, 7,  2, 9,  3,11,  3, 6,  3, 7,  3, 8,  4, 7,  4, 8,
 4, 9,  5,10,  5,11,  5, 6,  6,10,  6,11,  7, 8,  7, 9,  8,11,  9,10,
};
      ADDEDGES(30)
   }

	/*
		 The vertices _must_ form each face in the order given,
		 otherwise the normal will have the wrong sign, and the
		 object will fail the convex test.
	 */

#define Iface(a, i, j, k) \
	{ \
		f->verts[0] = i; \
		f->verts[1] = j; \
      f->verts[2] = k; \
      f++; \
	}
   f=b->faces;
	Iface( 0,  0,  1,  5);
	Iface( 1,  0,  5, 11);
	Iface( 2,  0, 11,  8);
	Iface( 3,  8,  4,  0);
	Iface( 4,  0,  4,  1);
	Iface( 5, 10,  5,  1);
	Iface( 6,  1,  4,  9);
	Iface( 7,  1,  9, 10);
	Iface( 8,  7,  4,  8);
	Iface( 9,  9,  4,  7);
	Iface(10,  6,  2,  3);
	Iface(11, 11,  6,  3);
	Iface(12,  8, 11,  3);
	Iface(13,  3,  7,  8);
	Iface(14,  2,  7,  3);
	Iface(15,  2,  6, 10);
	Iface(16,  9,  7,  2);
	Iface(17, 10,  9,  2);
	Iface(18, 11,  5,  6);
	Iface(19,  6,  5, 10);


	/* update misc. face info */
	for (i=0; i<b->num_faces; i++)
		if (!InitFace(&b->faces[i]))
         return NULL;

	b->Entity = M.WorldSpawn;
	b->Group = GetDefaultGroup();

   B_Link(b);

	return b;
}


static brush_t *AddBrush_Bucky(int x,int y,int z,int radius,float eccent)
{
	brush_t *b;
   face_t *f;
	int hexagons, pentagons;
	int i;

   /*
      60 vertices, 90 edges, 32 faces composed of 20 hexagons, 12 pentagons
   */

	hexagons = 20;
	pentagons = 12;

   b=AllocBrush(60,90,32);
   if (!b) return NULL;

	for (i=0; i<hexagons; i++)
	{
		b->faces[i].verts = (int *)Q_malloc(sizeof(int) * 6);
		b->faces[i].num_verts = 6;
		if (!b->faces[i].verts)
		{
			S_Error("Add Brush", "Unable to allocate brush face vertex list");
			return NULL;
		}
	}	

	for (i=hexagons; i<hexagons+pentagons; i++)
	{
		b->faces[i].verts = (int *)Q_malloc(sizeof(int) * 5);
		b->faces[i].num_verts = 5;
		if (!b->faces[i].verts)
		{
			S_Error("Add Brush", "Unable to allocate brush face vertex list");
			return NULL;
		}
	}

#define Bvert(a, j, k, l) \
	{ \
		b->verts[a].x = j; b->verts[a].y = k; b->verts[a].z = l; \
	}	

	Bvert( 0,  0.343279,  0.000000,  0.939234);
	Bvert( 1,  0.106079,  0.326477,  0.939234);
	Bvert( 2, -0.277718,  0.201774,  0.939234);
	Bvert( 3, -0.277718, -0.201774,  0.939234);
	Bvert( 4,  0.106079, -0.326477,  0.939234);
	Bvert( 5,  0.686557,  0.000000,  0.727076);
	Bvert( 6,  0.792636,  0.326477,  0.514918);
	Bvert( 7,  0.555436,  0.652955,  0.514918);
	Bvert( 8,  0.212158,  0.652955,  0.727076);
	Bvert( 9, -0.065560,  0.854729,  0.514918);
	Bvert(10, -0.449358,  0.730026,  0.514918);
	Bvert(11, -0.555436,  0.403548,  0.727076);
	Bvert(12, -0.833155,  0.201774,  0.514918);
	Bvert(13, -0.833155, -0.201774,  0.514918);
	Bvert(14, -0.555436, -0.403548,  0.727076);
	Bvert(15, -0.449358, -0.730026,  0.514918);
	Bvert(16, -0.065560, -0.854729,  0.514918);
	Bvert(17,  0.212158, -0.652955,  0.727076);
	Bvert(18,  0.555436, -0.652955,  0.514918);
	Bvert(19,  0.792636, -0.326477,  0.514918);
	Bvert(20,  0.964275, -0.201774,  0.171639);
	Bvert(21,  0.964275,  0.201774,  0.171639);
	Bvert(22,  0.898715,  0.403548, -0.171639);
	Bvert(23,  0.661515,  0.730026, -0.171639);
	Bvert(24,  0.489876,  0.854729,  0.171639);
	Bvert(25,  0.106079,  0.979432,  0.171639);
	Bvert(26, -0.106079,  0.979432, -0.171639);
	Bvert(27, -0.489876,  0.854729, -0.171639);
	Bvert(28, -0.661515,  0.730026,  0.171639);
	Bvert(29, -0.898715,  0.403548,  0.171639);
	Bvert(30, -0.964275,  0.201774, -0.171639);
	Bvert(31, -0.964275, -0.201774, -0.171639);
	Bvert(32, -0.898715, -0.403548,  0.171639);
	Bvert(33, -0.661515, -0.730026,  0.171639);
	Bvert(34, -0.489876, -0.854729, -0.171639);
	Bvert(35, -0.106079, -0.979432, -0.171639);
	Bvert(36,  0.106079, -0.979432,  0.171639);
	Bvert(37,  0.489876, -0.854729,  0.171639);
	Bvert(38,  0.661515, -0.730026, -0.171639);
	Bvert(39,  0.898715, -0.403548, -0.171639);
	Bvert(40,  0.833155, -0.201774, -0.514918);
	Bvert(41,  0.833155,  0.201774, -0.514918);
	Bvert(42,  0.555436,  0.403548, -0.727076);
	Bvert(43,  0.449358,  0.730026, -0.514918);
	Bvert(44,  0.065560,  0.854729, -0.514918);
	Bvert(45, -0.212158,  0.652955, -0.727076);
	Bvert(46, -0.555436,  0.652955, -0.514918);
	Bvert(47, -0.792636,  0.326477, -0.514918);
	Bvert(48, -0.686557,  0.000000, -0.727076);
	Bvert(49, -0.792636, -0.326477, -0.514918);
	Bvert(50, -0.555436, -0.652955, -0.514918);
	Bvert(51, -0.212158, -0.652955, -0.727076);
	Bvert(52,  0.065560, -0.854729, -0.514918);
	Bvert(53,  0.449358, -0.730026, -0.514918);
	Bvert(54,  0.555436, -0.403548, -0.727076);
	Bvert(55,  0.277718, -0.201774, -0.939234);
	Bvert(56,  0.277718,  0.201774, -0.939234);
	Bvert(57, -0.106079,  0.326477, -0.939234);
	Bvert(58, -0.343279,  0.000000, -0.939234);
	Bvert(59, -0.106079, -0.326477, -0.939234);

	for (i=0;i<b->num_verts;i++) {
		b->verts[i].x *= radius;
		b->verts[i].y *= eccent * radius;
		b->verts[i].z *= radius;
		b->verts[i].x += x;
		b->verts[i].y += y;
		b->verts[i].z += z;
		b->verts[i].x = rint(b->verts[i].x * 10000) / 10000;
		b->verts[i].y = rint(b->verts[i].y * 10000) / 10000;
		b->verts[i].z = rint(b->verts[i].z * 10000) / 10000;
	}

   {
static int edges[90*2]=
{
#define Bedge(i,x,y) x,y,
Bedge( 0,  0,  1)	Bedge( 1,  0,  4)	Bedge( 2,  0,  5)	Bedge( 3,  1,  2)
Bedge( 4,  1,  8)	Bedge( 5,  2,  3)	Bedge( 6,  2, 11)	Bedge( 7,  3,  4)
Bedge( 8,  3, 14)	Bedge( 9,  4, 17)	Bedge(10,  5,  6)	Bedge(11,  5, 19)
Bedge(12,  6,  7)	Bedge(13,  6, 21)	Bedge(14,  7,  8)	Bedge(15,  7, 24)
Bedge(16,  8,  9)	Bedge(17,  9, 10)	Bedge(18,  9, 25)	Bedge(19, 10, 11)
Bedge(20, 10, 28)	Bedge(21, 11, 12)	Bedge(22, 12, 13)	Bedge(23, 12, 29)
Bedge(24, 13, 14)	Bedge(25, 13, 32)	Bedge(26, 14, 15)	Bedge(27, 15, 16)
Bedge(28, 15, 33)	Bedge(29, 16, 17)	Bedge(30, 16, 36)	Bedge(31, 17, 18)
Bedge(32, 18, 19)	Bedge(33, 18, 37)	Bedge(34, 19, 20)	Bedge(35, 20, 21)
Bedge(36, 20, 39)	Bedge(37, 21, 22)	Bedge(38, 22, 23)	Bedge(39, 22, 41)
Bedge(40, 23, 24)	Bedge(41, 23, 43)	Bedge(42, 24, 25)	Bedge(43, 25, 26)
Bedge(44, 26, 27)	Bedge(45, 26, 44)	Bedge(46, 27, 28)	Bedge(47, 27, 46)
Bedge(48, 28, 29)	Bedge(49, 29, 30)	Bedge(50, 30, 31)	Bedge(51, 30, 47)
Bedge(52, 31, 32)	Bedge(53, 31, 49)	Bedge(54, 32, 33)	Bedge(55, 33, 34)
Bedge(56, 34, 35)	Bedge(57, 34, 50)	Bedge(58, 35, 36)	Bedge(59, 35, 52)
Bedge(60, 36, 37)	Bedge(61, 37, 38)	Bedge(62, 38, 39)	Bedge(63, 38, 53)
Bedge(64, 39, 40)	Bedge(65, 40, 41)	Bedge(66, 40, 54)	Bedge(67, 41, 42)
Bedge(68, 42, 43)	Bedge(69, 42, 56)	Bedge(70, 43, 44)	Bedge(71, 44, 45)
Bedge(72, 45, 46)	Bedge(73, 45, 57)	Bedge(74, 46, 47)	Bedge(75, 47, 48)
Bedge(76, 48, 49)	Bedge(77, 48, 58)	Bedge(78, 49, 50)	Bedge(79, 50, 51)
Bedge(80, 51, 52)	Bedge(81, 51, 59)	Bedge(82, 52, 53)	Bedge(83, 53, 54)
Bedge(84, 54, 55)	Bedge(85, 55, 56)	Bedge(86, 55, 59)	Bedge(87, 56, 57)
Bedge(88, 57, 58)	Bedge(89, 58, 59)
}; /* end edges[] */

      ADDEDGES(90)
   }


	/*
		 The vertices _must_ form each face in the order given,
		 otherwise the normal will have the wrong sign, and the
		 object will fail the convex test.
	*/

#define Bface(a, i, j, k, l, m, n) \
	{ \
		f->verts[0] = i; \
      f->verts[1] = j; \
      f->verts[2] = k; \
      f->verts[3] = l; \
      f->verts[4] = m; \
      f->verts[5] = n; \
      f++; \
	}

   f=b->faces;
	Bface( 0,  1,  8,  7,  6,  5,  0);
	Bface( 1,  2, 11, 10,  9,  8,  1);
	Bface( 2,  3, 14, 13, 12, 11,  2);
	Bface( 3,  4, 17, 16, 15, 14,  3);
	Bface( 4,  0,  5, 19, 18, 17,  4);
	Bface( 5,  7, 24, 23, 22, 21,  6);
	Bface( 6, 10, 28, 27, 26, 25,  9);
	Bface( 7, 13, 32, 31, 30, 29, 12);
	Bface( 8, 16, 36, 35, 34, 33, 15);
	Bface( 9, 19, 20, 39, 38, 37, 18);
	Bface(10, 40, 39, 20, 21, 22, 41);
	Bface(11, 43, 23, 24, 25, 26, 44);
	Bface(12, 46, 27, 28, 29, 30, 47);
	Bface(13, 49, 31, 32, 33, 34, 50);
	Bface(14, 52, 35, 36, 37, 38, 53);
	Bface(15, 55, 54, 40, 41, 42, 56);
	Bface(16, 56, 42, 43, 44, 45, 57);
	Bface(17, 57, 45, 46, 47, 48, 58);
	Bface(18, 58, 48, 49, 50, 51, 59);
	Bface(19, 59, 51, 52, 53, 54, 55);

#undef Bface
#define Bface(a, i, j, k, l, m) \
	{ \
		f->verts[0] = i; \
      f->verts[1] = j; \
      f->verts[2] = k; \
      f->verts[3] = l; \
      f->verts[4] = m; \
      f++; \
	}

	Bface(20,   4,   3,   2,    1,   0);
	Bface(21,   6,  21,  20,   19,   5);
	Bface(22,   9,  25,  24,    7,   8);
	Bface(23,  12,  29,  28,   10,  11);
	Bface(24,  15,  33,  32,   13,  14);
	Bface(25,  18,  37,  36,   16,  17);
	Bface(26,  42,  41,  22,   23,  43);
	Bface(27,  45,  44,  26,   27,  46);
	Bface(28,  48,  47,  30,   31,  49);
	Bface(29,  51,  50,  34,   35,  52);
	Bface(30,  54,  53,  38,   39,  40);
	Bface(31,  55,  56,  57,   58,  59);


	/* update misc. face info */
	for (i=0; i<b->num_faces; i++)
		if (!InitFace(&b->faces[i]))
         return NULL;

	b->Entity = M.WorldSpawn;
	b->Group = GetDefaultGroup();

   B_Link(b);

	return b;
}

static brush_t *AddBrush_Torus(int x,int y,int z,int nsides,int nradius,int osides,int oradius)
{
	brush_t *b;
	int i,j,k;
	float radius, height;
	float tempd1, tempd2, tempd3;

	for (k=0; k<osides; k++)
   {
      b=AllocBrush(nsides*2,nsides*3,2+nsides);
      if (!b) return NULL;

		/* Allocate top and bottom faces */
		for (i=0; i<2; i++)
		{
			b->faces[i].verts = (int *)Q_malloc(sizeof(int) * nsides);
			b->faces[i].num_verts = nsides;
			if (!b->faces[i].verts)
			{
				S_Error("Add Brush", "Out of memory!");
				return NULL;
			}
		}

		/* Allocate side faces */
		for (i=0; i<nsides; i++)
		{
			b->faces[i+2].verts = (int *)Q_malloc(sizeof(int) * 4);
			b->faces[i+2].num_verts = 4;
			if (!b->faces[i+2].verts)
			{
				S_Error("Add Brush", "Out of memory!");
				return NULL;
			}
		}

		/*
			 Create the vertices.
			 height is half the width of an torus edge 
			 tempd3 is the rotation angle to make the floor horizontal 

			 Method 
			 1. Make a prism about the y-axis.
			 2. Translate it so that (x) is +ve
			 3. Shape the ends w.r.t the number of edges.
			 4. Translate it back by the torus radius
			 5. Rotate each segment into position.

			 TODO
			 Make the math more efficient.
		 */

		height = oradius * sin(PI / osides);
		tempd1 = 2 * PI / nsides;
		tempd3 = ((nsides & 1) == 1) ? PI / 2 : PI / 4;

		for (i=0;i<nsides;i++)
		{
			tempd2 = cos(tempd3 - (i * tempd1));
			tempd2 *= ((tempd2<0.01)&&(tempd2>-0.01)) ? 0 : nradius;
			b->verts[i].x = tempd2 + nradius;

			tempd2 = sin(tempd3 - (i * tempd1));
			tempd2 *= ((tempd2<0.01)&&(tempd2>-0.01)) ? 0 : nradius;
			b->verts[i].z = b->verts[i+nsides].z = tempd2;

			b->verts[i].y = height - (b->verts[i].x * tan(PI / osides));
			b->verts[i].x += oradius * cos(PI / osides);

			radius = hypot(b->verts[i].x, b->verts[i].y);

			b->verts[i].y = radius * sin(2 * k * PI / osides);
			b->verts[i].x = radius * cos(2 * k * PI / osides);
			b->verts[i+nsides].y = radius * sin(2 * (k+1) * PI / osides);
			b->verts[i+nsides].x = radius * cos(2 * (k+1) * PI / osides);
		}

		/* Create top edges */
		for (i=0;i<nsides-1;i++)
		{
			b->edges[i].startvertex = i;
			b->edges[i].endvertex   = i + 1;
		}
		b->edges[nsides-1].startvertex = nsides - 1;
		b->edges[nsides-1].endvertex   = 0;

		/* Create bottom edges */
		for (i=0;i<nsides-1;i++)
		{
			b->edges[nsides+i].startvertex = nsides + i;
			b->edges[nsides+i].endvertex   = nsides + i + 1;
		}
		b->edges[nsides*2-1].startvertex = nsides*2 - 1;
		b->edges[nsides*2-1].endvertex   = nsides;

		/* Create side edges */
		b->edges[nsides*2].startvertex = 0;
		b->edges[nsides*2].endvertex   = nsides;
		for (i=1;i<nsides;i++)
		{
			b->edges[nsides*2+i].startvertex = i;
			b->edges[nsides*2+i].endvertex   = nsides + i;
		}

		/* Do the top face's vertices */
		for (j=0;j<nsides;j++)
		{
			b->faces[0].verts[j] = j;
		}

		/* Do the bottom face's vertices */
		for (j=0;j<nsides;j++)
		{
			b->faces[1].verts[j] =  nsides + nsides -1 -j;
		}

		/* Do the side faces' vertices */
		for (j=0; j<nsides; j++)
		{
			b->faces[j+2].verts[0] = j+nsides;
			b->faces[j+2].verts[1] = j+nsides+1;
			b->faces[j+2].verts[2] = j+1;
			b->faces[j+2].verts[3] = j;
		}
		b->faces[nsides+1].verts[1] = nsides;
		b->faces[nsides+1].verts[2] = 0;

		for (i=0; i<nsides+2; i++)
		if (!InitFace(&b->faces[i]))
         return NULL;
         
		for (i=0;i<nsides*2;i++)
		{
			b->verts[i].x += x;
			b->verts[i].y += y;
			b->verts[i].z += z;
			b->verts[i].x = rint(b->verts[i].x * 10000) / 10000;
			b->verts[i].y = rint(b->verts[i].y * 10000) / 10000;
			b->verts[i].z = rint(b->verts[i].z * 10000) / 10000;
		}
		b->Entity = M.WorldSpawn;
		b->Group = GetDefaultGroup();

      B_Link(b);
	}
	return NULL;
}


brush_t *AddBrush(int type,int x,int y,int z,float info1,float info2,float info3,float info4)
{
   texture_t *tex;

	switch (type)
	{
		case PRIM_TEXRECT:
			/* Make sure there is a texture to use, first */
			tex = S_T_LoadSize(cv_texture_current->s);
			if (tex)
			{
				info1 = tex->width;
				info2 = tex->width;
				info3 = tex->height;
			}
			else
			{
				info1=info2=info3=64;
			}
      case PRIM_CUBE:
			return AddBrush_Cube(x,y,z,info1,info2,info3);
         
		case PRIM_NPRISM:
			return AddBrush_Prism(x,y,z,info1,info2,info3,info4);

		case PRIM_PYRAMID:
			return AddBrush_Pyramid(x,y,z,info1,info2,info3,info4);

		case PRIM_DODEC:
			return AddBrush_Dodec(x,y,z,info1,info2);

		case PRIM_ICOS:
			return AddBrush_Icos(x,y,z,info1,info2);

		case PRIM_BUCKY:
			return AddBrush_Bucky(x,y,z,info1,info2);

		case PRIM_TORUS:	
			return AddBrush_Torus(x,y,z,info1,info2,info3,info4);
	}
	S_Error("Add Brush","Unknown brush type %i!",type);
	return NULL;
}

