#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gtk/gtk.h>

#include "g_menu.h"

#include "g_gui.h"
#include "g_input.h"

#include "cmds.h"


/* TODO: think of some better way to do this */
#define ARG_UNUSED(x) x __attribute__ ((unused))


#define MENU_DATA_FIELD "quest-menu"


GtkWindow *S_menu_window;
static GtkMenuBar *menu_bar;


static void menu_select(ARG_UNUSED(GtkWidget *w),char *str)
{
   Cmd_Run(str);
}


static gint menu_win_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),ARG_UNUSED(gpointer data))
{
   return TRUE;
}


struct S_Menu *S_RootMenu;


static GtkMenuItem *FindItem(GtkMenuShell *ms,const char *name)
{
   GList *ch;

   for (ch=ms->children;ch;ch=ch->next)
   {
      if (GTK_IS_LABEL(GTK_BIN(ch->data)->child))
         if (!strcmp(GTK_LABEL(GTK_BIN(ch->data)->child)->label,name))
            return GTK_MENU_ITEM(ch->data);
   }
   return NULL;
}

static void ClearItem(GtkMenuItem *m)
{
   if (m->submenu)
      gtk_menu_item_remove_submenu(m);
   else
   {
      gpointer d;

      d=gtk_object_get_data(GTK_OBJECT(m),MENU_DATA_FIELD);
      gtk_signal_disconnect_by_func(GTK_OBJECT(m),menu_select,d);
   }
}

struct S_Menu *S_Menu_Submenu(struct S_Menu *rm,const char *name)
{
   GtkMenuItem *mi;
   GtkMenu *m;

   mi=FindItem(GTK_MENU_SHELL(rm),name);
   if (mi && mi->submenu)
      return (struct S_Menu *)mi->submenu;

   if (mi)
   {
      ClearItem(mi);
   }
   else
   {
      mi=GTK_MENU_ITEM(gtk_menu_item_new_with_label(name));
      if (GTK_IS_MENU_BAR(rm))
         gtk_menu_bar_append(GTK_MENU_BAR(rm),GTK_WIDGET(mi));
      else
         gtk_menu_append(GTK_MENU(rm),GTK_WIDGET(mi));
      gtk_widget_show(GTK_WIDGET(mi));
   }
   
   m=GTK_MENU(gtk_menu_new());
   gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi),GTK_WIDGET(m));
   return (struct S_Menu *)m;
}

void S_Menu_AddItem(struct S_Menu *rm,const char *name,const char *cmd,int sep)
{
   GtkMenuItem *mi;

   mi=FindItem(GTK_MENU_SHELL(rm),name);
   if (mi)
   {
      ClearItem(mi);
   }
   else
   {
      if (sep&1)
      {
         GtkWidget *a;
      
         mi=GTK_MENU_ITEM(gtk_menu_item_new());
   
         a=gtk_hseparator_new();
         gtk_container_add(GTK_CONTAINER(mi),a);
         gtk_widget_show(a);
         
         gtk_widget_set(GTK_WIDGET(mi),
            "GtkWidget::sensitive",0,
            NULL);
         
         gtk_widget_show(GTK_WIDGET(mi));
         if (GTK_IS_MENU_BAR(rm))
            gtk_menu_bar_append(GTK_MENU_BAR(rm),GTK_WIDGET(mi));
         else
            gtk_menu_append(GTK_MENU(rm),GTK_WIDGET(mi));
      }
      
      mi=GTK_MENU_ITEM(gtk_menu_item_new_with_label(name));
      if (GTK_IS_MENU_BAR(rm))
         gtk_menu_bar_append(GTK_MENU_BAR(rm),GTK_WIDGET(mi));
      else
         gtk_menu_append(GTK_MENU(rm),GTK_WIDGET(mi));
      gtk_widget_show(GTK_WIDGET(mi));
      
      if (sep&2)
      {
         GtkWidget *a;
         GtkMenuItem *mi2;
      
         mi2=GTK_MENU_ITEM(gtk_menu_item_new());
   
         a=gtk_hseparator_new();
         gtk_container_add(GTK_CONTAINER(mi2),a);
         gtk_widget_show(a);
         
         gtk_widget_set(GTK_WIDGET(mi2),
            "GtkWidget::sensitive",0,
            NULL);
         
         gtk_widget_show(GTK_WIDGET(mi2));
         if (GTK_IS_MENU_BAR(rm))
            gtk_menu_bar_append(GTK_MENU_BAR(rm),GTK_WIDGET(mi2));
         else
            gtk_menu_append(GTK_MENU(rm),GTK_WIDGET(mi2));
      }
   }
   
   gtk_signal_connect(GTK_OBJECT(mi),"activate",GTK_SIGNAL_FUNC(menu_select),cmd);
   gtk_object_set_data(GTK_OBJECT(mi),MENU_DATA_FIELD,cmd);
}


void S_InitMenu(void)
{
   GtkWidget *mb;

   S_menu_window=GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
   gtk_window_set_title(S_menu_window,"Quest");
   gtk_signal_connect(GTK_OBJECT(S_menu_window),"delete_event",GTK_SIGNAL_FUNC(menu_win_delete),NULL);

   gtk_widget_set_events(GTK_WIDGET(S_menu_window),
      GDK_KEY_PRESS_MASK|GDK_KEY_RELEASE_MASK);
   gtk_signal_connect(GTK_OBJECT(S_menu_window),"key_press_event",GTK_SIGNAL_FUNC(S_in_key),NULL);
   gtk_signal_connect(GTK_OBJECT(S_menu_window),"key_release_event",GTK_SIGNAL_FUNC(S_in_key),NULL);


   mb=gtk_menu_bar_new();
   gtk_container_add(GTK_CONTAINER(S_menu_window),mb);
   gtk_widget_show(mb);

   menu_bar=GTK_MENU_BAR(mb);

   S_RootMenu=(struct S_Menu *)mb;


   { /* define a bunch of standard items (without commands) so they end in where we want them */
      struct S_Menu *m;

      m=S_Menu_Submenu(S_RootMenu,"File");
      S_Menu_AddItem(m,"New",NULL,0);
      S_Menu_AddItem(m,"Load...",NULL,0);
      S_Menu_AddItem(m,"Save",NULL,0);
      S_Menu_AddItem(m,"Save as...",NULL,2);
      S_Menu_AddItem(m,"Exit",NULL,2);

      m=S_Menu_Submenu(S_RootMenu,"Edit");
      S_Menu_AddItem(m,"Undo",NULL,2);
      S_Menu_AddItem(m,"Copy",NULL,0);
      S_Menu_AddItem(m,"Paste",NULL,2);
      S_Menu_AddItem(m,"Delete",NULL,2);

      m=S_Menu_Submenu(S_Menu_Submenu(S_RootMenu,"Brush"),"Create");
      m=S_Menu_Submenu(S_Menu_Submenu(S_RootMenu,"Brush"),"CSG");

      m=S_Menu_Submenu(S_RootMenu,"Curve");
      S_Menu_AddItem(m,"Create...",NULL,2);
      
      m=S_Menu_Submenu(S_RootMenu,"Entity");
      
      m=S_Menu_Submenu(S_RootMenu,"View");
      
      m=S_Menu_Submenu(S_RootMenu,"Texture");
      
      m=S_Menu_Submenu(S_RootMenu,"Camera");
      
      m=S_Menu_Submenu(S_RootMenu,"Misc");
      
      m=S_Menu_Submenu(S_RootMenu,"Help");
   }
   
   gtk_widget_show(GTK_WIDGET(S_menu_window));
   S_LoadWinPosition(S_menu_window,"menu");
}

void S_DoneMenu(void)
{
   S_SaveWinPosition(S_menu_window,"menu");
}

