#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>

/* these are redefined in defines.h */
#undef TRUE
#undef FALSE

#include "defines.h"
#include "types.h"

#include "g_3d.h"
#include "g_3d_all.h"
#include "g_input.h"
#include "g_menu.h"

#include "s_int.h"

#include "quest.h"
#include "s_error.h"


static g_viewport_t **vports;
static int num_vports;


viewport_t *S_cur_vport;
g_viewport_t *S_cur_gvport;
static int cur_vport;


void S_SetCurViewport(int v)
{
   if (cur_vport==v) return;
   cur_vport=v;
   S_cur_gvport=vports[cur_vport];
   S_cur_vport=&S_cur_gvport->vp;
}


static void DrawToScreen(g_viewport_t *v)
{ /* TODO: generate an expose_event instead (gtk_widget_draw)? */
   if (v->pic)
   {
      gdk_draw_pixmap(
         GTK_WIDGET(v->da)->window,
         GTK_WIDGET(v->da)->style->fg_gc[GTK_WIDGET_STATE(GTK_WIDGET(v->da))],
         v->pic,
         0,0,
         0,0,
         v->sx,v->sy);
   }
}


void S_UpdateCurViewport(void)
{
   S_3D_Render(S_cur_gvport);
   DrawToScreen(S_cur_gvport);
}

void S_UpdateAllViewports(void)
{
   int i;
   g_viewport_t **v;

   for (v=vports,i=num_vports;i;i--,v++)
   {
      S_3D_Render(*v);
      DrawToScreen(*v);
   }
}


void S_R_ForEachViewport(void (*func)(struct viewport_s *))
{
   int i;

   for (i=0;i<num_vports;i++)
      func(&vports[i]->vp);
}

/* TODO: figure out exactly what the return values from these do */
static gint vp_configure(GtkWidget *w,ARG_UNUSED(GdkEventConfigure *e),g_viewport_t *vp)
{
   if (vp->pic)
      gdk_pixmap_unref(vp->pic);

   vp->sx=w->allocation.width;
   vp->sy=w->allocation.height;

   vp->pic=gdk_pixmap_new(w->window,vp->sx,vp->sy,-1);

   S_3D_Render(vp);
      
   return TRUE;
}

static gint vp_expose(GtkWidget *w,GdkEventExpose *e,g_viewport_t *vp)
{
   if (vp->pic)
   {
      gdk_draw_pixmap(w->window,w->style->fg_gc[GTK_WIDGET_STATE(w)],
         vp->pic,
         e->area.x,e->area.y,
         e->area.x,e->area.y,
         e->area.width,e->area.height);
   }
   return FALSE;
}

static gint vp_focus(ARG_UNUSED(GtkWidget *w),GdkEventFocus *e,g_viewport_t *v)
{
   if (e->in && v->num!=cur_vport)
      S_SetCurViewport(v->num);
   return FALSE;
}


/* First created viewport needs to be handled specially as the
renderer isn't setup yet, and it can't be until it can create
the necessary gc:s and colormaps.
*/
static void NewViewport(int x,int y,int first)
{
   g_viewport_t *vp;

   {
      g_viewport_t **v;
      v=realloc(vports,sizeof(g_viewport_t *)*(num_vports+1));
      if (!v)
      {
         if (first)
            S_Abort("FirstViewport","Unable to allocate memory for new viewport!");
         else
            S_Error("FirstViewport","Unable to allocate memory for new viewport!");
      }
      vports=v;
   }
   vp=malloc(sizeof(g_viewport_t));
   if (!vp)
   {
      if (first)
         S_Abort("FirstViewport","Unable to allocate memory for new viewport!");
      else
         S_Error("FirstViewport","Unable to allocate memory for new viewport!");
   }
   vports[num_vports++]=vp;

   memset(vp,0,sizeof(g_viewport_t));

   vp->num=num_vports-1;

   vp->mode=M_2D_WIRE;

   vp->sx=x;
   vp->sy=y;

   vp->da=GTK_DRAWING_AREA(gtk_drawing_area_new());

   if (first)
   {
   /* TODO: figure out best way to do this */
/*      S_colormap=gdk_colormap_new(gtk_widget_get_visual(GTK_WIDGET(vp->da)),FALSE);*/
      S_colormap=gdk_colormap_get_system();
   }
   
   gtk_widget_set_colormap(GTK_WIDGET(vp->da),S_colormap);
   
   vp->win=GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL)); /* TODO: delete viewport when closed, guard against closing all */
   gtk_widget_set_colormap(GTK_WIDGET(vp->win),S_colormap);
   gtk_window_set_title(vp->win,"Viewport");
   gtk_window_set_policy(vp->win,1,1,0);
   gtk_window_set_transient_for(vp->win,S_menu_window);

   gtk_signal_connect(GTK_OBJECT(vp->win),"focus_in_event",GTK_SIGNAL_FUNC(vp_focus),vp);

   gtk_widget_set_events(GTK_WIDGET(vp->win),
      GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK|
      GDK_KEY_PRESS_MASK|GDK_KEY_RELEASE_MASK);
   gtk_signal_connect(GTK_OBJECT(vp->win),"key_press_event",GTK_SIGNAL_FUNC(S_in_key),NULL);
   gtk_signal_connect(GTK_OBJECT(vp->win),"key_release_event",GTK_SIGNAL_FUNC(S_in_key),NULL);
   gtk_signal_connect(GTK_OBJECT(vp->win),"button_press_event",GTK_SIGNAL_FUNC(S_in_mouse),&vp->num);
   gtk_signal_connect(GTK_OBJECT(vp->win),"button_release_event",GTK_SIGNAL_FUNC(S_in_mouse),&vp->num);

   gtk_widget_set_events(GTK_WIDGET(vp->da),GDK_EXPOSURE_MASK);
   gtk_widget_set_usize(GTK_WIDGET(vp->da),x,y);
   gtk_container_add(GTK_CONTAINER(vp->win),GTK_WIDGET(vp->da));


   if (first)
   {
      gtk_widget_realize(GTK_WIDGET(vp->da));
      gtk_widget_realize(GTK_WIDGET(vp->win));
      /* This needs a valid viewport to create gc:s and stuff, so
      call it now. */
      S_3D_Init(vp);
   }


   gtk_signal_connect(GTK_OBJECT(vp->da),"configure_event",GTK_SIGNAL_FUNC(vp_configure),vp);
   gtk_signal_connect(GTK_OBJECT(vp->da),"expose_event",GTK_SIGNAL_FUNC(vp_expose),vp);

   gtk_widget_show(GTK_WIDGET(vp->da));
   gtk_widget_show(GTK_WIDGET(vp->win));

   S_3D_Render(vp);
}


#include "plugin.h"
static void Plugin_Init(void);
static plugin_t g_3d_plugin=
{{Plugin_Init},{"basic/gtk_3d","Fundamental viewport commands (GTK)."}};


void S_InitViewports(void)
{
   NewViewport(320,240,1);
   cur_vport=-1;
   S_SetCurViewport(0);
   NewViewport(320,240,0);
   NewViewport(320,480,0);

   Plugin_Register(&g_3d_plugin);
}


#include "cmds.h"

#include "camera.h"

static void C_gtk_toggle_3d(void)
{
   if (S_cur_gvport->mode==M_3D_WIRE)
   {
      S_cur_gvport->mode=M_2D_WIRE;
      S_cur_gvport->vp.axis_aligned=1;
   }
   else
   {
      S_cur_gvport->mode=M_3D_WIRE;
      S_cur_gvport->vp.axis_aligned=0;
   }

   Cmd_SetRedraw(1);
}

static void C_gtk_new_viewport(void)
{
   NewViewport(320,240,0);
   InitViewport(&vports[num_vports-1]->vp);
   S_3D_Render(vports[num_vports-1]);
}


static cmd_reg_t g_3d_cmds[]={
{"gtk.toggle_3d",C_gtk_toggle_3d,{"View/Toggle 3d","f"},{},"Toggle 2d/3d view in current viewport."},
{"gtk.new_viewport",C_gtk_new_viewport,{"View/New viewport",NULL},{},"Create a new viewport."},
};

static void Plugin_Init(void)
{
   Cmd_Register(g_3d_cmds,sizeof(g_3d_cmds));
}

