#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "matrix.h"


void InitMatrix(matrix_t matrix)
{
	int i, j;

	for (i=0; i<4; i++)
	{
		for (j=0; j<4; j++)
		{
			if (i == j)
				matrix[i][j] = 1;
			else
				matrix[i][j] = 0;
		}
	}
}

void MultMatrix(matrix_t one,matrix_t two,matrix_t three)
{
	int      i, j;
	matrix_t prod;

	for (i=0; i<4; i++)
	{
		for (j=0; j<4; j++)
		{
			prod[i][j] = (one[i][0] * two[0][j]) + (one[i][1] * two[1][j]) +
							 (one[i][2] * two[2][j]) + (one[i][3] * two[3][j]);
		}
	}
   
	for (i=0; i<4; i++)
		for (j=0; j<4; j++)
			three[i][j] = prod[i][j];
}

scalar_t DetMatrix(matrix_t m)
{
   double d=0;

#define DET(a0,a1,a2, b0,b1,b2) \
   (m[a0][b0]*(m[a1][b1]*m[a2][b2]-m[a2][b1]*m[a1][b2]) - \
    m[a1][b0]*(m[a0][b1]*m[a2][b2]-m[a2][b1]*m[a0][b2]) + \
    m[a2][b0]*(m[a0][b1]*m[a1][b2]-m[a1][b1]*m[a0][b2]))

   d+=m[0][0]*DET(1,2,3, 1,2,3);
   d-=m[1][0]*DET(0,2,3, 1,2,3);
   d+=m[2][0]*DET(0,1,3, 1,2,3);
   d-=m[3][0]*DET(0,1,2, 1,2,3);
   return d;
}


int InvMatrix(matrix_t m,matrix_t dst)
{
   double d;

   d=DetMatrix(m);
   if (!d) return 0;
   d=1/d;

   dst[0][0]= d*DET(1,2,3, 1,2,3);
   dst[0][1]=-d*DET(0,2,3, 1,2,3);
   dst[0][2]= d*DET(0,1,3, 1,2,3);
   dst[0][3]=-d*DET(0,1,2, 1,2,3);
   
   dst[1][0]=-d*DET(1,2,3, 0,2,3);
   dst[1][1]= d*DET(0,2,3, 0,2,3);
   dst[1][2]=-d*DET(0,1,3, 0,2,3);
   dst[1][3]= d*DET(0,1,2, 0,2,3);
   
   dst[2][0]= d*DET(1,2,3, 0,1,3);
   dst[2][1]=-d*DET(0,2,3, 0,1,3);
   dst[2][2]= d*DET(0,1,3, 0,1,3);
   dst[2][3]=-d*DET(0,1,2, 0,1,3);
   
   dst[3][0]=-d*DET(1,2,3, 0,1,2);
   dst[3][1]= d*DET(0,2,3, 0,1,2);
   dst[3][2]=-d*DET(0,1,3, 0,1,2);
   dst[3][3]= d*DET(0,1,2, 0,1,2);

   return 1;
#undef DET
}


void PrintMatrix(const char *name,matrix_t m)
{
   int i,j;

   printf("%s=\n",name);
   for (i=0;i<4;i++)
   {
      printf("  (");
      for (j=0;j<4;j++)
      {
         printf(" %8.2f",m[i][j]);
      }
      printf(")\n");
   }
}


void GenerateIRotMatrix(matrix_t m,scalar_t rx,scalar_t ry,scalar_t rz)
{
   matrix_t	rot1,rot2,rot3,rot4;
   scalar_t r_sx,r_cx;
   scalar_t r_sy,r_cy;
   scalar_t r_sz,r_cz;
   int      i,j;

   r_sx=sin(rx*PI/180);
   r_cx=cos(rx*PI/180);

   r_sy=sin(ry*PI/180);
   r_cy=cos(ry*PI/180);

   r_sz=sin(rz*PI/180);
   r_cz=cos(rz*PI/180);

   InitMatrix(rot1);
   rot1[0][0]= 1;
   rot1[0][1]= 0;
   rot1[0][2]= 0;

   rot1[1][0]= 0;
   rot1[1][1]= r_cx;
   rot1[1][2]= r_sx;

   rot1[2][0]= 0;
   rot1[2][1]=-r_sx;
   rot1[2][2]= r_cx;

   InitMatrix(rot2);
   rot2[0][0]= r_cz;
   rot2[0][1]= r_sz;
   rot2[0][2]= 0;

   rot2[1][0]=-r_sz;
   rot2[1][1]= r_cz;
   rot2[1][2]= 0;

   rot2[2][0]= 0;
   rot2[2][1]= 0;
   rot2[2][2]= 1;

   InitMatrix(rot3);
   rot3[0][0]= r_cy;
   rot3[0][1]= r_sy;
   rot3[0][2]= 0;

   rot3[1][0]=-r_sy;
   rot3[1][1]= r_cy;
   rot3[1][2]= 0;

   rot3[2][0]= 0;
   rot3[2][1]= 0;
   rot3[2][2]= 1;

   MultMatrix(rot2,rot1,rot4);

   MultMatrix(rot4,rot3,rot2);

   for (i=0;i<4;i++)
      rot2[1][i]=-rot2[1][i];

   for (i=0;i<4;i++)
      for (j=0;j<4;j++)
         m[i][j]=rot2[i][j];
}

void GenerateRotMatrix(matrix_t m,scalar_t rx,scalar_t ry,scalar_t rz)
{
   matrix_t	rot1,rot2,rot3;
   scalar_t r_sx,r_cx;
   scalar_t r_sy,r_cy;
   scalar_t r_sz,r_cz;
   int      i,j;

   r_sx=sin(rx*PI/180);
   r_cx=cos(rx*PI/180);

   r_sy=sin(ry*PI/180);
   r_cy=cos(ry*PI/180);

   r_sz=sin(rz*PI/180);
   r_cz=cos(rz*PI/180);

   InitMatrix(rot1);
   rot1[0][0]= 1;
   rot1[0][1]= 0;
   rot1[0][2]= 0;

   rot1[1][0]= 0;
   rot1[1][1]= r_cx;
   rot1[1][2]= r_sx;

   rot1[2][0]= 0;
   rot1[2][1]=-r_sx;
   rot1[2][2]= r_cx;

   InitMatrix(rot2);
   rot2[0][0]= r_cz;
   rot2[0][1]= r_sz;
   rot2[0][2]= 0;

   rot2[1][0]=-r_sz;
   rot2[1][1]= r_cz;
   rot2[1][2]= 0;

   rot2[2][0]= 0;
   rot2[2][1]= 0;
   rot2[2][2]= 1;

   MultMatrix(rot1,rot2,rot3);

   InitMatrix(rot2);
   rot2[0][0]= r_cy;
   rot2[0][1]= r_sy;
   rot2[0][2]= 0;

   rot2[1][0]=-r_sy;
   rot2[1][1]= r_cy;
   rot2[1][2]= 0;

   rot2[2][0]= 0;
   rot2[2][1]= 0;
   rot2[2][2]= 1;

   MultMatrix(rot2,rot3,rot1);

   for (i=0;i<4;i++)
      rot1[i][1]=-rot1[i][1];

   for (i=0;i<4;i++)
      for (j=0;j<4;j++)
         m[i][j]=rot1[i][j];
}

