#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "camera.h"
#include "edprim.h"
#include "edsel.h"
#include "error.h"
#include "memory.h"
#include "quest.h"
#include "fe/3d.h"
#include "fe/gui.h"


static void C_cube(void)
{
   vec3_t d;
   struct brush_s *b;

   Move(S_cur_vport, MOVE_FORWARD, &d, 100);
   b=AddBrush(PRIM_TEXRECT,
      S_cur_vport->camera_pos.x + d.x,
      S_cur_vport->camera_pos.y + d.y,
      S_cur_vport->camera_pos.z + d.z,
      32,
      32,
      32,
      0);
   if (b)
   {
      ClearSelAll();
      AddSelBrush(b);
   }
}
static func_t F_cube=
{"brush.create.cube",FUNC_VOID,{fv:C_cube},0,0,0,
{"Create a cube"}};


/* Create a room made of six wall brushes */
/* TODO: default wall/ceiling/floor textures? */
static void C_room(void)
{
	int    i;
	char **keys;
	char **values;
	int    x, y, z, t;
	vec3_t d;
   struct brush_s *b;

	Move(S_cur_vport, MOVE_FORWARD, &d, 100);

	keys = (char **)Q_malloc(sizeof(char *) * 4);
	values = (char **)Q_malloc(sizeof(char *) * 4);
	if (!keys || !values)
	{
		Error_Add("brush_create_room: Out of memory!");
		return;
	}

	for (i=0; i<4; i++)
	{
		keys[i] = (char *)Q_malloc(sizeof(char) * 30);
		values[i] = (char *)Q_malloc(sizeof(char) * 10);
		if (!keys[i] || !values[i])
		{
			Error_Add("brush_create_room: Out of memory!");
			return;
		}
	}

	strcpy(keys[0], "X size");
	strcpy(keys[1], "Y size");
	strcpy(keys[2], "Z size");
	strcpy(keys[3], "Wall thickness");
	sprintf(values[0], "256");
	sprintf(values[1], "256");
	sprintf(values[2], "256");
	sprintf(values[3], "16");

	if (S_GUI_PopEntity("Create Room", keys, values, 4))
	{
		x = atoi(values[0]);
		y = atoi(values[1]);
		z = atoi(values[2]);
		t = atoi(values[3]);

      ClearSelAll();

		b=AddBrush(PRIM_CUBE,
				S_cur_vport->camera_pos.x + d.x + ((x + t)>>1),
				S_cur_vport->camera_pos.y + d.y,
				S_cur_vport->camera_pos.z + d.z,
				t, y, z, 1
						);
      if (b) AddSelBrush(b);
      
		b=AddBrush(PRIM_CUBE,
				S_cur_vport->camera_pos.x + d.x - ((x + t)>>1),
				S_cur_vport->camera_pos.y + d.y,
				S_cur_vport->camera_pos.z + d.z,
				t, y, z, 1
						);
      if (b) AddSelBrush(b);
      
		b=AddBrush(PRIM_CUBE,
				S_cur_vport->camera_pos.x + d.x,
				S_cur_vport->camera_pos.y + d.y + ((y + t)>>1),
				S_cur_vport->camera_pos.z + d.z,
				x, t, z, 1
						);
      if (b) AddSelBrush(b);
      
		b=AddBrush(PRIM_CUBE,
				S_cur_vport->camera_pos.x + d.x,
				S_cur_vport->camera_pos.y + d.y - ((y + t)>>1),
				S_cur_vport->camera_pos.z + d.z,
				x, t, z, 1
						);
      if (b) AddSelBrush(b);
      
		b=AddBrush(PRIM_CUBE,
				S_cur_vport->camera_pos.x + d.x,
				S_cur_vport->camera_pos.y + d.y,
				S_cur_vport->camera_pos.z + d.z + ((z + t)>>1),
				x, y, t, 1
						);
      if (b) AddSelBrush(b);
      
		b=AddBrush(PRIM_CUBE,
				S_cur_vport->camera_pos.x + d.x,
				S_cur_vport->camera_pos.y + d.y,
				S_cur_vport->camera_pos.z + d.z - ((z + t)>>1),
				x, y, t, 1
						);
      if (b) AddSelBrush(b);
	}

	for (i=0; i<4; i++)
	{
		Q_free(keys[i]);
		Q_free(values[i]);
	}
	Q_free(keys);
	Q_free(values);
}
static func_t F_room=
{"brush.create.room",FUNC_VOID,{fv:C_room},0,0,0,
{"Create a room"}};


static void CreateBrush(int type)
{
#define MAX_KEYS 4
typedef struct
{
   int typenum;
   int num;
   const char *name;
   const char *(keys[MAX_KEYS]);
   const char *(def_val[MAX_KEYS]);
} prim_type_t;
/* TODO: global table, pass pointer to struct */
static prim_type_t prims[]=
{

/*
These MUST match the order of parameters to the actual functions
above.
*/
{PRIM_NPRISM,4,"Prism",
  {"Sides","Radius","Height","Eccentricity"},
  {"12"   ,"64"    ,"64"    ,"1.0"}
},

{PRIM_PYRAMID,4,"Pyramid",
  {"Sides","Radius","Height","Eccentricity"},
  {"4"    ,"64"    ,"64"    ,"1.0"}
},

{PRIM_DODEC,2,"Dodecahedron",
  {"Radius","Eccentricity"},
  {"100"   ,"1.0"}
},

{PRIM_ICOS,2,"Icosahedron",
  {"Length of edge","Eccentricity"},
  {"100"           ,"1.0"}
},

{PRIM_BUCKY,2,"Buckyball",
  {"Radius","Eccentricity"},
  {"100"   ,"1.0"}
},

{PRIM_TORUS,4,"Torus",
  {"Minor Sides","Minor Radius","Major Sides","Major Radius"},
  {"10"         ,"20"          ,"8"          ,"100"}
},

{0}
};

	int i;
	const char **keys;
	char **values;
	float a[MAX_KEYS];
   vec3_t d;

   prim_type_t *p;

   for (p=prims;p->typenum;p++)
      if (p->typenum==type)
         break;
   if (!p->typenum)
   {
      ERROR("CreateBrush: Unknown brush type %i!",type);
      return;
   }

	Move(S_cur_vport, MOVE_FORWARD, &d, 100);

	keys = p->keys;
	values = (char **)Q_malloc(sizeof(char *) * p->num);

	if (!values)
	{
		ERROR("CreateBrush: Out of memory!");
		return;
	}

	for (i=0; i<p->num; i++)
	{
		values[i] = (char *)Q_malloc(sizeof(char) * 10);
		if (!values[i])
		{
			ERROR("CreateBrush: Out of memory!");
			return;
		}

      strcpy(values[i],p->def_val[i]);
	}

	if (S_GUI_PopEntity(p->name,(char **)keys, values, p->num))
	{
      struct brush_s *b;
      
      for (i=0;i<p->num;i++)
         a[i]=atof(values[i]);

		b=AddBrush(type,
				S_cur_vport->camera_pos.x + d.x,
				S_cur_vport->camera_pos.y + d.y,
				S_cur_vport->camera_pos.z + d.z,
				a[0], a[1], a[2], a[3]
				);
      if (b)
      {
         ClearSelAll();
         AddSelBrush(b);
      }
	}

	for (i=0; i<p->num; i++)
	{
		Q_free(values[i]);
	}
	Q_free(values);
}


static void C_prism  (void) { CreateBrush(PRIM_NPRISM ); }
static func_t F_prism  ={"brush.create.prism"  ,FUNC_VOID,{fv:C_prism  },0,0,0,{"Create a prism"}};

static void C_pyramid(void) { CreateBrush(PRIM_PYRAMID); }
static func_t F_pyramid={"brush.create.pyramid",FUNC_VOID,{fv:C_pyramid},0,0,0,{"Create a pyramid"}};

static void C_dodec  (void) { CreateBrush(PRIM_DODEC  ); }
static func_t F_dodec  ={"brush.create.dodec"  ,FUNC_VOID,{fv:C_dodec  },0,0,0,{"Create a dodecahedron"}};

static void C_icos   (void) { CreateBrush(PRIM_ICOS   ); }
static func_t F_icos   ={"brush.create.icos"   ,FUNC_VOID,{fv:C_icos   },0,0,0,{"Create a icosahedron"}};

static void C_bucky  (void) { CreateBrush(PRIM_BUCKY  ); }
static func_t F_bucky  ={"brush.create.bucky"  ,FUNC_VOID,{fv:C_bucky  },0,0,0,{"Create a buckyball"}};

static void C_torus  (void) { CreateBrush(PRIM_TORUS  ); }
static func_t F_torus  ={"brush.create.torus"  ,FUNC_VOID,{fv:C_torus  },0,0,0,{"Create a torus"}};


static func_t *funcs[]=
{&F_cube,&F_room,&F_prism,&F_pyramid,&F_dodec,&F_icos,&F_bucky,&F_torus};

#if 0
/* TODO: document */
static cmd_reg_t cmds[]={
{"brush.create.cube",C_brush_create_cube,{"Brush/Create/Cube"   },{},"Create a cube."},
{"brush.create.room",C_brush_create_room,{"Brush/Create/Room..."},{},"Create a room."},

{"brush.create.prism"  ,C_prism  ,{"Brush/Create/Prism..."       },{},"Create a prism."        },
{"brush.create.pyramid",C_pyramid,{"Brush/Create/Pyramid..."     },{},"Create a pyramid."      },
{"brush.create.dodec"  ,C_dodec  ,{"Brush/Create/Dodecahedron..."},{},"Create a dodecahedron." },
{"brush.create.icos"   ,C_icos   ,{"Brush/Create/Icosahedron..." },{},"Create an icosahedron." },
{"brush.create.bucky"  ,C_bucky  ,{"Brush/Create/Buckyball..."   },{},"Create a buckyball."    },
{"brush.create.torus"  ,C_torus  ,{"Brush/Create/Torus..."       },{},"Create a torus."        },
};
#endif

#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"brush.create","Brush creation (cube, room, etc.)."}};

PLUGIN_FOOT_FUNCS

