#include <stdio.h>
#include <stdlib.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "brush.h"
#include "csg.h"
#include "edsel.h"
#include "error.h"
#include "group.h"
#include "memory.h"
#include "quest.h"
#include "fe/message.h"


static void AddBrushes(brush_t *b1, brush_t *b2)
{
	int i,j;
	cbrush_t A;
	brush_t *b;

	if (b1->bt->type!=BRUSH_NORMAL || b2->bt->type!=BRUSH_NORMAL)
	{
		S_NewMessage("CSG Intersect: Can only use normal brushes for CSG operations!");
		return;
	}

	A.num_faces = b1->num_faces + b2->num_faces;
	A.faces = (cface_t *)Q_malloc(sizeof(cface_t) * A.num_faces);

	if (!A.faces)
	{
		Error_Add("Out of memory!");
		return;
	}

	for (i=j=0;i<b1->num_faces;i++,j++)
		if (!CSG_FaceToCFace(&A.faces[j],b1,i))
		{
			Error_Add("Out of memory!");
			return;
		}

	for (i=0;i<b2->num_faces;i++,j++)
		if (!CSG_FaceToCFace(&A.faces[j],b2,i))
		{
			Error_Add("Out of memory!");
			return;
		}

	b=CSG_BuildBrush(NULL,&A);
	if (!b)
	{
		Error_Add("Can't create new brush! (brushes don't intersect?)");
		return;
	}

	ClearSelVerts();
	ClearSelBrushes();

	b->Entity=M.WorldSpawn;
	b->Group=FindVisGroup(b1->Group);
	B_Link(b);
	AddSelBrush(b);
}


static void C_intersect(void)
{
	if (M.sel.num_bselected!=2)
	{
		S_NewMessage("CSG Intersect: Must have exactly 2 brushes selected!");
		return;
	}

	AddBrushes(M.sel.bsel->Brush,M.sel.bsel->Next->Brush);
}

static func_t F_csg_intersect=
{"csg.intersect",FUNC_VOID,{fv:C_intersect},NULL,NULL,0,
{"CSG intersection.",
"This does a CSG intersection of two selected brushes (which must "
"intersect), i.e. it creates a new brush that covers the space where "
"the two brushes intersect."}
};

static func_t *funcs[]={&F_csg_intersect};


#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"csg.intersect","CSG intersect command."}};

PLUGIN_FOOT_FUNCS

