#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "brush.h"
#include "config.h"
#include "csg.h"
#include "edsel.h"
#include "error.h"
#include "group.h"
#include "memory.h"
#include "quest.h"
#include "status.h"
#include "fe/message.h"


static void CSG_MakeHollow(brush_t *b)
{
	int i,j;
	brush_t *temp;
	cbrush_t t;

	for (i=0;i<b->num_faces;i++)
	{
		t.num_faces=b->num_faces+1;
		t.faces=Q_malloc(sizeof(cface_t)*t.num_faces);
		if (!t.faces)
		{
			Error_Add("Out of memory!");
			return;
		}
		for (j=0;j<b->num_faces;j++)
		{
			if (!CSG_FaceToCFace(&t.faces[j],b,j))
			{
				Error_Add("Out of memory!");
				return;
			}
		}

/* new plane */
		if (!CSG_FaceToCFace(&t.faces[j],b,i))
		{
			Error_Add("Out of memory!");
			return;
		}
		t.faces[j].p.n.x =-t.faces[j].p.n.x;
		t.faces[j].p.n.y =-t.faces[j].p.n.y;
		t.faces[j].p.n.z =-t.faces[j].p.n.z;
		t.faces[j].p.dist=-t.faces[j].p.dist+cv_edit_snap_size->i;

		temp=CSG_BuildBrush(NULL,&t);
		if (!temp)
		{
			Error_Add("Error building brush!");
			return;
		}

		temp->Entity=M.WorldSpawn;
		temp->Group=FindVisGroup(b->Group);
		B_Link(temp);
	}
}

static void C_hollow(void)
{
	bsel_t *b;
	int warn;

	if (!M.sel.num_bselected)
	{
		S_NewMessage("CSG Hollow: No brushes selected!");
		return;
	}
	ClearSelVerts();

	warn=0;
	for (b=M.sel.bsel;b;b=b->Next)
	{
		if (b->Brush->bt->type!=BRUSH_NORMAL)
		{
			if (!warn)
			{
				S_NewMessage("Warning: Can only use normal brushes for CSG operations.");
				warn=1;
			}
		}
		else
		{
			CSG_MakeHollow(b->Brush);
			B_Unlink(b->Brush);
			B_Free(b->Brush);
		}
	}

	ClearSelBrushes();
}


static void CSG_MakeRoom(brush_t *b)
{
	int i,j;
	brush_t *temp;
	cbrush_t t;

	for (i=0;i<b->num_faces;i++)
	{
		t.num_faces=b->num_faces*2;
		t.faces=Q_malloc(sizeof(cface_t)*t.num_faces);
		if (!t.faces)
		{
			Error_Add("Out of memory!");
			return;
		}
		for (j=0;j<b->num_faces;j++)
		{
			if (!CSG_FaceToCFace(&t.faces[j],b,j))
			{
				Error_Add("Out of memory!");
				return;
			}
		}
/* add new planes */
		for (;j<t.num_faces;j++)
		{
			if (!CSG_FaceToCFace(&t.faces[j],b,j-b->num_faces))
			{
				Error_Add("Out of memory!");
				return;
			}

			t.faces[j].p.dist-=cv_edit_snap_size->i;
			if (j-b->num_faces==i)
			{
				t.faces[j].p.n.x=-t.faces[j].p.n.x;
				t.faces[j].p.n.y=-t.faces[j].p.n.y;
				t.faces[j].p.n.z=-t.faces[j].p.n.z;
				t.faces[j].p.dist=-t.faces[j].p.dist;
			}
		}

/* Create final brush */
		temp=CSG_BuildBrush(NULL,&t);
		if (!temp)
		{
			Error_Add("Error building brush!");
			return;
		}

		temp->Entity=M.WorldSpawn;
		temp->Group=FindVisGroup(b->Group);
		B_Link(temp);
	}
}

static void C_room(void)
{
	bsel_t *b;
	int warn;
	
	if (!M.sel.num_bselected)
	{
		S_NewMessage("CSG Room: No brushes selected!");
		return;
	}
	ClearSelVerts();

	warn=0;
	for (b=M.sel.bsel;b;b=b->Next)
	{
		if (b->Brush->bt->type!=BRUSH_NORMAL)
		{
			if (!warn)
			{
				S_NewMessage("Warning: Can only use normal brushes for CSG operations.");
				warn=1;
			}
		}
		else
		{
			CSG_MakeRoom(b->Brush);
			B_Unlink(b->Brush);
			B_Free(b->Brush);
		}
	}

	ClearSelBrushes();
}


static func_t F_csg_hollow=
{"csg.hollow",FUNC_VOID,{fv:C_hollow},NULL,NULL,0,
{"CSG hollow."}
};
static func_t F_csg_room=
{"csg.room",FUNC_VOID,{fv:C_room},NULL,NULL,0,
{"CSG room."}
};

static func_t *funcs[]={&F_csg_hollow,&F_csg_room};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"csg.rmhl","CSG make hollow/room command."}};

PLUGIN_FOOT_FUNCS

