#include <stdio.h>
#include <malloc.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "memory.h"
#include "quest.h"

/*
																								of memory 1572
brushes		 2788	  1		 126	 1			710   1       499	 1		 228     228    14.5
faces			16129	  5.78	 757	 6.00	  4302   6.05   3004	 6.02  192    1152    73.3
vertices		24187	  8.67	1010	 8.01	  5720	8.05   4012	 8.04   12      96     6.1
fvertices	67017	 24.03	3030	24.04	 17204  24.23	12036	24.12    4      96     6.1

entities		  591	  0.21	  14	 0.11		 56	0.07     42	 0.08

*/


static void debug_heap(mem_heap_t *h)
{
	int ts,tf;
	int i;
	printf("          heap: %p\n",h);
	printf("        chunks: %4i\n",h->n_chunks);
	printf("	   chunk size: %4i (%8i)\n",h->csize,h->csize*h->esize);
	printf("  element size: %4i\n",h->esize);
	for (ts=tf=i=0;i<h->n_chunks;i++)
	{
		printf("  chunk %3i: %4i/%4i (%8i/%8i)\n",
			i,h->c[i].free,h->csize,h->c[i].free*h->esize,h->csize*h->esize);
		ts+=h->csize,tf+=h->c[i].free;
	}
	printf("total elements: %4i (%8i)\n",ts,ts*h->esize);
	printf(" free elements: %4i (%8i)\n",tf,tf*h->esize);
	printf(" used elements: %4i (%8i)\n",ts-tf,ts*h->esize-tf*h->esize);
}


static void C_dbg(void)
{
	struct mallinfo mi;
	mi=mallinfo();
	printf("----- memory info -----\n");
	printf("                arena: %8i\n",mi.arena);
	printf("    chunks not in use: %4i\n",mi.ordblks);
	printf("allocated mmap chunks: %4i\n",mi.hblks);
	printf(" allocated mmap bytes: %8i\n",mi.hblkhd);
	printf("      allocated bytes: %8i\n",mi.uordblks);
	printf("           free bytes: %8i\n",mi.fordblks);
	printf("----- struct info -----\n");
	printf(
		"  brush_t: %4i %4i\n"
		" entity_t: %4i %4i\n"
		"  group_t: %4i\n"
		"   face_t: %4i %4i\n",
		sizeof(brush_t),mem_brush.esize,
		sizeof(entity_t),mem_entity.esize,
		sizeof(group_t),
		sizeof(face_t),mem_face.esize);
	{
		int nbrush,nface,nent,ngroup,nvert,fvert;
		brush_t *b;
		entity_t *e;
		group_t *g;
		int i;

		nface=nbrush=nvert=0;
		fvert=0;
		for (b=M.BrushHead;b;b=b->Next)
		{
			nbrush++;
			nface+=b->num_faces;
			nvert+=b->num_verts;
			for (i=0;i<b->num_faces;i++)
				fvert+=b->faces[i]->num_verts;
		}
		for (nent=0,e=M.EntityHead;e;e=e->Next)
			nent++;
		for (ngroup=0,g=M.GroupHead;g;g=g->Next)
			ngroup++;

		printf("----- map info -----\n");
		printf(
			"  brushes: %5i (%4i) %8i\n"
			"      faces: %6i         %8i\n"
			"   vertices: %6i         %8i\n"
			" f vertices: %6i         %8i\n"
			" entities: %5i (%4i) %8i\n"
			"   groups: %5i (%4i) %8i\n",
			nbrush,M.num_brushes,nbrush*sizeof(brush_t),
			nface,nface*sizeof(face_t),
			nvert,nvert*sizeof(vec3_t),
			fvert,fvert*sizeof(int),
			nent,M.num_entities,nent*sizeof(entity_t),
			ngroup,M.num_groups,ngroup*sizeof(group_t));
	}
/*	printf("----- brush heap -----\n");
	debug_heap(&mem_brush);
	printf("----- entity heap -----\n");
	debug_heap(&mem_entity);
	printf("----- face heap -----\n");
	debug_heap(&mem_face);*/
}


static cmd_reg_t cmds[]={
{"gtk.debug_info",C_dbg,{"Misc/Debug info","control+alt+d"},{CMD_NOUNDO},"Dumps some information."},
};

#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/gtk/debug",""}};

PLUGIN_FOOT_CMDS

