#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gtk/gtk.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "hook.h"
#include "s_message.h"


#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/gtk/help","Interactive help for commands."}};


static gint help_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),gpointer data)
{
   gtk_widget_destroy(GTK_WIDGET(data));
   return TRUE;
}

static void help_ok(ARG_UNUSED(GtkWidget *b),gpointer data)
{
   gtk_widget_destroy(GTK_WIDGET(data));
}

static void Help(cmd_reg_t *cmd)
{
   GtkWindow *win;
   GtkText *text;
   GtkWidget *tb,*vb;
   
   GdkFont *bold_font;

   Hook_RemoveFunc(hk_cmd_start,(hook_func_void_t)Help);
   
   win=GTK_WINDOW(gtk_window_new(GTK_WINDOW_DIALOG));
   gtk_window_set_title(win,"Help");
   gtk_window_set_position(win,GTK_WIN_POS_MOUSE);
   gtk_signal_connect(GTK_OBJECT(win),"delete_event",GTK_SIGNAL_FUNC(help_delete),win);
   gtk_container_set_border_width(GTK_CONTAINER(win),8);
   gtk_window_set_default_size(win,300,225);
   gtk_window_set_policy(win,1,1,0);

   text=GTK_TEXT(gtk_text_new(NULL,NULL));
   gtk_text_set_word_wrap(text,TRUE);

   tb=gtk_table_new(2,2,FALSE);
   gtk_table_attach(GTK_TABLE(tb),GTK_WIDGET(text),0,1,0,1,
      GTK_FILL|GTK_EXPAND|GTK_SHRINK,
      GTK_FILL|GTK_EXPAND|GTK_SHRINK,0,0);

   vb=gtk_vscrollbar_new(text->vadj);
   gtk_table_attach(GTK_TABLE(tb),vb,1,2,0,1,
      GTK_FILL,GTK_EXPAND|GTK_FILL|GTK_SHRINK,0,0);
   gtk_widget_show(vb);

   vb=gtk_button_new_with_label("Close");
   gtk_signal_connect(GTK_OBJECT(vb),"clicked",GTK_SIGNAL_FUNC(help_ok),win);
   gtk_table_attach(GTK_TABLE(tb),vb,0,1,1,2,
      GTK_EXPAND|GTK_SHRINK,0,0,0);
   gtk_widget_show(vb);
   
   gtk_container_add(GTK_CONTAINER(win),tb);
   gtk_widget_grab_focus(vb);

   gtk_widget_show(GTK_WIDGET(text));
   gtk_widget_show(tb);

   gtk_text_freeze(text);

   bold_font=gdk_font_load("-*-times-bold-r-*-*-*-120-*-*-*-*-*-*");
   
   gtk_text_insert(text,bold_font,NULL,NULL,
      "Command: ",-1);
   gtk_text_insert(text,NULL,NULL,NULL,
      cmd->name,-1);
   if (cmd->d.menu)
   {
      gtk_text_insert(text,bold_font,NULL,NULL,
         "\nMenu: ",-1);
      gtk_text_insert(text,NULL,NULL,NULL,
         cmd->d.menu,-1);
   }
   if (cmd->d.key)
   {
      gtk_text_insert(text,bold_font,NULL,NULL,
         "\nKey: ",-1);
      gtk_text_insert(text,NULL,NULL,NULL,
         cmd->d.key,-1);
   }
   if (cmd->desc)
   {
      gtk_text_insert(text,NULL,NULL,NULL,
         "\n\n",-1);
      gtk_text_insert(text,NULL,NULL,NULL,
         cmd->desc,-1);
   }
   if (cmd->help_text)
   {
      gtk_text_insert(text,NULL,NULL,NULL,
         "\n\n",-1);
      gtk_text_insert(text,NULL,NULL,NULL,
         cmd->help_text,-1);
   }

   gtk_text_thaw(text);

   gtk_widget_show(GTK_WIDGET(win));
}


static void C_help_mode(void)
{
   S_NewMessage("Help-mode activated, select a command to get help on it.");
   Hook_AddFunc(hk_cmd_start,HOOK_FUNC_VOID,(hook_func_void_t)Help,NULL,&plugin);
}

static cmd_reg_t cmds[]={
{"help_mode",C_help_mode,{"Help/Help-mode","control+h"},{},
"Activates help-mode.",
"This command activates help-mode. When you activate the next command (by "
"e.g pressing a key, clicking somewhere, selecting something from the "
"menu, or whatever), instead of running it a help box will be brought "
"up with information about the command, including what key and what "
"menu item activates it, and a description of what it does and how to "
"use it. After you have selected one command, everything goes back to "
"normal."
},
};

PLUGIN_FOOT_CMDS

