/* Group editor */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "glade_callbacks.h"
#include "glade_interface.h"
#include "glade_support.h"

#undef TRUE
#undef FALSE

#include "quest/defines.h"
#include "quest/types.h"


#include "quest/brush.h"
#include "quest/entity.h"
#include "quest/error.h"
#include "quest/group.h"
#include "quest/map.h"
#include "quest/quest.h"
#include "quest/uid.h"
#include "quest/undo.h"
#include "quest/fe/gui.h"
#include "quest/fe/render.h"


#define GROUP_DATA_FIELD "quest-group"

static GtkWidget *ge;
static GtkTree *tree;
static GtkEntry *name,*curgroup;
static GtkToggleButton *hidden,*sselect;

static group_t *selection;

static int changes;


static void fill_cur(void)
{
   group_t *g;

   g=UID_Group(M.CurGroup);
   if (g)
      gtk_entry_set_text(curgroup,g->fullname);
   else
      gtk_entry_set_text(curgroup,"(invalid)");
}

static void fill_tree_r(GtkTree *t,group_t *g)
{
   GtkWidget *w;
   group_t *c;
   map_uid_t uc;
   GtkTree *t2;

   w=gtk_tree_item_new_with_label(g->groupname);
   gtk_object_set_data(GTK_OBJECT(w),GROUP_DATA_FIELD,g);
   gtk_widget_show(w);
   gtk_tree_append(t,w);

   if (g->child)
   {
      t2=GTK_TREE(gtk_tree_new());
      for (uc=g->child;uc;uc=c->cnext)
      {
         c=UID_Group(uc);
         fill_tree_r(t2,c);
      }
      gtk_tree_item_set_subtree(GTK_TREE_ITEM(w),GTK_WIDGET(t2));
   }
   gtk_tree_item_expand(GTK_TREE_ITEM(w));
}

static void fill_tree(void)
{
   group_t *g;
   gtk_tree_clear_items(tree,0,1);
   g=UID_Group(M.WorldGroup);
   fill_tree_r(tree,g);
}


static void fill_data(void)
{
   if (!selection)
   {
      gtk_entry_set_text(name,"");
      gtk_toggle_button_set_active(hidden,0);
      gtk_toggle_button_set_active(sselect,0);
      return;
   }
   gtk_entry_set_text(name,selection->groupname);
   gtk_toggle_button_set_active(hidden,selection->flags&GROUP_HIDDEN);
   gtk_toggle_button_set_active(sselect,selection->flags&GROUP_SMARTSELECT);
}


gboolean
ge_delete_event                        (ARG_UNUSED(GtkWidget       *widget),
                                        ARG_UNUSED(GdkEvent        *event),
                                        ARG_UNUSED(gpointer         user_data))
{
   gtk_main_quit();
   return TRUE;
}

void
ge_select                              (ARG_UNUSED(GtkTree         *ptree),
                                        ARG_UNUSED(gpointer         user_data))
{
   GList *sel;
   sel=GTK_TREE_SELECTION(tree);
   if (sel && sel->data)
   {
      selection=gtk_object_get_data(GTK_OBJECT(sel->data),GROUP_DATA_FIELD);
      fill_data();
   }
}

void
ge_name                                (ARG_UNUSED(GtkEditable     *editable),
                                        ARG_UNUSED(gpointer         user_data))
{
   char *c;
   
   if (!selection)
      return;
   if (selection->uid==M.WorldGroup)
   {
      if (strcmp("World",gtk_entry_get_text(name)))
      {
         gtk_entry_set_text(name,"World");
			/* TODO: should probably be important warning */
         Warning("Group editor: You can't change the name of the world group!");
      }
      return;
   }
   changes=1;
   U_GroupModify(selection);

   c=gtk_entry_get_text(name);
   if (*c)
      strcpy(selection->groupname,c);
   else
      strcpy(selection->groupname,"unnamed");

   {
      GList *sel;
      GtkBin *ti;
      GtkLabel *l;
      
      sel=GTK_TREE_SELECTION(tree);
      if (sel && sel->data)
      {
         ti=GTK_BIN(sel->data);
         l=GTK_LABEL(ti->child);
         gtk_label_set_text(l,selection->groupname);
      }
   }
}

void
ge_hidden                              (ARG_UNUSED(GtkToggleButton *togglebutton),
                                        ARG_UNUSED(gpointer         user_data))
{
   if (!selection)
      return;
   changes=1;
   U_GroupModify(selection);
   selection->flags&=~GROUP_HIDDEN;
   if (gtk_toggle_button_get_active(hidden))
      selection->flags|=GROUP_HIDDEN;
}

void
ge_smart_select                        (ARG_UNUSED(GtkToggleButton *togglebutton),
                                        ARG_UNUSED(gpointer         user_data))
{
   if (!selection)
      return;
   changes=1;
   U_GroupModify(selection);
   selection->flags&=~GROUP_SMARTSELECT;
   if (gtk_toggle_button_get_active(hidden))
      selection->flags|=GROUP_SMARTSELECT;
}

void
ge_done                                (ARG_UNUSED(GtkButton       *button),
                                        ARG_UNUSED(gpointer         user_data))
{
   gtk_main_quit();
}

void
ge_delete                              (ARG_UNUSED(GtkButton       *button),
                                        ARG_UNUSED(gpointer         user_data))
{ /* TODO: recursive delete of groups with children */
   int i;
   
   if (!selection)
      return;
   if (selection->uid==M.WorldGroup)
   {
      Warning("Group delete: Can't delete world group!");
      return;
   }
   if (selection->child)
   {
      Warning("Group delete: Can't delete group with children!");
      return;
   }
   i=S_GUI_Dialog("Group delete",
      "Should the objects in this group be deleted, moved to this group's "
      "parent, or moved to the world group?",
      "Cancel","Delete","Move to parent","Move to world group",NULL);
   if (i==0)
      return;
   if (i==1)
   {
      brush_t *b,*bn;
      entity_t *e,*en;

      for (b=M.BrushHead;b;b=bn)
      {
         bn=b->Next;
         if (b->Group!=selection->uid) continue;
         B_Unlink(b);
         B_Free(b);
      }

      for (e=M.EntityHead;e;e=en)
      {
         en=e->Next;
         if (e->Group!=selection->uid) continue;
         E_Unlink(e);
         E_Free(e);
      }
   }
   else
   {
      map_uid_t dst;
      brush_t *b;
      entity_t *e;

      if (i==2)
         dst=selection->parent;
      else
         dst=M.WorldGroup;

      /* TODO? for a large group, this will waste lots of memory in undo
      objects */
      for (b=M.BrushHead;b;b=b->Next)
      {
         if (b->Group!=selection->uid) continue;
         U_BrushModify(b);
         b->Group=dst;
      }

      for (e=M.EntityHead;e;e=e->Next)
      {
         if (e->Group!=selection->uid) continue;
         U_EntityModify(e);
         e->Group=dst;
      }
   }
   Grp_Delete(selection->uid);
   selection=NULL;
   Map_Update();
   changes=1;

   gtk_widget_destroy(GTK_TREE_SELECTION(tree)->data);
}

void
ge_add_to                              (ARG_UNUSED(GtkButton       *button),
                                        ARG_UNUSED(gpointer         user_data))
{
   bsel_t *bs;
   esel_t *es;
   
   if (!selection) return;

   if (!M.sel.bsel && !M.sel.esel)
   {
      Warning("Add to group: No map objects selected!");
      return;
   }

   for (bs=M.sel.bsel;bs;bs=bs->Next)
   {
      U_BrushModify(bs->Brush);
      bs->Brush->Group=selection->uid;
   }
   for (es=M.sel.esel;es;es=es->Next)
   {
      U_EntityModify(es->Entity);
      es->Entity->Group=selection->uid;
   }
   Undo_Update();
}

void
ge_create                              (ARG_UNUSED(GtkButton       *button),
                                        ARG_UNUSED(gpointer         user_data))
{
   const char *key="Group name:";
   char name[128];
   char *n;
   map_uid_t ug;
   group_t *g;
   
   if (!selection) return;

   name[0]=0;
   n=name;
   if (!S_GUI_PopEntity("Create group",(char **)&key,&n,1))
      return;

   ug=Grp_Create(name,selection->uid);
   g=UID_Group(ug);

   {
      GtkWidget *w;
      GtkTree *t;
      GtkTreeItem *ti;

      ti=GTK_TREE_ITEM(GTK_TREE_SELECTION(tree)->data);

      w=GTK_TREE_ITEM_SUBTREE(ti);
      if (!w)
      {
         t=GTK_TREE(gtk_tree_new());
         gtk_tree_item_set_subtree(ti,GTK_WIDGET(t));
      }
      else
         t=GTK_TREE(w);
      
      w=gtk_tree_item_new_with_label(g->groupname);
      gtk_object_set_data(GTK_OBJECT(w),GROUP_DATA_FIELD,g);
      gtk_widget_show(w);
      gtk_tree_append(t,w);

      gtk_tree_item_expand(ti);
   }
}

void
ge_current                             (ARG_UNUSED(GtkButton       *button),
                                        ARG_UNUSED(gpointer         user_data))
{
   if (!selection)
      return;
   M.CurGroup=selection->uid;
   fill_cur();
}


void S_GUI_GroupEditor(void)
{
   changes=0;
   
   ge=create_group_editor();
   gtk_box_set_homogeneous(GTK_BOX(lookup_widget(ge,"dialog_action_area1")),0);
   tree=GTK_TREE(lookup_widget(ge,"tree"));
   name=GTK_ENTRY(lookup_widget(ge,"name"));
   hidden=GTK_TOGGLE_BUTTON(lookup_widget(ge,"hidden"));
   sselect=GTK_TOGGLE_BUTTON(lookup_widget(ge,"smart_select"));
   curgroup=GTK_ENTRY(lookup_widget(ge,"cur_group"));

   selection=NULL;
   fill_tree();
   fill_cur();
   
   gtk_widget_show(ge);
   gtk_tree_select_item(tree,0);
   gtk_main();
   gtk_widget_destroy(ge);
   selection=NULL;
   ge=NULL;

   if (changes)
   {
      Map_Update();
      S_R_SetUpdate(R_UPDATE_ALL);
   }
}

