#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "entclass.h"

#include "entity.h"
#include "error.h"
#include "memory.h"
#include "token.h"
#include "fe/dir.h"


/* TODO:
rework all this
*/


eclass_t eclass_dummy=
{
"dummy_class",ECLASS_POINT,
{1,1,1},
{-8,-8,-8},{8,8,8},
};


eclass_t **eclasses;
int num_eclasses;

static eclass_t **base_classes;
static int num_base_classes;

static FILE *f;
static char l[256];


static const char *str_x="x";


static void FixStr(char *str)
{
   if ((str[0]=='"') && (str[strlen(str)-1]=='"'))
   {
      strcpy(str,&str[1]);
      str[strlen(str)-1]=0;
   }
}


static int AddKeyChoice(e_key_t *key,e_key_choice_t *c)
{
   key->x.ch.choices=
      Q_realloc(key->x.ch.choices,
         (key->x.ch.num_choices+1)*sizeof(e_key_choice_t));
   if (!key->x.ch.choices)
      return 1;
   key->x.ch.choices[key->x.ch.num_choices++]=*c;
   return 0;
}

static int AddKey(eclass_t *c,e_key_t *key)
{
   int i;

   for (i=0;i<c->num_keys;i++)
      if (!strcmp(c->keys[i].name,key->name))
      {
         c->keys[i]=*key;
         return 0;
      }

   c->keys=Q_realloc(c->keys,(c->num_keys+1)*sizeof(e_key_t));
   if (!c->keys)
      return 1;
   c->keys[c->num_keys++]=*key;
   return 0;
}


static char *extra_comment;

static int AddBase(eclass_t *c,char *basename)
{
   int i,err;
   eclass_t *b;

   for (i=0;i<num_base_classes;i++)
   {
      if (!strcmp(base_classes[i]->name,basename))
         break;
   }
   if (i==num_base_classes)
   {
      for (i=0;i<num_eclasses;i++)
      {
         if (!strcmp(eclasses[i]->name,basename))
            break;
      }

      if (i==num_eclasses)
      {
			Error_Add("ParseSpecial: Unknown base class '%s' in class '%s'!",
				basename,c->name);
         return 3;
      }

      b=eclasses[i];
   }
   else
   {
      b=base_classes[i];
   }
   for (i=0;i<b->num_keys;i++)
   {
      if ((err=AddKey(c,&b->keys[i]))) return err;
   }

   if (b->comment)
   {
      if (extra_comment)
      {
         extra_comment=
            Q_realloc(extra_comment,
               strlen(extra_comment)+strlen(b->comment)+1);
         if (!extra_comment)
            return 1;
         strcat(extra_comment,b->comment);
      }
      else
      {
         extra_comment=Q_strdup(b->comment);
         if (!extra_comment)
            return 1;
      }
   }
   return 0;
}


#define GETTOKEN() \
   if (!TokenGet(1,-1)) \
   { \
      Error_Add("ParseSpecial: Parse error in class '%s'!",c->name); \
      return 3; \
   }

#define EXPECT(x) \
   GETTOKEN(); \
   if (strcmp(token,x)) \
   { \
      Error_Add("ParseSpecial: Parse error in class '%s' (expected '%s')!",c->name,x); \
      return 3; \
   }


static int GetStr(char **str,eclass_t *c)
{
   GETTOKEN();
   FixStr(token);
   *str=Q_strdup(token);
   if (!*str)
      return 1;
   return 0;
}

static int ParseSpecial(eclass_t *c)
{
   e_key_t temp;
   e_key_choice_t temp_c;
   
   int i;

   TokenBuf(c->comment,TOKEN_C|TOKEN_NAME|TOKEN_MISC|TOKEN_STRING,NULL);

   EXPECT("{");
   TokenGet(1,-1);
   while (strcmp(token,"}"))
   {
      memset(&temp,0,sizeof(e_key_t));

/* Key types */
      if (!strcmp(token,"choice"))
      {
         temp.type=EKEY_CHOICE;
         if ((i=GetStr(&temp.name,c))) return i;
         EXPECT("(");
         GETTOKEN();
         while (strcmp(token,")"))
         {
            if (strcmp(token,"("))
            {
               Error_Add("ParseSpecial: Parse error in class '%s'!",c->name);
               return 3;
            }

            if ((i=GetStr(&temp_c.val,c))) return i;
            EXPECT(",");
            if ((i=GetStr(&temp_c.name,c))) return i;
            if ((i=AddKeyChoice(&temp,&temp_c))) return i;
            EXPECT(")");
            GETTOKEN();
         }
         EXPECT(";");
         if ((i=AddKey(c,&temp))) return i;
      }
      else
      if (!strcmp(token,"list"))
      {
         temp.type=EKEY_LIST;
         if ((i=GetStr(&temp.name,c))) return i;
         EXPECT("(");
         if ((i=GetStr(&temp.x.list.name,c))) return i;
         EXPECT(")");
         EXPECT(";");
         if ((i=AddKey(c,&temp))) return i;
      }
      else
      if (!strcmp(token,"texture"))
      {
         temp.type=EKEY_TEXTURE;
         if ((i=GetStr(&temp.name,c))) return i;
         if ((i=AddKey(c,&temp))) return i;
         EXPECT(";");
      }
/* Default values */
      else
      if (!strcmp(token,"default"))
      {
         c->defs=Q_realloc(c->defs,sizeof(e_default_t)*(c->num_defs+1));
         if (!c->defs)
            return 1;

         EXPECT("(");
         if ((i=GetStr(&c->defs[c->num_defs].key,c))) return i;
         EXPECT(",");
         if ((i=GetStr(&c->defs[c->num_defs++].value,c))) return i;
         EXPECT(")");
         EXPECT(";");
      }

/* Game specific stuff */
      else
      if (!strcmp(token,"q2_color"))
      {
         temp.type=EKEY_Q2_COLOR;
         if ((i=GetStr(&temp.name,c))) return i;
         if ((i=AddKey(c,&temp))) return i;
         EXPECT(";");
      }

      else
      if (!strcmp(token,"hl_color"))
      {
         temp.type=EKEY_HL_COLOR;
         if ((i=GetStr(&temp.name,c))) return i;
         if ((i=AddKey(c,&temp))) return i;
         EXPECT(";");
      }
      else
      if (!strcmp(token,"hl_light"))
      {
         temp.type=EKEY_HL_LIGHT;
         if ((i=GetStr(&temp.name,c))) return i;
         if ((i=AddKey(c,&temp))) return i;
         EXPECT(";");
      }

/* Commands */
      else
      if (!strcmp(token,"base"))
      {
         EXPECT("(");
         GETTOKEN();
         FixStr(token);
         if ((i=AddBase(c,token))) return i;
         EXPECT(")");
         EXPECT(";");
      }
      else
      if (!strcmp(token,"model"))
      {
         EXPECT("(");
         GETTOKEN();
         FixStr(token);
/*     if (status.load_models)
            c->mdl=LoadModel(token);*/
         EXPECT(")");
         EXPECT(";");
      }
      else
      {
         Error_Add("ParseSpecial: Unknown command '%s' in class '%s'!",
            token,c->name);
         return 3;
      }
      TokenGet(1,-1);
   }
   Q_free(c->comment);

   {
      char *d;

      for (d=token_curpos;*d && *d<=32;d++) { }
      if (*d)
      {
         c->comment=Q_strdup(d);
         if (!c->comment)
            return 1;
      }
      else
         c->comment=NULL;
   }

   TokenDone();
#undef GETTOKEN
#undef EXPECT
   return 0;
}


static int ParseClass(void)
{
   eclass_t *temp;
   char *tok;
   char tempstr[256];
#define PGET_TOKEN() tok=strtok(0," \n()\t")

   tok=strtok(l," "); /* \/\*QUAKED */

   PGET_TOKEN(); /* item_shells */
   if (FindClass(tok))
      return 0;

   temp=Q_malloc(sizeof(eclass_t));
   if (!temp)
      return 1;
   memset(temp,0,sizeof(eclass_t));

   temp->name=Q_strdup(tok);
   if (!temp->name)
      return 1;

   PGET_TOKEN(); /* (0 .5 .8) */
   if (tok)
   {
      temp->color[0]=atof(tok);
      PGET_TOKEN();
      temp->color[1]=atof(tok);
      PGET_TOKEN();
      temp->color[2]=atof(tok);
   
      PGET_TOKEN(); /* (0 0 0) */
      if (!strcmp(tok,"?"))
      {
         temp->type=ECLASS_MODEL;
      }
      else
      {
         temp->type=ECLASS_POINT;
   
         temp->min.x=atof(tok);
         PGET_TOKEN();
         temp->min.y=atof(tok);
         PGET_TOKEN();
         temp->min.z=atof(tok);
   
         PGET_TOKEN(); /* (32 32 32) */
         temp->max.x=atof(tok);
         PGET_TOKEN();
         temp->max.y=atof(tok);
         PGET_TOKEN();
         temp->max.z=atof(tok);
      }
   
      temp->num_flags=0;
      temp->flags=NULL;
      for (PGET_TOKEN();tok;PGET_TOKEN()) /* big */
      {
         if (*tok=='"')
         {
            strcpy(tempstr,&tok[1]);
            while (tok[strlen(tok)-1]!='"')
            {
               PGET_TOKEN();
               strcat(tempstr," ");
               strcat(tempstr,tok);
            }
            tempstr[strlen(tempstr)-1]=0;
         }
         else
         {
            strcpy(tempstr,tok);
         }

         temp->num_flags++;
         temp->flags=Q_realloc(temp->flags,temp->num_flags*sizeof(char *));
         if (!temp->flags)
            return 1;
         if (strcmp(tempstr,str_x))
         {
            temp->flags[temp->num_flags-1]=Q_strdup(tempstr);
            if (!temp->flags[temp->num_flags-1])
               return 1;
         }
         else  /* handle "x" specially, as it occurs frequently */
            temp->flags[temp->num_flags-1]=str_x;
      }
/*      if (Game.entity.qdiff)
      {
         temp->flags=Q_realloc(temp->flags,sizeof(char *)*12);
         if (!temp->flags)
            return 1;
         while (temp->num_flags<12)
         {
            if (temp->num_flags<8)
               temp->flags[temp->num_flags]="x";
            else
            if (temp->num_flags==8)
               temp->flags[temp->num_flags]="!Easy";
            else
            if (temp->num_flags==9)
               temp->flags[temp->num_flags]="!Medium";
            else
            if (temp->num_flags==10)
               temp->flags[temp->num_flags]="!Hard";
            else
            if (temp->num_flags==11)
               temp->flags[temp->num_flags]="!Deathmatch";
            temp->num_flags++;
         }
      }*/
   }
   else
      temp->type=ECLASS_BASE;

   temp->comment=NULL;
   fgets(l,256,f);
   while (l[0] && l[strlen(l)-1]<=32) l[strlen(l)-1]=0;
   while (strncmp(&l[(strlen(l)>2)?strlen(l)-2:0],"*/",2))
   {
/* Check for new, special fields */
#define F_MDL ":model=" /* specifies which mdl/md2 file this classname uses */
      if (!strncmp(l,F_MDL,strlen(F_MDL)))
      {
/*         if (status.load_models)
         {
            strcpy(l,&l[strlen(F_MDL)]);
            while (l[strlen(l)-1]<=32) l[strlen(l)-1]=0;
            temp->mdl=LoadModel(l);
         }*/
         goto done;
      }

      strcat(l,"\n"); // add newline again
      if (temp->comment)
      {
         temp->comment=Q_realloc(temp->comment,strlen(temp->comment)+strlen(l)+1);
         if (!temp->comment)
            return 1;
         strcat(temp->comment,l);
      }
      else
      {
         temp->comment=Q_strdup(l);
         if (!temp->comment)
            return 1;
      }
done:
      if (!fgets(l,256,f))
         return 2;
      while (l[0] && l[strlen(l)-1]<=32) l[strlen(l)-1]=0;
   }

   if (temp->comment)
   {
      extra_comment=NULL;
      if (temp->comment[0]=='{')
      {
         int err;
         if ((err=ParseSpecial(temp))) return err;
      }
      if (extra_comment)
      {
         temp->comment=Q_realloc(temp->comment,strlen(extra_comment)+
            (temp->comment?strlen(temp->comment):0)+1);
         if (!temp->comment)
            return 1;
         strcat(temp->comment,extra_comment);
         Q_free(extra_comment);
      }
   }

   if (temp->type==ECLASS_BASE)
   {
      base_classes=Q_realloc(base_classes,(num_base_classes+1)*sizeof(eclass_t *));
      if (!base_classes)
         return 1;
      num_base_classes++;
      base_classes[num_base_classes-1]=temp;
   }
   else
   {
      eclasses=Q_realloc(eclasses,(num_eclasses+1)*sizeof(eclass_t *));
      if (!eclasses)
         return 1;
      num_eclasses++;
      eclasses[num_eclasses-1]=temp;
   }
#undef PGET_TOKEN
   return 0;
}

static int ParseFile(char *name)
{
   int i;
   
   f=fopen(name,"rt");
   if (!f) return 0;
   while (!feof(f))
   {
      if (!fgets(l,256,f)) break;
      while (l[0] && l[strlen(l)-1]<=32) l[strlen(l)-1]=0;

      if (!strncmp(l,"/*QUAKED",8))
      {
         i=ParseClass();
         if (i) return i;
      }
   }
   fclose(f);
   return 0;
}


eclass_t *FindClass(const char *classname)
{
   int i;

   if (!classname)
      return NULL;
   for (i=0;i<num_eclasses;i++)
      if (!strcmp(classname,eclasses[i]->name))
         return eclasses[i];
   return NULL;
}


void SetEntityDefaults(entity_t *e)
{
   eclass_t *cl;
   const char *c;
   int i;

   if (!(c=GetKeyValue(e,"classname")))
      return;

   cl=FindClass(c);
   if (!cl)
      return;

   for (i=0;i<cl->num_defs;i++)
      SetKeyValue(e,cl->defs[i].key,cl->defs[i].value);
}


static int class_sort(const void *e1, const void *e2)
{
   const eclass_t *c1,*c2;

   c1=*(eclass_t * const *)e1;
   c2=*(eclass_t * const *)e2;
   return strcmp(c1->name,c2->name);
}


#include "qfunc.h"


static void C_entity_set_load(qatom_t l)
{
static const char *errors[]={
"Out of memory!",
"Parse error!"
};

	char *filename=QSTRING(QLIST(&l,0));

	int error;

	error=ParseFile(filename);
	if (error)
	{
		if (error>=0 && error<=2)
			Error_Add("entity_set.load: %s",errors[error]);
		else
			Error_Add("entity_set.load: %s","Error loading entity classes!");
		Error_Display("entity_set.load: while loading '%s'",filename);
	}

	qsort(eclasses,num_eclasses,sizeof(eclass_t *),class_sort);
}
static func_t F_entity_set_load=
{"entity_set.load",FUNC_VOID,{fva:C_entity_set_load},NULL,"c",0};


static void FreeClass(eclass_t *c)
{
   int j;
   
   if (!c)
      return;
      
   if (c->name)
      Q_free(c->name);
   if (c->comment)
      Q_free(c->comment);

   if (c->flags)
   {
      for (j=0;j<c->num_flags;j++)
         if (c->flags[j] && c->flags[j]!=str_x)
            Q_free(c->flags[j]);
      Q_free(c->flags);
   }

   if (c->keys)
   { /* we can't free the keys, other classes might have pointers to them */
      Q_free(c->keys);
   }

   if (c->defs)
   {
      for (j=0;j<c->num_defs;j++)
      {
         if (c->defs[j].key)
            Q_free(c->defs[j].key);
         if (c->defs[j].value)
            Q_free(c->defs[j].value);
      }
      Q_free(c->defs);
   }

   Q_free(c);
}

void EClass_Reset(void)
{
   int i;
   eclass_t **e;

   if (eclasses)
   {
      for (e=eclasses,i=num_eclasses;i;i--,e++)
         FreeClass(*e);
      Q_free(eclasses);
      eclasses=NULL;
      num_eclasses=0;
   }

   if (base_classes)
   {
      for (e=base_classes,i=num_base_classes;i;i--,e++)
         FreeClass(*e);
      Q_free(base_classes);
      base_classes=NULL;
      num_base_classes=0;
   }
}


#include "funcs.h"

func_t *funcs_eclass[]={&F_entity_set_load,NULL};

