#ifndef HOOKS_H
#define HOOKS_H

/*
Standard hooks
*/

extern hook_t
/*
plugin.remove(struct plugin_s *p) (HOOK_ALL)
  Called when a plugin is deactivated. p is the plugin id, so systems that
pointers to stuff in plugins should have a function here that removes the
items from that plugin when it is deactivated.
*/
   *hk_plugin_remove,

/*
cmd.start(cmd_reg_t *c) (HOOK_HANDLE)
cmd.end(cmd_reg_t *c) (HOOK_ALL)
  Called before and after commands are executed. Not called for recursive
commands. If cmd.start is handled, the command is not run (nor is cmd.end).
cmd.start is called before undo is started for the command, and cmd.end
is called before undo is completed for the command. Thus cmd.end can modify
stuff, but cmd.start can't.
*/
   *hk_cmd_start,
   *hk_cmd_end,

/*
map.new(void) HOOK_ALL
  Called when new map is called, after an empty valid map has been
constructed.

map.unload(void) HOOK_ALL
  Called right before a map is unloaded.

map.load(void) HOOK_ALL
  Called after a map has been loaded.

map.save(const char *name) HOOK_ALL
  Called just before a map is saved.

map.saved(const char *name) HOOK_ALL
  Called just after a map has been saved.
*/
   *hk_map_new,
   *hk_map_unload,
   *hk_map_load,
   *hk_map_save,
   *hk_map_saved;


#endif

