#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "error.h"
#include "err_catch.h"
#include "qatom.h"
#include "ql.h"
#include "qsymbol.h"



static char CharForType(int type)
{
	switch (type)
	{
	case QAT_SYMBOL: return 's';
	case QAT_STRING: return 'c';
	case QAT_FUNC:   return 'f';
	case QAT_MENU:   return 'm';
	case QAT_INT:    return 'i';
	default: ERROR("Unknown atom type %05x\n",type);
	}
}

static int Match(qatom_t a,char *mb,int type)
{
	if (!type)
	{
		if (*mb=='.') return 1;
		if (CharForType(a.type)==*mb)
			return 1;
		if (a.type==QAT_SYMBOL && !a.x.symbol && *mb=='n')
			return 1;
		return 0;
	}
	if (strchr(mb,CharForType(a.type)))
		return 1;
	if (a.type==QAT_SYMBOL && !a.x.symbol && strchr(mb,'n'))
		return 1;
	return 0;
}

static char *WriteArg(char *b,qatom_t a)
{
	if (QAT_ISLIST(a.type))
	{
		int i;
		*b++='<';
		for (i=0;i<a.type;i++)
			b=WriteArg(b,a.x.l[i]);
		*b++='>';
	}
	else
	switch (a.type)
	{
	case QAT_SYMBOL: if (!a.x.symbol) *b++='n'; else *b++='s'; break;
	case QAT_STRING: *b++='c'; break;
	case QAT_MENU:   *b++='m'; break;
	case QAT_FUNC:   *b++='f'; break;
	case QAT_INT:    *b++='i'; break;
	default: ABORT("Unknown atom type %05x\n",a.type);
	}
	return b;
}

struct qatom_s Func_Call(func_t *f,struct qatom_s arglist)
{
	qatom_t ret;
	ASSERT(QAT_ISLIST(arglist.type));

	if (f->type==FUNC_QL)
	{
		qatom_t *l=f->data;
		int i;
		if (arglist.type)
			Error("'%s': Function doesn't take any arguments",QL_SymbolName(f->symbol));
		for (i=1;i<l->type;i++)
			QA_Free(QL_Eval(l->x.l[i]));
		ret=qa_undef;
		return ret;
	}

	if (!f->args)
	{
		if (arglist.type)
			Error("'%s': Function doesn't take any arguments",QL_SymbolName(f->symbol));
	}
	else
	{
		char abuf[128];
		char *b;
		int i;
		for (b=abuf,i=0;i<arglist.type;i++)
			b=WriteArg(b,arglist.x.l[i]);
		*b=0;

	{ /* TODO: cleanup */
		char mbuf[64];
		const char *c;
		char *d;
		int i;
		int type;

		i=0;
		c=f->args;
		while (*c)
		{
			if (*c=='[')
			{
				c++;
				for (d=mbuf;*c && *c!=']';*d++=*c++);
				ASSERT(*c);
				c++;
				*d=0;
				type=1;
			}
			else
			if (*c=='<')
				ABORT("TODO: unimplemented"); /* do we really need this? */
			else
				mbuf[0]=*c++,mbuf[1]=type=0;
			if (*c=='*')
			{
				c++;
				for (;i<arglist.type && Match(arglist.x.l[i],mbuf,type);i++);
			}
			else
			if (*c=='+')
			{
				c++;
				if (i==arglist.type)
					Error("'%s': Too few arguments (%s/%s)",
						QL_SymbolName(f->symbol),abuf,f->args);
				if (!Match(arglist.x.l[i],mbuf,type))
					Error("'%s': Argument %i type mismatch (%s/%s)",
						QL_SymbolName(f->symbol),i,abuf,f->args);
				for (i++;i<arglist.type && Match(arglist.x.l[i],mbuf,type);i++);
			}
			else
			if (*c=='?')
			{
				c++;
				if (i<arglist.type)
				{
					if (!Match(arglist.x.l[i],mbuf,type))
						Error("'%s': Argument %i type mismatch (%s/%s)",
							QL_SymbolName(f->symbol),i,abuf,f->args);
					i++;
				}
			}
			else
			{
				if (i==arglist.type)
					Error("'%s': Too few arguments (%s/%s)",
						QL_SymbolName(f->symbol),abuf,f->args);
				if (!Match(arglist.x.l[i],mbuf,type))
					Error("'%s': Argument %i type mismatch (%s/%s)",
						QL_SymbolName(f->symbol),i,abuf,f->args);
				i++;
			}
		}
	}
	}

/* TODO: check hook */

	ERR_TRY
	{
		switch (f->type)
		{
		case FUNC_VOID:
			f->f.fvd(arglist,f->data);
			ret=qa_T;
			break;
		case FUNC_INT:
			switch (f->f.fid(arglist,f->data))
			{
			case -1: ret=qa_undef; break;
			default:
			case  0: ret=qa_nil; break;
			case  1: ret=qa_T; break;
			}
			break;
		case FUNC_ATOM:
			f->f.fad(&ret,arglist,f->data);
			break;
		}
	}
	ERR_CATCH
		Error("while in '%s':",QL_SymbolName(f->symbol));
	ERR_END
	return ret;
}


void Func_RegisterOne(struct plugin_s *p,func_t *f)
{
	ASSERT(f->name);
	f->symbol=QL_SymbolDef(f->name);
	f->plugin=p;
	QL_SymbolSetFunc(f->symbol,f);
}

void Func_Register(struct plugin_s *p,func_t **f,int size)
{
	int i;
	for (i=0,size/=sizeof(func_t *);i<size;i++)
		Func_RegisterOne(p,f[i]);
}

void Func_RegisterP(struct plugin_s *p,func_t **f)
{
	for (;*f;f++)
		Func_RegisterOne(p,*f);
}


#include "err_catch.h"
#include "undo.h"
#include "fe/render.h"

void Func_RunCmd(func_t *f)
{ /* TODO: do more properly? */
	ERR_TRY
	{
		volatile qatom_t r,a;
		a.type=0;
		r=Func_Call(f,a);
		QA_Free(r);

		if (Error_Have())
			Error_Display("Error while running command '%s'",QL_SymbolName(f->symbol));
	}
	ERR_CATCH
	{
		Error_Display("Error while running command '%s'",QL_SymbolName(f->symbol));
	}
	ERR_END
	Undo_Done();
	S_R_HandleUpdate();
}

