#ifndef QFUNC_H
#define QFUNC_H

#include "qatom.h"


typedef struct
{
	const char *desc;
	const char *help;
	int undo; /* TODO: use this? */
#define FUNC_UNDO_NORMAL    0
#define FUNC_UNDO_NOCHANGE  1 /* The function doesn't change any map data */
#define FUNC_UNDO_CLEAR     2 /* Clear all undo information */
} func_attr_t;


struct qatom_s;
typedef struct func_s
{
	const char *name;
	int type;
#define FUNC_VOID     1
#define FUNC_INT      2
#define FUNC_ATOM     3
#define FUNC_QL       4
	union
	{
		void (*fad)(qatom_t *ret,qatom_t args,void *data);
		int (*fid)(qatom_t args,void *data);  /* -1=qa_undef, 0=qa_nil, 1=qa_T */
		void (*fvd)(qatom_t args,void *data); /* returns qa_T */

		void (*faa)(qatom_t *ret,qatom_t args);
		int (*fia)(qatom_t args);
		void (*fva)(qatom_t args);

		void (*fa)(qatom_t *ret);
		int (*fi)(void);
		void (*fv)(void);
	} f;
	void *data;

	const char *args;

	int flags;
/* Arguments should not be evaluated before calling */
#define FUNC_NOT_EVAL_ARGS   0x01
/* Not a user visible function (shouldn't be bound to keys, etc.) */
#define FUNC_NOT_USER        0x02

	func_attr_t attr;
/* The following fields are set when the function is registered. */
	int symbol;
	struct hook_s *h;
	struct plugin_s *plugin;
} func_t;


struct qatom_s Func_Call(func_t *f,struct qatom_s arglist);

void Func_RunCmd(func_t *f);

void Func_RegisterOne(struct plugin_s *p,func_t *f);
void Func_Register(struct plugin_s *p,func_t **f,int size);
void Func_RegisterP(struct plugin_s *p,func_t **f);

#endif

