/*
quest.h file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#ifndef QUEST_H
#define QUEST_H


typedef struct
{
   int valid;

   struct brush_s *BrushHead;   /* head of linked list of brushes */
   struct brush_s *VBrushHead;  /* head of linked list of visible brushes */
   int num_brushes;
   
   entity_t       *EntityHead;  /* head of linked list of entities */
   entity_t       *VEntityHead; /* head of linked list of visible entities */
   int num_entities;
   
   group_t        *GroupHead;   /* head of linked list of groups */
   int num_groups;
   
   map_uid_t       WorldSpawn;  /* points to the worldspawn entity */
   map_uid_t       WorldGroup;  /* root group in the group hierarchy */
   
   map_uid_t       CurGroup;    /* new objects should be placed in this
                                   group */

   /* The primary selection */
   sel_t sel;
   /* Should have these flags on entry and exit from commands */
#define SELECTION_PRIMARY (SEL_DRAW|SEL_SELECTED)

   sel_t **rsel;
   int num_rsel;

   char mapfilename[256];

   struct undo_s *undo;
   int modified;
} map_t;


extern map_t M;


/*
Rendering flags, kept in 'render.flags'.
*/                                    /* requires */
#define RENDER_BRUSH            0x002
#define RENDER_BRUSH_FOCAL      0x004 /* BRUSH */
#define RENDER_BRUSH_VERT       0x008 /* BRUSH_FOCAL */
#define RENDER_FACE_FOCAL       0x010 /* BRUSH */
#define RENDER_FACE_VERT        0x020 /* FACE_FOCAL */
#define RENDER_ENTITY           0x100
#define RENDER_ENTITY_FOCAL     0x200 /* ENTITY */
#define RENDER_ENTITY_LINK      0x400 /* ENTITY_FOCAL */


#endif

