#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "sel.h"

#include "brush.h"
#include "entity.h"
#include "error.h"
#include "memory.h"
#include "quest.h"
#include "fe/3d.h"
#include "fe/render.h"


/* TODO: should these really call S_R_SetUpdate()? */


vsel_t *Sel_FindVert(const sel_t *s,struct brush_s *b,int num)
{
   vsel_t *v;

   for (v=s->vsel; v; v=v->Next)
      if (v->Brush==b && v->num==num)
         return v;
   return NULL;
}

void Sel_FAddVert(sel_t *s,struct brush_s *b,int num)
{
   vsel_t *v;

   if (s->vsel!=NULL)
   {
      v = (vsel_t *)Q_malloc(sizeof(vsel_t));
      if (v==NULL)
      {
         ERROR("Select Verts: Could not allocate select list");
         return;
      }
      v->Brush=b;
      v->num=num;
      v->Last = NULL;
      v->Next = s->vsel;
      s->vsel->Last=v;
      s->vsel = v;
      s->num_vselected++;
   }
   else
   {
      s->vsel = (vsel_t *)Q_malloc(sizeof(vsel_t));
      if (s->vsel==NULL)
      {
         ERROR("Select Verts: Could not allocate select list");
         return;
      }
      s->vsel->Brush=b;
      s->vsel->num=num;
      s->vsel->Last = NULL;
      s->vsel->Next = NULL;
      s->num_vselected=1;
   }
   S_R_SetUpdate(R_UPDATE_SEL);
}

void Sel_AddVert(sel_t *s,struct brush_s *b,int num)
{
   if (Sel_FindVert(s,b,num))
      return;
   Sel_FAddVert(s,b,num);
}

void Sel_RemoveVert(sel_t *s,vsel_t *vs)
{
	if (vs->Last!=NULL) vs->Last->Next = vs->Next;
	if (vs->Next!=NULL) vs->Next->Last = vs->Last;
	if (vs==M.sel.vsel) M.sel.vsel=vs->Next;
	Q_free(vs);
	s->num_vselected--;
	S_R_SetUpdate(R_UPDATE_SEL);
}


fsel_t *Sel_FindFace(const sel_t *s,struct brush_s *b,int num)
{
   fsel_t *f;

   for (f=s->fsel; f; f=f->Next)
      if (f->Brush==b && f->num==num)
         return f;
   return NULL;
}

void Sel_FAddFace(sel_t *s,struct brush_s *b,int num)
{
   fsel_t *f;

   if (s->fsel!=NULL)
   {
      f = (fsel_t *)Q_malloc(sizeof(fsel_t));
      if (f==NULL)
      {
         ERROR("Select Face: Could not allocate select list");
         return;
      }
      f->Brush=b;
      f->num=num;
      f->Last = NULL;
      f->Next = s->fsel;
      s->fsel->Last=f;
      s->fsel = f;
      s->num_fselected++;
   }
   else
   {
      s->fsel = (fsel_t *)Q_malloc(sizeof(fsel_t));
      if (s->fsel==NULL)
      {
         ERROR("Select Face: Could not allocate select list");
         return;
      }
      s->fsel->Brush=b;
      s->fsel->num=num;
      s->fsel->Last = NULL;
      s->fsel->Next = NULL;
      s->num_fselected=1;
   }
	S_R_SetUpdate(R_UPDATE_SEL);
}

void Sel_AddFace(sel_t *s,struct brush_s *b,int num)
{
   if (Sel_FindFace(s,b,num))
      return;
   Sel_FAddFace(s,b,num);
}

void Sel_RemoveFace(sel_t *s,fsel_t *fs)
{ /* TODO: figure out more exactly which selected vertices to deselect */
   vsel_t *v,*vn;

	for (v=s->vsel; v; v=vn)
   {
      vn=v->Next;
		if (v->Brush==fs->Brush)
         Sel_RemoveVert(s,v);
   }
   
	if (fs->Last!=NULL) (fs->Last)->Next = fs->Next;
	if (fs->Next!=NULL) (fs->Next)->Last = fs->Last;
	if (fs==s->fsel) s->fsel=fs->Next;
	Q_free(fs);
	s->num_fselected--;
	S_R_SetUpdate(R_UPDATE_SEL);
}


bsel_t *Sel_FindBrush(const sel_t *s,struct brush_s *b)
{
   bsel_t *bs;

   for (bs=s->bsel; bs; bs=bs->Next)
      if (bs->Brush == b)
         return bs;
   return NULL;
}

void Sel_FAddBrush(sel_t *s,struct brush_s *sel_brush)
{
	bsel_t *b;

   if (Sel_FindBrush(s,sel_brush))
      return;

	if (s->bsel!=NULL)
	{
		b = (bsel_t *)Q_malloc(sizeof(bsel_t));
		if (b==NULL)
		{
			ERROR("Select Brush: Could not allocate select list");
			return;
		}
		b->Brush = sel_brush;
		b->Last = NULL;
		b->Next = s->bsel;
		s->bsel->Last=b;
		s->bsel = b;

		s->num_bselected++;
	}
	else
	{
		s->bsel = (bsel_t *)Q_malloc(sizeof(bsel_t));
		if (s->bsel==NULL)
		{
			ERROR("Select Brush: Could not allocate select list");
			return;
		}
		s->bsel->Brush = sel_brush;
		s->bsel->Last = NULL;
		s->bsel->Next = NULL;
		s->num_bselected=1;
	}
	S_R_SetUpdate(R_UPDATE_SEL);
}

void Sel_AddBrush(sel_t *s,struct brush_s *sel_brush)
{
   if (Sel_FindBrush(s,sel_brush))
      return;
   Sel_FAddBrush(s,sel_brush);
}
      
void Sel_RemoveBrush(sel_t *s,bsel_t *b)
{
	vsel_t *v,*vn;

	for (v=s->vsel; v; v=vn)
   {
      vn=v->Next;
		if (v->Brush==b->Brush)
         Sel_RemoveVert(s,v);
   }
   
	if (b->Last!=NULL) (b->Last)->Next = b->Next;
	if (b->Next!=NULL) (b->Next)->Last = b->Last;
	if (b==s->bsel) s->bsel=b->Next;
	Q_free(b);
	s->num_bselected--;
	S_R_SetUpdate(R_UPDATE_SEL);
}


esel_t *Sel_FindEnt(const sel_t *s,entity_t *e)
{
   esel_t *er;

   for (er=s->esel;er;er=er->Next)
      if (er->Entity==e)
         return er;
   return NULL;
}

void Sel_FAddEnt(sel_t *s,entity_t *e)
{
   esel_t *er;

   if (Sel_FindEnt(s,e))
      return;

   er=Q_malloc(sizeof(esel_t));
   if (!er)
   {
      ERROR("AddSelEnt: Out of memory!");
      return;
   }
   er->Entity=e;
   er->Next=s->esel;
   er->Last=NULL;
   if (s->esel)
      s->esel->Last=er;
   s->esel=er;
   s->num_eselected++;
	S_R_SetUpdate(R_UPDATE_SEL);
}

void Sel_AddEnt(sel_t *s,entity_t *e)
{
   if (Sel_FindEnt(s,e))
      return;
   Sel_FAddEnt(s,e);
}

void Sel_RemoveEnt(sel_t *s,esel_t *es)
{
   if (es->Next) es->Next->Last=es->Last;
   if (es->Last) es->Last->Next=es->Next;
   if (es==s->esel) s->esel=es->Next;
   Q_free(es);
   s->num_eselected--;
	S_R_SetUpdate(R_UPDATE_SEL);
}


void Sel_AddDisp(sel_t *s)
{
   sel_t **r;
   r=Q_realloc(M.rsel,sizeof(sel_t *)*(M.num_rsel+1));
   if (!r)
   {
      ERROR("Sel_AddDisp: Out of memory!");
      return;
   }
   M.rsel=r;
   r[M.num_rsel++]=s;
	S_R_SetUpdate(R_UPDATE_SEL);
}

void Sel_RemoveDisp(sel_t *s)
{
   int i,j;
   
   for (i=0;i<M.num_rsel;i++)
      if (M.rsel[i]==s)
         break;
   ASSERT(i!=M.num_rsel);

   for (j=i;j<M.num_rsel-1;j++)
      M.rsel[j]=M.rsel[j+1];
   M.num_rsel--;

   if (M.num_rsel)
   {
      sel_t **r;
      r=Q_realloc(M.rsel,sizeof(sel_t *)*M.num_rsel);
      if (r) M.rsel=r;
   }
   else
   {
      Q_free(M.rsel);
      M.rsel=NULL;
   }
	S_R_SetUpdate(R_UPDATE_SEL);
}


void Sel_Update(sel_t *s)
{
   bsel_t *bs;
   esel_t *es;
   fsel_t *fs;

   for (bs=s->bsel;bs;bs=bs->Next)
      B_Update(bs->Brush);
   for (fs=s->fsel;fs;fs=fs->Next)
      B_Update(fs->Brush);
   for (es=s->esel;es;es=es->Next)
      E_Update(es->Entity);
	S_R_SetUpdate(R_UPDATE_SEL);
}


void Sel_Clear(sel_t *s)
{
   bsel_t *b,*b2;
   vsel_t *v,*v2;
   esel_t *e,*e2;
   fsel_t *f,*f2;

   for (v=s->vsel;v;v=v2)
   {
      v2=v->Next;
      Q_free(v);
   }
   for (b=s->bsel;b;b=b2)
   {
      b2=b->Next;
      Q_free(b);
   }
   for (f=s->fsel;f;f=f2)
   {
      f2=f->Next;
      Q_free(f);
   }
   for (e=s->esel;e;e=e2)
   {
      e2=e->Next;
      Q_free(e);
   }
}

void Sel_Free(sel_t *s)
{
   bsel_t *b,*b2;
   vsel_t *v,*v2;
   esel_t *e,*e2;
   fsel_t *f,*f2;

   for (v=s->vsel;v;v=v2)
   {
      v2=v->Next;
      Q_free(v);
   }
   for (f=s->fsel;f;f=f2)
   {
      f2=f->Next;
      Q_free(f);
   }
   for (b=s->bsel;b;b=b2)
   {
      B_Free(b->Brush);
      b2=b->Next;
      Q_free(b);
   }
   for (e=s->esel;e;e=e2)
   {
      E_Free(e->Entity);
      e2=e->Next;
      Q_free(e);
   }
}

