#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "texalign.h"

#include "brush.h"
#include "error.h"
#include "memory.h"



static float baseaxis[18][3] =
{
{ 0, 0, 1},{ 1, 0, 0},{ 0,-1, 0}, /* floor */
{ 0, 0,-1},{ 1, 0, 0},{ 0,-1, 0}, /* ceiling */
{ 1, 0, 0},{ 0, 1, 0},{ 0, 0,-1}, /* west wall */
{-1, 0, 0},{ 0, 1, 0},{ 0, 0,-1}, /* east wall */
{ 0, 1, 0},{ 1, 0, 0},{ 0, 0,-1}, /* south wall */
{ 0,-1, 0},{ 1, 0, 0},{ 0, 0,-1}  /* north wall */
};
static int flipped[6]={0,1,0,1,1,0};

int TextureAxisFromPlane(vec3_t pnorm,vec3_t *norm,vec3_t *xv,vec3_t *yv)
{
	int	bestaxis;
	float	dot,best;
	int	i;
	
	best = 0;
	bestaxis = 0;

	for (i=0 ; i<6 ; i++)
	{
		dot = (pnorm.x*baseaxis[i*3][0])+
            (pnorm.y*baseaxis[i*3][1])+
            (pnorm.z*baseaxis[i*3][2]);
		if (dot-best>0.01)
		{
			best = dot;
			bestaxis = i;
		}
	}
   if (norm)
   {
      norm->x=baseaxis[bestaxis*3][0];
      norm->y=baseaxis[bestaxis*3][1];
      norm->z=baseaxis[bestaxis*3][2];
   }
   if (xv)
   {
      xv->x  =baseaxis[bestaxis*3+1][0];
      xv->y  =baseaxis[bestaxis*3+1][1];
      xv->z  =baseaxis[bestaxis*3+1][2];
   }
   if (yv)
   {
      yv->x  =baseaxis[bestaxis*3+2][0];
      yv->y  =baseaxis[bestaxis*3+2][1];
      yv->z  =baseaxis[bestaxis*3+2][2];
   }

   return flipped[bestaxis];
}


void MakeVectors(scalar_t vecs[2][3],brush_normal_align_t *t)
{
	int	sv, tv;
	double ang, sinv, cosv;
	double ns, nt;

   double rotate;
   double scale[2];
   int i,j;

   rotate=t->rotate;

   scale[0]=t->scale.x;
   scale[1]=t->scale.y;
   if (fabs(scale[0])<0.001) scale[0]=1;
   if (fabs(scale[1])<0.001) scale[1]=1;
   scale[0]=1/scale[0];
   scale[1]=1/scale[1];

   /* avoid round-off errors in common cases */
	if (rotate == 0)
   {
      sinv = 0;
      cosv = 1;
   }
	else
	if (rotate == 90)
	{
	   sinv = 1;
	   cosv = 0;
   }
	else
	if (rotate == 180)
	{
	   sinv = 0;
	   cosv = -1;
   }
	else
	if (rotate == 270)
   {
      sinv = -1;
      cosv = 0;
   }
	else
	{	
		ang = rotate / 180.0 * PI;
		sinv = sin(ang);
		cosv = cos(ang);
	}

	if (vecs[0][0])
		sv = 0;
	else
   if (vecs[0][1])
		sv = 1;
	else
		sv = 2;
				
	if (vecs[1][0])
		tv = 0;
	else
   if (vecs[1][1])
		tv = 1;
	else
		tv = 2;

	for (i=0 ; i<2 ; i++)
	{
		ns = cosv * vecs[i][sv] - sinv * vecs[i][tv];
		nt = sinv * vecs[i][sv] + cosv * vecs[i][tv];
		vecs[i][sv] = ns;
		vecs[i][tv] = nt;
	}

   for (i=0 ; i<2 ; i++)
		for (j=0 ; j<3 ; j++)
			vecs[i][j] *= scale[i];
}


void CalcTexv(brush_t *b,int facenum)
{
   face_t *f;
   brush_normal_align_t *a;
   scalar_t vecs[2][3];
   int i;
   vec3_t *v;

   if (!b->bt->flags&BRUSH_F_NORMAL) return;
   
   f=b->faces[facenum];
   a=&f->a;

   TextureAxisFromPlane(f->normal,NULL,(vec3_t *)(vecs[0]),(vec3_t *)(vecs[1]));
   MakeVectors(vecs,a);

	/* TODO: handle f->num_verts==0 */
   f->texv=Q_realloc(f->texv,sizeof(vec2_t)*f->num_verts);
   if (!f->texv)
      ABORT("UpdateTexv: Out of memory!");

   for (i=0;i<f->num_verts;i++)
   {
      v=&b->verts[f->verts[i]];
      f->texv[i].x=
         v->x*vecs[0][0]+
         v->y*vecs[0][1]+
         v->z*vecs[0][2]+
         a->shift.x;
      f->texv[i].y=
         v->x*vecs[1][0]+
         v->y*vecs[1][1]+
         v->z*vecs[1][2]+
         a->shift.y;
   }
}

