#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "transform.h"

#include "brush.h"
#include "entity.h"
#include "error.h"
#include "geom.h"
#include "map.h"
#include "matrix.h"
#include "quest.h"
#include "sel.h"
#include "status.h"
#include "texlock.h"
#include "undo.h"
#include "fe/message.h"


int Trans_Setup(const sel_t *src,sel_t *dst)
{
   bsel_t *bs;
   fsel_t *fs;
   vsel_t *vs;
   esel_t *es;

   brush_t *b;
   face_t *f;

   int i;
   

   if (!(src->bsel || src->esel || src->fsel || src->vsel))
   {
      S_NewMessage("Nothing selected!");
      return 0;
   }


   memset(dst,0,sizeof(sel_t));

   for (bs=src->bsel;bs;bs=bs->Next)
   {
      if (bs->Brush->disp)
         continue;
         
      Sel_AddBrush(dst,bs->Brush);
      bs->Brush->disp=B_Duplicate(NULL,bs->Brush);
      if (!bs->Brush->disp)
         ABORT("Trans_Setup: Out of memory!");
   }

   for (fs=src->fsel;fs;fs=fs->Next)
   {
      if (fs->Brush->disp)
         continue;
      Sel_AddBrush(dst,fs->Brush);
      fs->Brush->disp=B_Duplicate(NULL,fs->Brush);
      if (!fs->Brush->disp)
         ABORT("Trans_Setup: Out of memory!");
   }

   if (src->vsel)
   {
      for (vs=src->vsel;vs;vs=vs->Next)
         Sel_FAddVert(dst,vs->Brush,vs->num);
   }
   else
   {
      for (bs=src->bsel;bs;bs=bs->Next)
      {
         b=bs->Brush;
         for (i=0;i<b->num_verts;i++)
            Sel_FAddVert(dst,b,i);
      }
      for (fs=src->fsel;fs;fs=fs->Next)
      {
         b=fs->Brush;
         f=b->faces[fs->num];
         for (i=0;i<f->num_verts;i++)
            Sel_AddVert(dst,b,f->verts[i]);
      }
   }

   for (es=src->esel;es;es=es->Next)
   {
      Sel_AddEnt(dst,es->Entity);
      es->Entity->disp=E_Duplicate(NULL,es->Entity);
      if (!es->Entity->disp)
         ABORT("Trans_Setup: Out of memory!");
      for (b=M.VBrushHead;b;b=b->VNext)
      {
         if (b->Entity!=es->Entity->uid) continue;

         if (!b->disp)
         {
            Sel_AddBrush(dst,b);
            b->disp=B_Duplicate(NULL,b);
            if (!b->disp)
               ABORT("Trans_Setup: Out of memory!");
            for (i=0;i<b->num_verts;i++)
               Sel_FAddVert(dst,b,i);
         }
         else
         {
            if (src->vsel)
            {
               for (i=0;i<b->num_verts;i++)
               {
                  if (!Sel_FindVert(src,b,i))
                     Sel_FAddVert(dst,b,i);
               }
            }
         }
      }
   }

   return 1;
}

void Trans_Done(sel_t *s,int restore)
{
   bsel_t *bs;
   brush_t *b;
   
   esel_t *es;
   entity_t *e;

   for (bs=s->bsel;bs;bs=bs->Next)
   {
      b=bs->Brush;
      ASSERT(b->disp);
      B_Free(b->disp);
      b->disp=NULL;
   }
   for (es=s->esel;es;es=es->Next)
   {
      e=es->Entity;
      ASSERT(e->disp);
      E_Free(e->disp);
      e->disp=NULL;
   }

   Sel_Clear(s);
   
   if (restore)
   {
      Undo_Restore();
/*      vsel_t *vs;
      face_t *f,*f2;
      int i,j;
      vec3_t *v1,*v2;
      
      entity_t *e;
      esel_t *es;
      const char *c;

      for (vs=s->vsel;vs;vs=vs->Next)
      {
         v2=&vs->Brush->verts[vs->num];
         v1=&vs->Brush->disp->verts[vs->num];

         *v2=*v1;
      }
      
      for (bs=s->bsel;bs;bs=bs->Next)
      {
         if (!(bs->Brush->bt->flags&BRUSH_F_NORMAL)) continue;
         for (f2=bs->Brush->faces,f=bs->Brush->disp->faces,i=bs->Brush->num_faces;
              i;
              i--,f++,f2++)
            for (j=0;j<f->num_verts;j++)
               f2->verts[j]=f->verts[j];
      }

      for (es=s->esel;es;es=es->Next)
      {
         e=es->Entity;
         c=GetKeyValue(e->disp,"origin");
         if (!c) continue;
         SetKeyValue(e,"origin",c);
      }*/
   }
}

void Trans_Linear(matrix_t m,sel_t *s)
{
   bsel_t *bs;
   esel_t *es;
   vec3_t *v1,*v2;
   vsel_t *vs;

   for (bs=s->bsel;bs;bs=bs->Next)
      U_BrushModify(bs->Brush);
   for (es=s->esel;es;es=es->Next)
      U_EntityModify(es->Entity);

   for (vs=s->vsel;vs;vs=vs->Next)
   {
      v2=&vs->Brush->verts[vs->num];
      v1=&vs->Brush->disp->verts[vs->num];

      v2->x=
         m[0][0]*v1->x+
         m[0][1]*v1->y+
         m[0][2]*v1->z+
         m[0][3];
      v2->y=
         m[1][0]*v1->x+
         m[1][1]*v1->y+
         m[1][2]*v1->z+
         m[1][3];
      v2->z=
         m[2][0]*v1->x+
         m[2][1]*v1->y+
         m[2][2]*v1->z+
         m[2][3];
   }

   if (DetMatrix(m)<0)
   { /* TODO: flip vertex order in bezier patches (or something) */
      face_t **f,**f2;
      int i,j;
      
      for (bs=s->bsel;bs;bs=bs->Next)
      {
         if (!(bs->Brush->bt->flags&BRUSH_F_NORMAL)) continue;

         for (f2=bs->Brush->faces,f=bs->Brush->disp->faces,i=bs->Brush->num_faces;
              i;
              i--,f++,f2++)
            for (j=0;j<(*f)->num_verts;j++)
               (*f2)->verts[(*f)->num_verts-j-1]=(*f)->verts[j];
      }
   }
   else
   {
      face_t **f,**f2;
      int i,j;
      
      for (bs=s->bsel;bs;bs=bs->Next)
      {
         if (!(bs->Brush->bt->flags&BRUSH_F_NORMAL)) continue;
         for (f2=bs->Brush->faces,f=bs->Brush->disp->faces,i=bs->Brush->num_faces;
              i;
              i--,f++,f2++)
            for (j=0;j<(*f)->num_verts;j++)
               (*f2)->verts[j]=(*f)->verts[j];
      }
   }

   if (c_texture_locking)
   {
      for (bs=s->bsel;bs;bs=bs->Next)
      {
         RecalcNormals(bs->Brush);
         TexLock(bs->Brush,bs->Brush->disp);
      }
   }

   {
      const char *c;
      vec3_t v,v2;
      entity_t *e;
      char buf[32];

      /* TODO: change angle/mangle/angles key as well */
      for (es=s->esel;es;es=es->Next)
      {
         e=es->Entity;
         c=GetKeyValue(e->disp,"origin");
         if (!c) continue;

         v.x=v.y=v.z=0;
         sscanf(c,"%f %f %f",&v.x,&v.y,&v.z);
         v2.x=
            m[0][0]*v.x+
            m[0][1]*v.y+
            m[0][2]*v.z+
            m[0][3];
         v2.y=
            m[1][0]*v.x+
            m[1][1]*v.y+
            m[1][2]*v.z+
            m[1][3];
         v2.z=
            m[2][0]*v.x+
            m[2][1]*v.y+
            m[2][2]*v.z+
            m[2][3];

         snprintf(buf,sizeof(buf),"%0.0f %0.0f %0.0f",v2.x,v2.y,v2.z);
         SetKeyValue(e,"origin",buf);
      }
   }
}

void Trans_Move(vec3_t delta,sel_t *s)
{
   bsel_t *bs;
   esel_t *es;
   vec3_t *v1,*v2;
   vsel_t *vs;

   for (bs=s->bsel;bs;bs=bs->Next)
      U_BrushModify(bs->Brush);
   for (es=s->esel;es;es=es->Next)
      U_EntityModify(es->Entity);

   for (vs=s->vsel;vs;vs=vs->Next)
   {
      v2=&vs->Brush->verts[vs->num];
      v1=&vs->Brush->disp->verts[vs->num];

      v2->x=v1->x+delta.x;
      v2->y=v1->y+delta.y;
      v2->z=v1->z+delta.z;
   }

   if (c_texture_locking)
   {
      for (bs=s->bsel;bs;bs=bs->Next)
         TexLock(bs->Brush,bs->Brush->disp);
   }
   
   {
      const char *c;
      vec3_t v,v2;
      entity_t *e;
      char buf[32];

      for (es=s->esel;es;es=es->Next)
      {
         e=es->Entity;
         c=GetKeyValue(e->disp,"origin");
         if (!c) continue;

         v.x=v.y=v.z=0;
         sscanf(c,"%f %f %f",&v.x,&v.y,&v.z);
         v2.x=v.x+delta.x;
         v2.y=v.y+delta.y;
         v2.z=v.z+delta.z;

         snprintf(buf,sizeof(buf),"%0.0f %0.0f %0.0f",v2.x,v2.y,v2.z);
         SetKeyValue(e,"origin",buf);
      }
   }
}


void Trans_Center(const sel_t *s,vec3_t *center)
{
   vsel_t *vs;
   vec3_t min,max;

   vs=s->vsel;
   if (!vs)
   {
      center->x=center->y=center->z=0;
      return;
   }
   min=max=vs->Brush->verts[vs->num];

   for (vs=vs->Next;vs;vs=vs->Next)
   {
      VEC_CheckMin(min,vs->Brush->verts[vs->num]);
      VEC_CheckMax(max,vs->Brush->verts[vs->num]);
   }
   center->x=(min.x+max.x)/2;
   center->y=(min.y+max.y)/2;
   center->z=(min.z+max.z)/2;
}


void TransformLinear(matrix_t m,sel_t *s)
{
   sel_t t;

   if (!Trans_Setup(s,&t))
      return;
   Trans_Linear(m,&t);
   Trans_Done(&t,0);
}

void TransformMove(vec3_t delta,sel_t *s)
{
   sel_t t;

   if (!Trans_Setup(s,&t))
      return;
   Trans_Move(delta,&t);
   Trans_Done(&t,0);
}

