/*
uid.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#include "defines.h"
#include "types.h"

#include "uid.h"

#include "error.h"


/* 32 bits should be more than enough. If they do wrap around, it'll be detected,
and the user asked to restart Quest. It will then start searching for unused uid:s
and reusing them, so nothing bad should happen (besides a performance hit).
*/

#ifndef ALPHA
static unsigned int uid_brush =1;
static unsigned int uid_entity=1;
static unsigned int uid_group =1;
#else
/* For now (at least) make the uid:s unique for each map object to catch
cross-referencing errors. */
static unsigned int uid_u=1;
#endif


unsigned int GetBrushUID(void)
{
   unsigned int i;

   i=uid_u++;
   if (!uid_u)  /* wraparound, duplicate values, not good */
   { /* TODO: should be warning, but an important warning */
      Warning("GetBrushUID: UID overflow! You should restart Quest as soon as possible!");
      uid_u--;
      for (i=1;i<(1<<31);i++)
         if (!UID_Brush(i))
            return i;
      ASSERT(1);
   }
   return i;
}


unsigned int GetEntityUID(void)
{
   unsigned int i;

   i=uid_u++;
   if (!uid_u)  /* wraparound, duplicate values, not good */
   {
      Warning("GetEntityUID: UID overflow! You should restart Quest as soon as possible!");
      uid_u--;
      for (i=1;i<(1<<31);i++)
         if (!UID_Entity(i))
            return i;
      ASSERT(1);
   }
   return i;
}

unsigned int GetGroupUID(void)
{
   unsigned int i;

   i=uid_u++;
   if (!uid_u)  /* wraparound, duplicate values, not good */
   {
      Warning("GetEntityUID: UID overflow! You should restart Quest as soon as possible!");
      uid_u--;
      for (i=1;i<(1<<31);i++)
         if (!UID_Group(i))
            return i;
      ASSERT(1);
   }
   return i;
}


#include "brush.h"
#include "quest.h"

struct entity_s *UID_Entity(map_uid_t u)
{
   entity_t *e;
   for (e=M.EntityHead;e;e=e->Next)
      if (e->uid==u)
         return e;
   return NULL;
}

struct brush_s *UID_Brush(map_uid_t u)
{
   brush_t *b;
   for (b=M.BrushHead;b;b=b->Next)
      if (b->uid==u)
         return b;
   return NULL;
}

struct group_s *UID_Group(map_uid_t u)
{
   group_t *g;
   for (g=M.GroupHead;g;g=g->Next)
      if (g->uid==u)
         return g;
   return NULL;
}

